/***************************************************************************
                          arrayitem.h  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Sat Aug 28 1999
    copyright            : (C) 2017 by Werner Stille
    email                : kpl@werner-stille.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef ARRAYITEM_H
#define ARRAYITEM_H

#include <QUrl>
#include "scaleditem.h"

class QIODevice;
class QTextStream;

/**
  * Array item class. Class for array items showing data arrays by lines
  * or markers.
  * @author Werner Stille
  */
class Q_DECL_EXPORT ArrayItem : public ScaledItem
{
public:
  /**
    * Decimal character. Determines decimal character for reading of floating
    * point numbers.
    */
  enum DecimalCharacter {
    Point,    ///< Decimal point.
    Comma,    ///< Decimal comma.
    Settings  ///< Use Kpl Settings (only for internal use in Kpl application).
  };
  /** Specific QGraphicsItem type number for function type() */
  enum {
    Type = UserType + Array
    /**<
      * Specific QGraphicsItem type number of ArrayItem for function
      * type().
      */
  };
  /** Constructor. */
  ArrayItem();
  /**
    * Constructor. Initializes corresponding to arguments.
    * @param color color for data plot.
    * @param colX x column index.
    * @param colY y column index.
    * @param colErr y error column index.
    * @param errorBars true to plot error bars.
    * @param symbol symbol type.
    * @param decimalChar decimal symbol of floating point numbers used in file.
    * @param url URL of data file.
    * @param fx normalization factor for x values.
    * @param fy normalization factor for y values.
    * @param parent pointer to parent item.
    * @param n number of data points to be plotted.
    * @param rowStart start row index.
    * @param fillStyle fill style.
    * @param xOffset offset for x values.
    * @param yOffset offset for x values.
    * @param visible true for visible item.
    * @param internal true for internal stored array data.
    * @param relSize relative size of lines or symbols.
    */
  ArrayItem(unsigned color, int colX = 0, int colY = 0, int colErr = 0,
            bool errorBars = false, Symbol symbol = OpenCircle,
            DecimalCharacter decimalChar = Point, const QUrl& url = QUrl(),
            double fx = 1, double fy = 1, QGraphicsItem* parent = 0, int n = 0,
            int rowStart = 0, FillStyle fillStyle = NoBrush,
            double xOffset = 0, double yOffset = 0, bool visible = true,
            bool internal = false, double relSize = 1);
  /**
    * Copy constructor.
    * @param a source item.
    */
  ArrayItem(const ArrayItem& a);
  /**
    * Constructor. Initializes corresponding to plot file.
    * @param plo pointer to KSimpleConfig object containing item data.
    * @param urlPlo URL of plot file.
    * @param color default color for data plot.
    * @param colX default x column index.
    * @param colY default y column index.
    * @param colErr default y error column index.
    * @param errorBars default, true to plot error bars.
    * @param symbol default symbol type.
    * @param parent pointer to parent item.
    */
  ArrayItem(const KConfigGroup* plo, const QUrl& urlPlo, unsigned color = 0,
            int colX = 0, int colY = 0, int colErr = 0, bool errorBars = false,
            Symbol symbol = OpenCircle, QGraphicsItem* parent = 0);
  /** Destructor. */
  virtual ~ArrayItem();
  /**
    * Sets and limits the column indices.
    * @param colX x column.
    * @param colY y column.
    * @param colErr error column.
    */
  void adjustCols(int colX, int colY, int colErr);
  /** Limits the column indices. */
  inline void adjustCols() { adjustCols(m_colX, m_colY, m_colErr); };
  /** Calculates bounding rectangle. */
  virtual QRectF boundingRect() const;
  /** Returns y error column index. */
  inline int colErr() const { return m_colErr; };
  /** Returns x column index. */
  inline int colX() const { return m_colX; };
  /** Returns y column index. */
  inline int colY() const { return m_colY; };
  /**
    * Copies array storage.
    * @param data two-dimensional source array.
    * @param nCols number of columns.
    * @param nRows number of rows.
    * @return pointer to copy of array.
    */
  static double** copyArray(double **data, int nCols, int nRows);
  /** Returns number of data points to be plotted. */
  inline int count() const { return m_count; };
  /**
    * Returns pointer to internal array.
    * @return pointer to two-dimensional array.
    */
  inline double** data() const { return m_data; };
  /** Returns decimal symbol of floating point numbers. */
  inline DecimalCharacter decimalCharacter() const
  { return m_decimalCharacter; };
  /** Returns true for plotting error bars. */
  inline bool errorBars() const { return m_errorBars; };
  /**
   * Calculates normalization factors.
   * @param expX reference to logarithm of x normalization factor.
   * @param expY reference to logarithm of x normalization factor.
   * @param fx reference to x normalization factor.
   * @param fy reference to y normalization factor.
   */
  void exponent(int& expX, int& expY, double& fx, double& fy) const;
  /**
    * Deallocates array storage.
    * @param data pointer to two-dimensional array.
    */
  static void freeArray(double*** data);
  /** Returns true for internal stored array data. */
  inline bool internal() const { return m_internal; };
  /**
   * Calculates minimum and maximum values.
   * @param xMin reference to minimum of x values.
   * @param xMax reference to maximum of x values.
   * @param yMin reference to minimum of y values.
   * @param yMax reference to maximum of y values.
   */
  void limits(double& xMin, double& xMax, double& yMin, double& yMax) const;
  /** Returns number of columns. */
  inline int numCols() const { return m_numCols; };
  /** Returns number of rows. */
  inline int numRows() const { return m_numRows; };
  /** Paints item. */
  virtual void paint(QPainter* painter, const QStyleOptionGraphicsItem* option,
                     QWidget* widget = 0);
  /**
    * Reads double values from the URL.
    * @return number of rows, 0 for error.
    */
  virtual int readFile();
  /**
    * Reads double values from a file into an array.
    * @param url URL of data file.
    * @param nCols reference to number of columns.
    * @param data pointer to two-dimensional array.
    * @param decimalChar decimal symbol of floating point numbers used in file.
    * @return number of rows, 0 for error.
    */
  static int readFile(const QUrl& url, int& nCols, double*** data,
                      DecimalCharacter decimalChar = Point);
  /**
    * Reads double values from an I/O device into an array.
    * @param device I/O device to read from.
    * @param nCols reference to number of columns.
    * @param data pointer to two-dimensional array.
    * @param decimalChar decimal character: Point or Comma.
    * @return number of rows, 0 for error.
    */
  static int readFile(QIODevice* device, int& nCols, double*** data,
                      DecimalCharacter decimalChar = Point);
  /**
    * Reads double values from a string into an array.
    * @param text string to read from.
    * @param nCols reference to number of columns.
    * @param data pointer to two-dimensional array.
    * @param decimalChar decimal character: Point or Comma.
    * @return number of rows, 0 for error.
    */
  static int readFile(QString* text, int& nCols, double*** data,
                      DecimalCharacter decimalChar = Point);
  /**
    * Reads double values from a string.
    * @return number of rows, 0 for error.
    */
  int readFile(QString* text);
  /**
    * Reads double values from a text stream into an array.
    * @param t text stream to read from.
    * @param nCols number of columns.
    * @param data pointer to array.
    * @param decimalChar decimal character: Point or Comma.
    * @return number of lines, 0 for error.
    */
  static int readLines(QTextStream& t, int nCols, double*** data,
                       DecimalCharacter decimalChar = Point);
  /** Returns start row index. */
  inline int rowStart() const { return m_rowStart; };
  /**
    * Sets error column index.
    * @param colErr y error column index.
    */
  inline void setColErr(int colErr) { m_colErr = colErr; };
  /**
    * Sets x column index.
    * @param colX y error column index.
    */
  inline void setColX(int colX) { m_colX = colX; };
  /**
    * Sets y column index.
    * @param colY y column index.
    */
  inline void setColY(int colY) { m_colY = colY; };
  /**
    * Sets number of data points.
    * @param n number of data points.
    */
  inline void setCount(int n) { m_count = n; };
  /** Sets offset for dashed lines. */
  void setDashOffset();
  /**
    * Sets array pointer.
    * @param data two-dimensional array.
    */
  void setData(double** data);
  /**
    * Sets decimal character.
    * @param decimalChar decimal symbol of floating point numbers used in file.
    */
  inline void setDecimalCharacter(DecimalCharacter decimalChar)
  { m_decimalCharacter = decimalChar; };
  /**
    * Sets error bars plotting.
    * @param errorBars true to plot error bars.
    */
  inline void setErrorBars(bool errorBars) { m_errorBars = errorBars; };
  /**
    * Sets storage mode.
    * @param internal true for internal stored array data.
    */
  inline void setInternal(bool internal) { m_internal = internal; };
  /**
    * Sets number of columns.
    * @param nCols number of columns.
    */
  inline void setNumCols(int nCols) { m_numCols = nCols; };
  /**
    * Sets number of rows.
    * @param nRows number of Rows.
    */
  inline void setNumRows(int nRows) { m_numRows = nRows; };
  /**
    * Sets and limits the start row index and the number of points to be
    * plotted.
    * @param rowStart start row index.
    * @param n number of data points.
    */
  void setRange(int rowStart, int n);
  /**
    * Sets start row index.
    * @param rowStart start row index.
    */
  inline void setRowStart(int rowStart) { m_rowStart = rowStart; };
  /**
    * Sets URL.
    * @param url URL of data file.
    */
  inline void setUrl(const QUrl& url) { m_url = url; };
  /** Returns the shape of the plot item. */
  virtual QPainterPath shape() const;
  /**
   * Translates exponent and decimal symbols.
   * @param line line from data or parameter file.
   * @param decimalChar decimal symbol of floating point numbers used in file.
   */
  static void translate(QString& line, DecimalCharacter decimalChar);
  /** Returns specific QGraphicsItem type number. */
  virtual int type() const { return Type; };
  /** Returns URL of data file. */
  inline const QUrl& url() const { return m_url; };
  /**
   * Writes into plot file.
   * @param plo KConfigGroup object.
   * @param urlPlo URL of plot file.
   * @param absolute true to save absolute paths.
   */
  void writePlo(KConfigGroup* plo, const QUrl& urlPlo, bool absolute) const;

protected:
  /**
   * Paints or calculates bounding area.
   * @param painter Pointer to QPainter object, 0 for calculating bounding
   * area.
   */
  void paintCond(QPainter* painter) const;
  bool m_internal, m_errorBars;
  int m_colX, m_colY, m_colErr, m_rowStart, m_count, m_numRows, m_numCols;
  DecimalCharacter m_decimalCharacter;
  double** m_data;
  QUrl m_url;
};

#endif
