/***************************************************************************
                          fun3ditem.h  -  description
                             -------------------
    begin                : Tue Dec 25 2001
    copyright            : (C) 2017 by Werner Stille
    email                : kpl@werner-stille.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef FUN3DITEM_H
#define FUN3DITEM_H

#include "funitem.h"
#include "threeditem.h"

/**
  * 3D function item class. Provides methods for 3D plotting of functions.
  * @author Werner Stille
  */
class Q_DECL_EXPORT Fun3DItem : public ThreeDItem
{
public:
  /** Specific QGraphicsItem type number for function type() */
  enum {
    Type = UserType + Function3D
    /**< Specific QGraphicsItem type number of Fun3DItem for function
      * type().
      */
  };
  /**
    * Constructor. Initializes corresponding to arguments.
    * @param x0 left page margin in centimeters.
    * @param y0 bottom page margin in centimeters.
    * @param width width in centimeters.
    * @param height height in centimeters.
    * @param colorFrame color for frame and labels.
    * @param colorGrid color for grids and tics.
    * @param colorData color for data plot.
    * @param logX true for logarithmic x axis.
    * @param logY true for logarithmic y axis.
    * @param logZ true for logarithmic y axis.
    * @param gridMode grid mode.
    * @param mode3D 3D mode.
    * @param xMax maximum x value.
    * @param yMax maximum y value.
    * @param zMax maximum z value.
    * @param textX x axis text.
    * @param textY y axis text.
    * @param textZ z axis text.
    * @param header heading text.
    * @param relSize relative size of text, markers and lines.
    * @param numDigX number of digits for x axis.
    * @param numDigY number of digits for y axis.
    * @param numDigZ number of digits for z axis.
    * @param minTicX number of intervals between major tics of x axis.
    * @param minTicY number of intervals between major tics of y axis.
    * @param minTicZ number of intervals between major tics of z axis.
    * @param expX logarithm of normalization factor for x axis.
    * @param expY logarithm of normalization factor for y axis.
    * @param expZ logarithm of normalization factor for z axis.
    * @param xMin minimum x value.
    * @param yMin minimum y value.
    * @param zMin minimum z value.
    * @param xTic distance between major tics of x axis.
    * @param yTic distance between major tics of y axis.
    * @param zTic distance between major tics of z axis.
    * @param deltaX x step size.
    * @param deltaZ z step size.
    * @param phi angle of rotation about y axis in deg.
    * @param theta tilt angle of y axis in deg.
    * @param bx x box width in cm.
    * @param by y box width in cm.
    * @param bz z box width in cm.
    * @param xr x position of center of box bottom plane.
    * @param yr y position of center of box bottom plane.
    * @param fx normalization factor for x values.
    * @param fy normalization factor for y values.
    * @param fz normalization factor for z values.
    * @param name function name.
    * @param url URL of the function library.
    * @param showFrame true to show frame.
    * @param collapsed true for collapsed in list view.
    * @param visible true for visible item.
    */
  Fun3DItem(double x0 = 4, double y0 = 3, double width = 15,
            double height = 10, unsigned colorFrame = 0,
            unsigned colorGrid = 0, unsigned colorData = 0, bool logX = false,
            bool logY = false, bool logZ = false,
            GridMode gridMode = AxesWithLabels, Mode3D mode3D = Grid,
            double xMax = 15, double yMax = 10, double zMax = 15,
            const QString& textX = "", const QString& textY = "",
            const QString& textZ = "", const QString& header = "",
            double relSize = 1, int numDigX = -1, int numDigY = -1,
            int numDigZ = -1, int minTicX = 5, int minTicY = 2,
            int minTicZ = 5, int expX = 0, int expY = 0, int expZ = 0,
            double xMin = 0, double yMin = 0, double zMin = 0, double xTic = 5,
            double yTic = 2, double zTic = 5, double deltaX = 0,
            double deltaZ = 0,
            double phi = 20, double theta = -20, double bx = 7.5,
            double by = 7.5, double bz = 7.5, double xr = 7.5, double yr = 4,
            double fx = 1, double fy = 1, double fz = 1,
            const QString& name = "", const QUrl& url = QUrl(),
            bool showFrame = true, bool collapsed = false,
            bool visible = true);
  /** Copy constructor. */
  Fun3DItem(const Fun3DItem& f);
  /**
    * Constructor. Initializes corresponding to plot file.
    * @param plo pointer to KConfigGroup object containing item data.
    * @param urlPlo URL of plot file.
    * @param x0 default value for left page margin in centimeters.
    * @param y0 default value for bottom page margin in centimeters.
    * @param width default value for width in centimeters.
    * @param height default value for height in centimeters.
    * @param colorFrame default value for color for frame and labels.
    * @param colorGrid default value for color for grids and tics.
    * @param colorData color for data plot.
    * @param gridMode default grid mode.
    */
  Fun3DItem(const KConfigGroup* plo, const QUrl& urlPlo, double x0 = 4,
            double y0 = 3, double width = 15, double height = 10,
            unsigned colorFrame = 0, unsigned colorGrid = 0,
            unsigned colorData = 0, GridMode gridMode = AxesWithLabels);
  /** Destructor. */
  ~Fun3DItem();
  /**
   * Calculates function value table.
   * @return number of data points.
   */
  int calcTable() const;
  /**
   * Exports function value table to file.
   * @param stream stream for output file.
   * @param separator column separator character.
   * @param format C format character.
   * @param precision precision.
   */
  void exportTable(QTextStream& stream, char separator, char format,
                   int precision) const;
  double (*funcY)(double, double, const double*);
  /** Returns pointer to library for y function. */
  inline const QLibrary* libY() const { return m_libY; };
  /** Returns y function name. */
  inline QString nameY() const { return m_nameY; };
  /** Paints item. */
  virtual void paint(QPainter* painter, const QStyleOptionGraphicsItem* option,
                     QWidget* widget = 0);
  /** Returns y function parameter array pointer. */
  inline double* parY() { return m_parY; };
  /**
   * Resolves y function address.
   * @param url URL of the function library.
   * @param name function name.
   * @return true for success.
   */
  bool setFuncY(const QUrl& url, const QString& name);
  /**
    * Sets parameter.
    * @param i parameter index.
    * @param value value.
    */
  inline void setPar(int i, double value) { m_parY[i] = value; };
  /** Returns specific QGraphicsItem type number. */
  virtual int type() const { return Type; };
  /** Applies changes and forces redrawing. */
  void update();
  /** Returns URL of y function library. */
  inline const QUrl& urlY() const { return m_urlY; };
  /**
   * Writes into plot file.
   * @param plo KConfigGroup object.
   * @param urlPlo URL of plot file.
   * @param absolute true to save absolute paths.
   * @param format format character for value display.
   * @param precision precision (number of digits) for value display.
   */
  void writePlo(KConfigGroup* plo, const QUrl& urlPlo, bool absolute,
                char format, int precision) const;

protected:
  /** Unloads y function library. */
  void unload();
  mutable double (*funcYo)(double, double, const double*);
  double m_parY[FunItem::NumParMax];
  mutable double parYo[FunItem::NumParMax];
  QString m_nameY;
  QUrl m_urlY;
  QLibrary* m_libY;
};

#endif
