/***************************************************************************
                          funitem.h  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Sun Aug 29 1999
    copyright            : (C) 2017 by Werner Stille
    email                : kpl@werner-stille.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef FUNITEM_H
#define FUNITEM_H

#include <QVector>
#include <QUrl>
#include "scaleditem.h"

class QLibrary;
class QTextStream;
class ArrayItem;
class LMFit;

/**
  * Function item class. Class for function items showing ordinary functions
  * by lines or markers.
  * @author Werner Stille
  */
class Q_DECL_EXPORT FunItem : public ScaledItem
{
public:
  /** Dimension of parameter arrays. */
  static const int NumParMax = 20;
  /** Specific QGraphicsItem type number for function type() */
  enum {
    Type = UserType + Function
    /**< Specific QGraphicsItem type number of FunItem for function type(). */
  };
/**
    * Data error structure. Contains data error settings for parameter fits.
    */
  struct DataErrorStruct {
    /**
      * Error model function argument. Column of array item to be used as
      * argument for the error model function.
      */
    enum ErrorModelArgument {
      ColX, ///< x column.
      ColY  ///< y column.
    };
    /** Initializes structure. */
    DataErrorStruct() : errModArg(ColY) {};
    /**
      * True for using error column of array item instead of error model
      * function.
      */
    bool fitErrCol;
    /** Error model function argument. */
    ErrorModelArgument errModArg;
    /** Parameter array for error model function. */
    double parErrMod[NumParMax];
    /** Name of error model function. */
    QString errModName;
    /** Path of library containing error model function. */
    QString errModPath;
  };
  /** Constructor. */
  FunItem();
  /**
    * Copy constructor.
    * @param a source item.
    */
  FunItem(const FunItem& a);
  /**
    * Constructor. Initializes corresponding to plot file.
    * @param plo pointer to KConfigGroup object containing item data.
    * @param urlPlo URL of plot file.
    * @param color default color for function plot.
    * @param symbol default symbol type.
    * @param parent pointer to parent item.
    */
  FunItem(const KConfigGroup* plo, const QUrl& urlPlo, unsigned color = 0,
          Symbol symbol = SolidLine, QGraphicsItem* parent = 0);
  /**
    * Constructor. Initializes corresponding to arguments.
    * @param url URL of the function library.
    * @param name function name.
    * @param color color for function plot.
    * @param symbol symbol type.
    * @param fx normalization factor for x values.
    * @param fy normalization factor for y values.
    * @param parent pointer to parent item.
    * @param xMin minimum argument value.
    * @param xMax maximum argument value.
    * @param deltaX argument step size, 0 for automatic.
    * @param visible true for visible item.
    * @param fillStyle fill style.
    * @param xOffset offset for x values.
    * @param yOffset offset for x values.
    * @param relSize relative size of lines or symbols.
    */
  FunItem(const QUrl& url, const QString& name, unsigned color = 0,
          Symbol symbol = SolidLine, double fx = 1, double fy = 1,
          QGraphicsItem* parent = 0, double xMin = 0, double xMax = 0,
          double deltaX = 0, bool visible = true,
          FillStyle fillStyle = NoBrush, double xOffset = 0,
          double yOffset = 0, double relSize = 1);
  /** Destructor. */
  virtual ~FunItem();
  /** Calculates bounding rectangle. */
  virtual QRectF boundingRect() const;
  /**
    * Calculates function value table.
    * @param x array with argument values.
    * @param values array for function values.
    * @param func function.
    * @param par parameter array.
    * @param n number of values.
    */
  static void calcTable(const double* x, double* values,
                        double (*func)(double, const double*),
                        const double* par, int n);
  /**
   * Calculates function value table.
   * @return number of rows.
   */
  int calcTable() const;
  /**
   * Calculates function value table.
   * @param logX true for logarithmic steps
   * @return number of rows.
   */
  int calcTable(bool logX) const;
  /** Returns argument step size. */
  inline double deltaT() const { return m_deltaT; };
  /**
   * Checks data error structures for differences.
   * @param e1 first data error structure.
   * @param e2 second data error structure.
   * @return true for different structure contents.
   */
  static bool errorStructChanged(const DataErrorStruct& e1,
                                 const DataErrorStruct& e2);
  /**
   * Calculates normalization factors.
   * @param expX reference to logarithm of x normalization factor.
   * @param expY reference to logarithm of x normalization factor.
   * @param fx reference to x normalization factor.
   * @param fy reference to y normalization factor.
   * @param logX true for logarithmic x axis.
   */
  void exponent(int& expX, int& expY, double& fx, double& fy, bool logX) const;
  /**
   * Exports function value table to file.
   * @param stream stream for output file.
   * @param separator column separator character.
   * @param format C format character.
   * @param precision number of digits.
   */
  void exportTable(QTextStream& stream, char separator, char format,
                   int precision) const;
  /**
   * Performs fit of parameters of a set of functions to a set of arrays.
   * @param aList array item list.
   * @param fList function item list.
   * @param bFit array containing true values for parameters to be fitted.
   * @param par parameter array.
   * @param resVector array for weighted residuals.
   * @param sigma array for data errors.
   * @param err array of structures containing the data error settings.
   * @param nonLin true for nonlinear fit.
   * @param tol tolerance value for stopping nonlinear fit iterations.
   * @param itMax maximum number of nonlinear fit iterations.
   * @param lm object for Levenberg-Marquardt fits.
   * @param chiSquare chi-square.
   * @param corr array for parameter correlation matrix.
   * @param parErr array for parameter erros.
   * @param avgErr average error.
   * @param parent parent object for lm.
   * @return error code.
   */
  static int fit(const QList<const ArrayItem*>* aList, QList<FunItem*>* fList,
                 const bool* bFit, double* par,
                 QVector<double>* resVector = 0, QVector<double>* sigma = 0,
                 FunItem::DataErrorStruct* err = 0, bool nonLin = true,
                 double tol = 1e-14, int itMax = 100, LMFit* lm = 0,
                 double* chiSquare = 0, double* corr = 0, double* parErr = 0,
                 double* avgErr = 0, QObject* parent = 0);
  /** y function. */
  double (*funcY)(double, const double*);
  /** Returns pointer to library for y function. */
  inline const QLibrary* libY() const { return m_libY; };
  /**
   * Calculates minimum and maximum values.
   * @param xMin reference to minimum of x values.
   * @param xMax reference to maximum of x values.
   * @param yMin reference to minimum of y values.
   * @param yMax reference to maximum of y values.
   * @param logX true for logarithmic x axis.
   */
  void limits(double& xMin, double& xMax, double& yMin, double& yMax,
              bool logX) const;
  /** Returns y function name. */
  inline QString nameY() const { return m_nameY; };
  /** Paints item. */
  virtual void paint(QPainter* painter, const QStyleOptionGraphicsItem* option,
                     QWidget* widget = 0);
  /**
   * Checks parameter arrays for differences.
   * @param p1 first parameter array.
   * @param p2 second parameter array.
   * @return true for different array contents.
   */
  static bool parChanged(const double* p1, const double* p2);
  /**
   * Converts parameter array to string.
   * @param par parameter array.
   * @param format C format character.
   * @param precision number of digits.
   * @return space separated string.
   */
  static QString parToString(const double* par, char format, int precision);
  /** Returns y function parameter array pointer. */
  inline double* parY() { return m_parY; };
  /** Sets offset for dashed lines. */
  void setDashOffset();
  /**
    * Sets argument step size.
    * @param deltaT step size.
    */
  inline void setDeltaT(double deltaT) { m_deltaT = deltaT; };
  /**
   * Resolves y function address.
   * @param url url of the function library.
   * @param name function name.
   * @return true for success.
   */
  bool setFuncY(const QUrl& url, const QString& name);
  /**
    * Sets parameter.
    * @param i parameter index.
    * @param value value.
    * @param yFunc true for y value function.
    */
  virtual void setPar(int i, double value, bool yFunc = true);
  /**
    * Sets maximum argument value.
    * @param xMax maximum value.
    */
  inline void setTMax(double xMax) { m_tMax = xMax; };
  /**
    * Sets minimum argument value.
    * @param xMin maximum value.
    */
  inline void setTMin(double xMin) { m_tMin = xMin; };
  /** Returns the shape of the plot item. */
  virtual QPainterPath shape() const;
  /**
   * Converts string to parameter array values.
   * @param s string.
   * @param par parameter array.
   */
  static void stringToPar(const QString& s, double* par);
  /** Returns maximum argument value. */
  inline double tMax() const { return m_tMax; };
  /** Returns minimum argument value. */
  inline double tMin() const { return m_tMin; };
  /** Returns pointer to argument value array. */
  const double* tValues() const;
  /** Returns specific QGraphicsItem type number. */
  virtual int type() const { return Type; };
  /** Returns URL of y function library. */
  inline const QUrl& urlY() const { return m_urlY; };
  /**
   * Writes into plot file.
   * @param plo KConfigGroup object.
   * @param urlPlo URL of plot file.
   * @param absolute true to save absolute paths.
   * @param format C format character.
   * @param precision number of digits.
   */
  void writePlo(KConfigGroup* plo, const QUrl& urlPlo, bool absolute,
                char format, int precision) const;
  /** Returns pointer to y function value array. */
  double* yValues() const;

protected:
  /** Performs initialization. */
  void init();
  /**
   * Paints or calculates bounding area.
   * @param painter QPainter object, 0 for calculating bounding area.
   */
  void paintCond(QPainter* painter) const;
  /** Unloads y function library. */
  void unload();
  mutable bool logXo;
  double m_tMin, m_tMax, m_deltaT;
  mutable double tMino, tMaxo, deltaTo;
  mutable double (*funcYo)(double, const double*);
  double m_parY[NumParMax];
  mutable double parYo[NumParMax];
  mutable QVector<double> m_tValues, m_yValues;
  QString m_nameY;
  QUrl m_urlY;
  QLibrary* m_libY;
};

#endif
