/***************************************************************************
                          imageitem.cpp  -  description
                             -------------------
    begin                : Don Jun 5 2003
    copyright            : (C) 2017 by Werner Stille
    email                : kpl@werner-stille.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <QFile>
#include <QPainter>
#include <QPixmapCache>
#include <QProcess>
#include <QStyleOptionGraphicsItem>
#include <KConfigGroup>
#include "imageitem.h"
#include "kplscene.h"
#include "utils.h"

/**
 * Reads bounding box coordinates from an EPS file.
 * @param filename path of EPS file.
 * @param x1 left x value.
 * @param y1 bottom y value.
 * @param x2 right x value.
 * @param y2 top y value.
 * @return true for success.
 */
static bool boundingBox(QString& filename, double* x1, double* y1, double* x2,
                        double* y2)
{
  bool success = false;
  QFile f(filename);
  if (f.open(QIODevice::ReadOnly))
    while (!f.atEnd()) {
      QString line = f.readLine();
      if (line.indexOf("%%BoundingBox:") >= 0) {
        QStringList list = line.mid(14).trimmed().split(" ",
          QString::SkipEmptyParts);
        if (list.count() == 4) {
          *x1 = list[0].toDouble();
          *y1 = list[1].toDouble();
          *x2 = list[2].toDouble();
          *y2 = list[3].toDouble();
          if ((*x2 > *x1) && (*y2 > *y1))
            success = true;
        }
        break;
      }
    }
  return success;
}

ImageItem::ImageItem(double x0, double y0, QGraphicsItem* parent,
                     const QUrl& url, bool visible, double scaleX,
                     double scaleY, double dir) :
 DecoItem(0, x0, y0, 1), QGraphicsPixmapItem(parent), m_scaleX(scaleX),
 m_scaleY(scaleY), m_dir(dir), m_url(url)
{
  if (!visible)
    hide();
  setFlag(QGraphicsItem::ItemIsMovable, true);
  QPixmapCache::clear();
}

ImageItem::ImageItem(const ImageItem& a) :
 DecoItem(a), QGraphicsPixmapItem(), m_scaleX(a.m_scaleX),
 m_scaleY(a.m_scaleY), m_dir(a.m_dir), m_url(a.m_url)
{
  setZValue(a.zValue());
  setFlags(QGraphicsItem::ItemIsMovable | QGraphicsItem::ItemIsSelectable);
  setSelected(a.isSelected());
  setVisible(a.isVisible());
}

ImageItem::ImageItem(const KConfigGroup* plo, const QUrl& urlPlo,
                     QGraphicsItem* parent) :
 DecoItem(plo), QGraphicsPixmapItem(parent)
{
  if (!plo->readEntry("active", true))
    hide();
  m_url = absoluteUrl(plo->readEntry("path", ""), urlPlo);
  m_scaleX = plo->readEntry("fx", 1.0);
  m_scaleY = plo->readEntry("fy", 1.0);
  m_dir = plo->readEntry("dir", 0.0);
  setFlags(QGraphicsItem::ItemIsMovable | QGraphicsItem::ItemIsSelectable);
  QPixmapCache::clear();
}

ImageItem::~ImageItem()
{
}

bool ImageItem::boundingBox(const QUrl& url, QString& filename, double* x1,
                            double* y1, double* x2, double* y2)
{
  bool success = false;
  if (url.path().isEmpty())
    return success;
  if (url.isLocalFile())
    filename = url.path();
  else
    if (!Utils::download(url, filename, 0))
      return success;
  success = ::boundingBox(filename, x1, y1, x2, y2);
  if (!success && !url.isLocalFile())
    QFile::remove(filename);
  return success;
}

bool ImageItem::checkBoundingBox(const QUrl& url)
{
  QString filename;
  double x1, y1, x2, y2;
  return boundingBox(url, filename, &x1, &y1, &x2, &y2);
}

void ImageItem::paint(QPainter* painter,
                      const QStyleOptionGraphicsItem* option, QWidget* widget)
{
  QPaintDevice* device = painter->device();
  if (device) {
    QPixmap pm;
    int dpiX = qMin(150, device->logicalDpiX());
    int dpiY = qMin(150, device->logicalDpiY());
    renderPS(&pm, dpiX, dpiY);
    QWidget w;
    int rx = w.logicalDpiX();
    int ry = w.logicalDpiY();
    if ((dpiX != rx) || (dpiY != ry)) {
      QRectF r = QRectF(pm.rect());
      QRectF r2 = QRectF(0, 0, r.width() * rx / dpiX, r.height() * ry / dpiY);
      painter->drawPixmap(r2.translated(0, -r2.height()), pm, r);
      return;
    }
  }
  QGraphicsPixmapItem::paint(painter, option, widget);
}

void ImageItem::renderPS(QPixmap* pm, int dpiX, int dpiY)
{
  double xScale = m_scaleX * dpiX / 72;
  double yScale = m_scaleY * dpiY / 72;
  QString key = m_url.url() + QString("_%1x%2").arg(xScale).arg(yScale);
  if (QPixmapCache::find(key, pm) &&
      (pm->cacheKey() != QGraphicsPixmapItem::pixmap().cacheKey())) {
    setPixmap(*pm);
    setOffset(0, -pm->height());
  }
  if (QGraphicsPixmapItem::pixmap().isNull() ||
      (pm->cacheKey() != QGraphicsPixmapItem::pixmap().cacheKey())) {
    QString filename;
    if (m_url.path().isEmpty())
      return;
    if (m_url.isLocalFile())
      filename = m_url.path();
    else
      if (!Utils::download(m_url, filename, 0))
        return;
    double x0, y0, x1, y1;
    if (::boundingBox(filename, &x0, &y0, &x1, &y1)) {
      QString imageFile = Utils::tempFileName();
      renderPS(filename, imageFile, qRound(xScale * (x1 - x0)),
               qRound(yScale * (y1 - y0)), x0, y0, xScale, yScale, "ppm");
      pm->load(imageFile, "ppm");
      QFile::remove(imageFile);
      setPixmap(*pm);
      setOffset(0, -pm->height());
      QPixmapCache::insert(key, *pm);
    }
    if (!m_url.isLocalFile())
      QFile::remove(filename);
  }
}

void ImageItem::renderPS(const QString& psFile, const QString& imageFile,
                         int width, int height, double x0, double y0,
                         double xScale, double yScale, const char format[])
{
  QProcess p;
  p.start(QString("gs -sOutputFile=%1 -q -g%2x%3 -dSAFER -dBATCH -dNOPAUSE "
                  "-sDEVICE=%4 -dGraphicsAlphaBits=4"
                  " -dTextAlphaBits=4 -").arg(imageFile).arg(width).
                                          arg(height).arg(format));
  p.waitForStarted();
  p.write(QString("\n%1 %2 translate\n%3 %4 scale\n")
          .arg(-x0 * xScale).arg(-y0 * yScale)
          .arg(xScale).arg(yScale).toLatin1());
  QFile f(psFile);
  f.open(QIODevice::ReadOnly);
  p.write(f.readAll());
  p.closeWriteChannel();
  p.waitForFinished(-1);
}

void ImageItem::setProperties(const QTransform& m)
{
  m_matrix = m;
  setProperties();
}

void ImageItem::update()
{
  prepareGeometryChange();
  setProperties();
  QGraphicsItem::update();
}

void ImageItem::writePlo(KConfigGroup* plo, const QUrl& urlPlo, bool abs) const
{
  plo->writeEntry("Type", "IMAGEITEM");
  if (!isVisible())
    plo->writeEntry("active", false);
  plo->writeEntry("x", m_x0);
  plo->writeEntry("y", m_y0);
  plo->writeEntry("fx", m_scaleX);
  plo->writeEntry("fy", m_scaleY);
  plo->writeEntry("path", path(urlPlo, m_url, abs));
  if (m_dir)
    plo->writeEntry("dir", m_dir);
}

void ImageItem::mouseReleaseEvent(QGraphicsSceneMouseEvent* event)
{
  QGraphicsPixmapItem::mouseReleaseEvent(event);
  ((KplScene*)scene())->updatePos(this);
}

void ImageItem::setProperties()
{
  setPos(m_matrix.map(QPointF(m_x0, m_y0)));
  setRotation(-m_dir);
  QPixmap pm;
  QWidget w;
  renderPS(&pm, w.logicalDpiX(), w.logicalDpiY());
}
