/***************************************************************************
                          kplitem.h  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Wed Aug 25 1999
    copyright            : (C) 2017 by Werner Stille
    email                : kpl@werner-stille.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef KPLITEM_H
#define KPLITEM_H

#include <QBrush>
#include <QPen>
#include <QPixmap>
#include <QString>

class QGraphicsItem;
class QPainter;
class QPointF;
class QRectF;
class KConfigGroup;
class QUrl;

/**
  * Kpl item class. Base class for plot items.
  * @author Werner Stille
  */
class Q_DECL_EXPORT KplItem
{
public:
  /** Fill styles. Types of fill styles for filled geometric forms. */
  enum FillStyle {
    NoBrush,         ///< No brush pattern.
    SolidPattern,    ///< Uniform color.
    HorPattern,      ///< Horizontal lines.
    VerPattern,      ///< Vertical lines.
    CrossPattern,    ///< Crossing horizontal and vertical lines.
    BDiagPattern,    ///< Backward diagonal lines.
    FDiagPattern,    ///< Forward diagonal lines.
    DiagCrossPattern ///< Crossing diagonal lines.
  };
  /** Item types. Types of plot items. */
  enum ItemType {
    Frame = 1,   ///< Frame item with axes or grid.
    Array,       ///< Item with array of 2D data points.
    Function,    ///< 2D function item.
    ParFunction, ///< 2D parametric function item.
    Spline,      ///< 2D spline interpolation item.
    Array3D,     ///< Item with array of 3D data points.
    Function3D,  ///< 3D function item.
    Legend,      ///< Legend item for explaining of symbols or lines.
    Text,        ///< Text item.
    Line,        ///< Item with straight line.
    Arrow,       ///< Item with single or double-headed arrow.
    Arc,         ///< Item with circular or elliptical arc.
    Rectangle,   ///< Item with rectangle or square.
    Ellipse,     ///< Item with ellipse or circle.
    Image,       ///< Item with EPS image from external file.
    ScaleBar     ///< Scalebar item.
  };
  /**
   * Pen types. Pens with different relative widths corresponding to different
   * parts of the plot. */
  enum PenType {
    FramePen, ///< For rectangular frame.
    GridPen,  ///< For grids and tics.
    DataPen   ///< For array and function items.
  };
  /** Symbol types. Types of plot symbols and lines. */
  enum Symbol {
    Invisible = -17, ///< No painting.
    FilledOctagon,   ///< Filled octagon.
    FilledHourglass, ///< Filled hourglass.
    FilledTriangle,  ///< Filled triangle.
    FilledSquare,    ///< Filled square.
    FilledDiamond,   ///< Filled diamond.
    FilledCircle,    ///< Filled circle.
    Umbrella,        ///< Umbrella-like shape.
    Ypsilon,         ///< Ypsilon-like shape.
    Plus,            ///< Plus sign cross.
    Cross,           ///< Diagonal cross.
    OpenOctagon,     ///< Open octagon.
    OpenHourglass,   ///< Open hourglass.
    OpenTriangle,    ///< Open triangle.
    OpenSquare,      ///< Open square.
    OpenDiamond,     ///< Open Diamond
    Dot,             ///< Dot-like filled circle.
    OpenCircle,      ///< Open circle.
    SolidLine,       ///< Solid line.
    DashLine1,       ///< Line with wide dashes and spaces.
    DashLine2,       ///< Line with short dashes and wide spaces.
    DotLine1,        ///< Dotted line with wide spaces.
    DashLine3,       ///< Line with wide dashes and short spaces.
    DashDotLine,     ///< Dash-dotted line.
    DashLine4,       ///< Line with wide and short dashes.
    DashLine5,       ///< Line with short dashes and short spaces.
    DotLine2         ///< Dotted line with short spaces.
  };
  /**
    * Constructor. Initializes corresponding to arguments.
    * @param relSize relative size of text and symbols.
    */
  KplItem(double relSize = 1);
  /**
    * Copy constructor.
    * @param a source item.
    */
  KplItem(const KplItem& a);
  /**
   * Constructor. Initializes corresponding to plot file.
   * @param plo KSimpleConfig object containing item data.
   */
  KplItem(const KConfigGroup* plo);
  /** Destructor. */
  virtual ~KplItem();
  /**
   * Constructs absolute file URL.
   * @param path path of file (can be relative).
   * @param urlPlo URL of referring plot file.
   * @return resulting absolute URL of file.
   */
  static QUrl absoluteUrl(const QString& path, const QUrl& urlPlo);
  /**
   * Creates brush.
   * @param color brush color.
   * @param fillStyle fill style.
   * @return created brush.
   */
  static QBrush createBrush(unsigned color, FillStyle fillStyle = NoBrush);
  /**
   * Creates pen.
   * @param width pen width in pixel units.
   * @param color pen color.
   * @param penType pen type.
   * @param symbol symbol type.
   * @return created pen.
   */
  static QPen createPen(double width, unsigned color, PenType penType,
                        Symbol symbol);
  /**
   * Paints symbol.
   * @param painter pointer to QPainter.
   * @param rect bounding rectangle.
   * @param symbol symbol type.
   */
  static void paintSymbol(QPainter* painter, const QRectF& rect,
                          Symbol symbol = OpenCircle);
  /**
   * Returns black decoration pixmap of plot symbol for use in dialog GUI.
   * @param symbol symbol type.
   * @return pixmap.
   */
  static QPixmap pixmap(Symbol symbol);
  /**
   * Returns decoration pixmap of plot symbol for use in item lists.
   * @param pen pen with symbol color.
   * @param symbol symbol type.
   * @return pixmap.
   */
  static QPixmap pixmap(QPen pen, Symbol symbol = SolidLine);
  /**
    * Initializes corresponding to plot file.
    * @param plo pointer to KConfigGroup object containing item data.
    */
  void readPlo(const KConfigGroup* plo);
  /**
   * Constructs absolute or (if possible) relative file path.
   * @param urlPlo URL of referring plot file.
   * @param url file URL.
   * @param absolute construct absolute path if true.
   * @return resulting path.
   */
  static QString path(const QUrl& urlPlo, const QUrl& url,
                      bool absolute = false);
  /** Returns relative size of item. */
  inline double relSize() const { return m_relSize; };
  /** Resets offset for dashed lines. */
  static void resetDashOffset();
  /**
   * Converts rgb string to QRgb.
   * @param color hexadecimal coded Qt1 rgb value.
   * @return QRgb value for actual Qt version.
   */
  static unsigned rgb(const QString& color);
  /**
   * Converts QRgb to Qt1 QRgb.
   * @param rgb QRgb value of actual Qt version.
   * @return QRgb value for actual Qt1 version.
   */
  static unsigned rgbQt1(unsigned rgb);
  /**
    * Sets relative size of item.
    * @param relSize relative size.
    */
  inline void setRelSize(double relSize) { m_relSize = relSize; };
  /**
   * Writes into plot file.
   * @param plo KConfigGroup object.
   * @param visible false for hidden item.
   */
  void writePlo(KConfigGroup* plo, bool visible = true) const;

protected:
  /**
   * Changes pen width.
   * @param pen reference to pen.
   * @param width pen width in pixel units.
   * @return reference to changed pen.
   */
  QPen& changePenWidth(QPen& pen, double width) const;
  /**
   * Creates pen for plot item.
   * @param item plot item.
   * @param color pen color.
   * @param penType pen type.
   * @param symbol symbol type.
   * @return created pen.
   */
  QPen createPen(QGraphicsItem* item, unsigned color = 0,
                 PenType penType = DataPen, Symbol symbol = SolidLine) const;
  /**
   * Returns width of item's parent frame or width of scene if the item has no
   * parent item.
   * @param item plot item.
   * @return width in pixel units.
   */
  static double parentWidth(const QGraphicsItem* item);
  /**
   * Calculates pen width for plot item.
   * @param item plot item.
   * @param penType pen type.
   * @return width in pixel units.
   */
  double penWidth(const QGraphicsItem* item, PenType penType = DataPen) const;
  /**
   * Returns factor to calculate pen width from frame width.
   * @param penType pen type.
   * @return factor.
   */
  static double penWidthFactor(PenType penType);
  static double m_dashOffset;
  double m_relSize;
};

#endif
