/***************************************************************************
                          kplview.h  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Sat Apr 24 15:14:00 MEST 1999

    copyright            : (C) 2017 by Werner Stille
    email                : kpl@werner-stille.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef KPLVIEW_H
#define KPLVIEW_H

#include <QGraphicsView>
#include <QPoint>
#include <QUrl>

class QRectF;
class QRubberBand;

/**
  * Kpl main view class. Displays the plot items of a KplScene. Provides
  * methods for scrolling and zooming.
  * @author Werner Stille
  */
class Q_DECL_EXPORT KplView : public QGraphicsView
{
  Q_OBJECT
public:
  /**
   * Constructor.
   * @param f zoom factor.
   * @param parent pointer to parent widget.
   */
  KplView(double f = 1, QWidget* parent = 0);
  /** Destructor. */
  virtual ~KplView();
  /**
   * Sets zoom factor.
   * @param f zoom factor.
   */
  void setZoomFactor(double f);
  /** Calculates recommended size of the widget. */
  virtual QSize sizeHint() const;
  /** Returns zoom factor. */
  inline double zoomFactor() const { return m_zoomFactor; };

signals:
  /** Emitted when mouse cursor leaves the widget. */
  void mouseLeaved();
  /**
   * Emitted at drop event.
   * @param list pointer to list of dropped URLs.
   */
  void urlsDropped(QList<QUrl>* list);
  /**
   * Emitted when rectangular region was selected using the mouse with its
   * right button pressed.
   * @param rect rectangle in pixel units.
   * units.
   */
  void regionSelected(const QRectF& rect);

protected:
  /** @param e pointer to event. */
  virtual void dragEnterEvent(QDragEnterEvent* e);
  /** @param e pointer to event. */
  virtual void dragMoveEvent(QDragMoveEvent* e);
  /**
   * Emits the signal #urlsDropped, transferring a list of URLs.
   * @param e pointer to event.
   */
  virtual void dropEvent(QDropEvent* e);
  /**
   * Implements scrolling of the window by key sequences.<ul>
   * <li><b>Home</b> scrolls to bottom margin.
   * <li><b>End</b> scrolls to top margin.
   * <li><b>Ctrl+Home</b> scrolls to left margin.
   * <li><b>Ctrl+End</b> scrolls to right margin.
   * <li><b>PgUp</b> scrolls page up.
   * <li><b>PgDown</b> scrolls page down.
   * </ul>
   * @param e pointer to event.
   */
  virtual void keyPressEvent(QKeyEvent* e);
  /**
   * Emits the signal #mouseLeaved.
   * @param e pointer to event.
   */
  virtual void leaveEvent(QEvent* e);
  /**
   * Implements selection of a rectangular region using the mouse with its
   * right button pressed.
   * @param e pointer to event.
   */
  virtual void mouseMoveEvent(QMouseEvent* e);
  /**
   * Implements initiation of selection of a rectangular region using the mouse
   * with its right button pressed.
   * @param e pointer to event.
   */
  virtual void mousePressEvent(QMouseEvent* e);
  /**
   * Implements finishing of selection of a rectangular region using the mouse
   * with its right button pressed. In this case the signal #regionSelected is
   * emitted, transferring the rectangle.
   * @param e pointer to event.
   */
  virtual void mouseReleaseEvent(QMouseEvent* e);
  double m_zoomFactor;
  QPoint startPos;
  QRubberBand* rubberBand;
};

#endif // KPLVIEW_H
