/***************************************************************************
                          parfunitem.h  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Sun Aug 29 1999
    copyright            : (C) 2017 by Werner Stille
    email                : kpl@werner-stille.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef PARFUNITEM_H
#define PARFUNITEM_H

#include "funitem.h"

/**
  * Parametric function item class. Class for function items showing parametric
  * functions by lines or markers.
  * @author Werner Stille
  */
class Q_DECL_EXPORT ParFunItem : public FunItem
{
public:
  /** Specific QGraphicsItem type number for function type() */
  enum {
    Type = UserType + ParFunction
    /**<
      * Specific QGraphicsItem type number of ParFunItem for function
      * type().
      */
  };
  /**
    * Constructor. Initializes corresponding to arguments.
    * @param urlX URL of the x function library.
    * @param nameX x function name.
    * @param urlY URL of the y function library.
    * @param nameY y function name.
    * @param color color for function plot.
    * @param symbol symbol type.
    * @param fx normalization factor for x values.
    * @param fy normalization factor for y values.
    * @param parent pointer to parent item.
    * @param tMin minimum argument value.
    * @param tMax maximum argument value.
    * @param deltaT argument step size.
    * @param visible true for visible item.
    * @param fillStyle fill style.
    * @param xOffset offset for x values.
    * @param yOffset offset for x values.
    * @param relSize relative size of lines or symbols.
    */
  ParFunItem(const QUrl& urlX = QUrl(), const QString& nameX = "",
             const QUrl& urlY = QUrl(), const QString& nameY = "",
             unsigned color = 0, Symbol symbol = SolidLine, double fx = 1,
             double fy = 1, QGraphicsItem* parent = 0, double tMin = 0,
             double tMax = 0, double deltaT = 0, bool visible = true,
             FillStyle fillStyle = NoBrush, double xOffset = 0,
             double yOffset = 0, double relSize = 1);
  /**
    * Copy constructor.
    * @param a source item.
    */
  ParFunItem(const ParFunItem& a);
  /**
    * Constructor. Initializes corresponding to plot file.
    * @param plo pointer to KSimpleConfig object containing item data.
    * @param urlPlo URL of plot file.
    * @param color default color for function plot.
    * @param symbol default symbol type.
    * @param parent parent item.
    */
  ParFunItem(const KConfigGroup* plo, const QUrl& urlPlo, unsigned color = 0,
             Symbol symbol = SolidLine, QGraphicsItem* parent = 0);
  /** Destructor. */
  virtual ~ParFunItem();
  /** Calculates bounding rectangle. */
  virtual QRectF boundingRect() const;
  /**
   * Calculates function value table.
   * @return number of rows.
   */
  int calcTable() const;
  /**
   * Calculates normalization factors.
   * @param expX reference to logarithm of x normalization factor.
   * @param expY reference to logarithm of x normalization factor.
   * @param fx reference to x normalization factor.
   * @param fy reference to y normalization factor.
   */
  void exponent(int& expX, int& expY, double& fx, double& fy) const;
  /**
   * Exports function value table to file.
   * @param stream stream for output file.
   * @param separator column separator character.
   * @param format C format character.
   * @param precision number of digits.
   */
  void exportTable(QTextStream& stream, char separator, char format,
                   int precision) const;
  /** y function. */
  double (*funcX)(double, const double*);
  /** Returns pointer to library for x function. */
  inline const QLibrary* libX() const { return m_libX; };
  /**
   * Calculates minimum and maximum values.
   * @param xMin reference to minimum of x values.
   * @param xMax reference to maximum of x values.
   * @param yMin reference to minimum of y values.
   * @param yMax reference to maximum of y values.
   */
  void limits(double& xMin, double& xMax, double& yMin, double& yMax) const;
  /** Returns x function name. */
  inline QString nameX() const { return m_nameX; };
  /** Paints item. */
  virtual void paint(QPainter* painter, const QStyleOptionGraphicsItem* option,
                     QWidget* widget = 0);
  /** Returns x function parameter array pointer. */
  inline double* parX() { return m_parX; };
  /** Sets offset for dashed lines. */
  void setDashOffset();
  /**
   * Resolves x function address.
   * @param url URL of the function library.
   * @param name function name.
   * @return true for success.
   */
  bool setFuncX(const QUrl& url, const QString& name);
  /**
    * Sets parameter.
    * @param i parameter index.
    * @param value value.
    * @param yFun true for y value function.
    */
  virtual void setPar(int i, double value, bool yFun);
  /** Returns specific QGraphicsItem type number. */
  virtual int type() const { return Type; };
  /** Returns URL of x function library. */
  inline const QUrl& urlX() const { return m_urlX; };
  /**
   * Writes into plot file.
   * @param plo KConfigGroup object.
   * @param urlPlo URL of plot file.
   * @param absolute true to save absolute paths.
   * @param format C format character.
   * @param precision number of digits.
   */
  void writePlo(KConfigGroup* plo, const QUrl& urlPlo, bool absolute,
                char format, int precision) const;
  /** Returns pointer to x function value array. */
  double* xValues() const;

protected:
  /** Performs initialization. */
  void init();
  /**
   * Paints or calculates bounding area.
   * @param painter QPainter object, 0 for calculating bounding area.
   */
  void paintCond(QPainter* painter) const;
  /** Unloads x function library. */
  void unloadX();
  mutable double (*funcXo)(double, const double*);
  double m_parX[NumParMax];
  mutable double parXo[NumParMax];
  mutable QVector<double> m_xValues;
  QString m_nameX;
  QUrl m_urlX;
  QLibrary* m_libX;
};

#endif
