//
// C++ Implementation: psgraph
//
// Description: 
//
//
// Author: Werner Stille <kpl@werner-stille.de>, (C) 2017
//
// Copyright: See COPYING file that comes with this distribution
//
//
#include <QtMath>
#include <QDateTime>
#include <QFile>
#include <QTextStream>
#include "arcitem.h"
#include "array3ditem.h"
#include "arrayitem.h"
#include "arrowitem.h"
#include "ellipseitem.h"
#include "frameitem.h"
#include "fun3ditem.h"
#include "imageitem.h"
#include "legenditem.h"
#include "lineitem.h"
#include "parfunitem.h"
#include "psgraph.h"
#include "rectitem.h"
#include "scalebaritem.h"
#include "splineitem.h"
#include "textitem.h"

const double PSGraph::pts = 72 / 2.54;

PSGraph::PSGraph(QTextStream* stream, const QList<QGraphicsItem*>& list,
                 const QSizeF& size, bool landscape, bool eps) :
 m_logX(false), m_logY(false), m_symbol(KplItem::OpenCircle), m_x0(0), m_y0(0),
 m_xMin(0), m_yMin(0), m_relSize(1), ts(stream), m_size(size)
{
  m_xMax = m_width = m_size.width();
  m_yMax = m_height = m_size.height();
  prolog(landscape, eps);
  drawItems(list);
  finish();
}

PSGraph::PSGraph(QTextStream* stream, const QSizeF& size,
                 const QPointF& topRight, const QPointF& bottomLeft,
                 bool landscape, bool eps) :
 m_logX(false), m_logY(false), m_symbol(KplItem::OpenCircle), m_x0(0), m_y0(0),
 m_xMin(0), m_yMin(0), m_relSize(1), ts(stream), m_size(size)
{
  m_xMax = m_width = m_size.width();
  m_yMax = m_height = m_size.height();
  prolog(landscape, eps, topRight, bottomLeft);
}

PSGraph::~PSGraph()
{
}

void PSGraph::drawArray(const double* x, const double* y, double fx, double fy,
                        int n, KplItem::FillStyle fillStyle, double xOff,
                        double yOff, bool clip)
{
  if (fillStyle) {
    bool pl = false;
    for (int i = 0; i < n; i++) {
      double xa = fx * (xOff + x[i]);
      double ya = fy * (yOff + y[i]);
      if (FrameItem::valid(xa, m_logX) && FrameItem::valid(ya, m_logY)) {
        xa = (((m_logX ? log10(xa) : xa) - m_xMin) / (m_xMax - m_xMin)
              * m_width + m_x0) * pts;
        ya = (((m_logY ? log10(ya) : ya) - m_yMin) / (m_yMax - m_yMin) *
              m_height + m_y0) * pts;
        if (pl)
          *ts << QString("%1 %2 LT\n").arg(xa).arg(ya);
        else {
          *ts << "\ngs\n";
          if (clip)
            *ts << "cl\n";
          *ts << QString("NP %1 %2 MT\n").arg(xa).arg(ya);
        }
        pl = true;
      }
    }
    if (pl)
      *ts << fillStyle << " bf\ngr\n";
  } else {
    int ie = 0;
    int ia;
    bool xNorm = (fx > 1e33) || (fx < 1e-33);
    bool yNorm = (fy > 1e33) || (fy < 1e-33);
    do {
      ia = qMin(n - 1, ie + 127);
      *ts << "\n";
      int fail = 0;
      for (int i = ia; i >= ie; i--) {
        double xa = xOff + x[i];
        if (xNorm)
          xa *= fx;
        double ya = yOff + y[i];
        if (yNorm)
          ya *= fy;
        bool b = FrameItem::valid(xa, m_logX) && FrameItem::valid(ya, m_logY);
        if (b) {
          if (m_symbol <= 0) {
            double t = xNorm ? xa : fx * xa;
            b = (fabs(((m_logX ? log10(t) : t) - m_xMin) /
                      (m_xMax - m_xMin) - 0.5) <= 1);
            if (b) {
              t = yNorm ? ya : fy * ya;
              b = (fabs(((m_logY ? log10(t) : t) - m_yMin) /
                        (m_yMax - m_yMin) - 0.5) <= 1);
            }
          } else {
            if (xa) {
              if (fabs(xa) <= 1e-37)
                xa = (xa > 0) ? 1e-37 : -1e-37;
              else
                xa = qBound(-1e37, xa, 1e37);
            }
            if (ya) {
              if (fabs(ya) <= 1e-37)
                ya = (ya > 0) ? 1e-37 : -1e-37;
              else
                ya = qBound(-1e37, ya, 1e37);
            }
          }
        }
        if (b)
          *ts << xa << " " << ya << "\n";
        else
          fail++;
      }
      int np = ia - ie - fail + 1;
      if (np)
        *ts << (xNorm ? 1 : fx) << " " << (yNorm ? 1 : fy) << " " << np
            << (clip ? " T" : " F") << " plarry\n";
      ie = ia;
      if (m_symbol <= 0)
        ie++;
    } while (ia != (n - 1));
  }
}

void PSGraph::drawItems(const QList<QGraphicsItem*>& list)
{
  foreach (QGraphicsItem* item, list)
    if (item->isVisible())
      switch (item->type()) {
        case ArcItem::Type:
          drawArc(qgraphicsitem_cast<ArcItem*>(item));
          break;
        case ArrayItem::Type:
          drawArray(qgraphicsitem_cast<ArrayItem*>(item));
          break;
        case Array3DItem::Type:
          drawArray3D(qgraphicsitem_cast<Array3DItem*>(item));
          break;
        case ArrowItem::Type:
          drawArrow(qgraphicsitem_cast<ArrowItem*>(item));
          break;
        case EllipseItem::Type:
          drawEllipse(qgraphicsitem_cast<EllipseItem*>(item));
          break;
        case FrameItem::Type:
          drawFrame(qgraphicsitem_cast<FrameItem*>(item));
          break;
        case LegendItem::Type:
          drawLegend(qgraphicsitem_cast<LegendItem*>(item));
          break;
        case LineItem::Type:
          drawLine(qgraphicsitem_cast<LineItem*>(item));
          break;
        case FunItem::Type:
          drawFunc(qgraphicsitem_cast<FunItem*>(item));
          break;
        case Fun3DItem::Type:
          drawFunc3D(qgraphicsitem_cast<Fun3DItem*>(item));
          break;
        case ImageItem::Type:
          drawImage(qgraphicsitem_cast<ImageItem*>(item));
          break;
        case ParFunItem::Type:
          drawParFunc(qgraphicsitem_cast<ParFunItem*>(item));
          break;
        case RectItem::Type:
          drawRect(qgraphicsitem_cast<RectItem*>(item));
          break;
        case ScaleBarItem::Type:
          drawScaleBar(qgraphicsitem_cast<ScaleBarItem*>(item));
          break;
        case SplineItem::Type:
          drawSpline(qgraphicsitem_cast<SplineItem*>(item));
          break;
        case TextItem::Type:
          drawText(qgraphicsitem_cast<TextItem*>(item));
          break;
      }
}

void PSGraph::drawText(const QString& text, double x, double y,
                       TextItem::Alignment alignment)
{
  psStr(text);
  *ts << " " << pts * x << " " << pts * y << " " << alignment << " text\n";
}

void PSGraph::finish()
{
  *ts << "\nshowpage gr\n\n%%Trailer\n%%EOF\n";
}

void PSGraph::setColorData(unsigned icol)
{
  *ts << QString("/cd 16#%1 d\ncd sC\n").arg(KplItem::rgbQt1(icol), 0, 16);
}

void PSGraph::setDir(double ang)
{
  *ts << "/di " << ang << " d\n";
}

double PSGraph::setRelSize(double r)
{
  double old = m_relSize;
  m_relSize = r;
  if (r != old)
    *ts << "/rs " << r << " d 0.00295 sw\n";
  return old;
}

void PSGraph::cm2r(double xcm, double ycm, double fx, double fy,
                   double* xu, double* yu) const
{
  *xu = m_xMin + (xcm - m_x0) * fx * (m_xMax - m_xMin) / m_width;
  if (m_logX)
    *xu = pow(10, *xu);
  *yu = m_yMin + (ycm - m_y0) * fy * (m_yMax - m_yMin) / m_height;
  if (m_logY)
    *yu = pow(10, *yu);
}

void PSGraph::draw3D(ThreeDItem* item)
{
  if (item->width() && item->height()) {
    *ts << m_size.width() << " " << m_size.height() << " format\n";
    setRelSize(item->relSize());
    window(item->x0(), item->width(), item->y0(), item->height());
    scale(0, item->width(), 0, item->height(), false, false);
    *ts << "/nx " << item->numDigX() << " d\n/ny " << item->numDigY()
        << " d\n";
    setColorFrame(item->colorFrame());
    setColorData(item->colorData());
    *ts << item->xTic() << " " << item->yTic() << " " << item->minTicX() << " "
        << item->minTicY() << " " << (item->showFrame() ? 0 : -1)
        << " 0 0 T F T F raster\n";
    psStr(item->header());
    *ts << " F LH\n";
    setSymbol(KplItem::SolidLine);
    item->paintCond(0, this);
    setColorData(item->colorData());
  }
}

void PSGraph::drawArc(ArcItem* item)
{
  setColorData(item->color());
  double sav = setRelSize(item->relSize());
  *ts << item->start() << " " << item->span() << " " << item->width() << " "
      << item->height() << " " << item->dir() << " " << item->x0() << " "
      << item->y0() << " " << item->symbol() << " arccm\n";
  setRelSize(sav);
}

void PSGraph::drawArray(ArrayItem* item)
{
  double** x = item->data();
  if (x && (item->count() > 0)) {
    setProperties(item);
    double sav = m_relSize;
    setRelSize(sav * item->relSize());
    drawArray(&x[item->colX()][item->rowStart()],
              &x[item->colY()][item->rowStart()], item->normX(), item->normY(),
              item->count(), item->fillStyle(), item->xOffset(),
              item->yOffset());
    if (item->errorBars())
      drawErrorBars(&x[item->colX()][item->rowStart()],
                    &x[item->colY()][item->rowStart()],
                    &x[item->colErr()][item->rowStart()], item->normX(),
                    item->normY(), item->count(), item->xOffset(),
                    item->yOffset());
    setRelSize(sav);
  }
}

void PSGraph::drawArray3D(Array3DItem* item)
{
  item->calcTable();
  draw3D(item);
}

void PSGraph::drawArrow(ArrowItem* item)
{
  setColorData(item->color());
  double sav = setRelSize(item->relSize());
  *ts << item->length() << " " << item->dir() << " " << item->x0() << " "
      << item->y0() << (item->doubleHeaded() ? " T " : " F ")
      << item->symbol() << " arrow\n";
  setRelSize(sav);
}

void PSGraph::drawEllipse(EllipseItem* item)
{
  setColorData(item->color());
  double sav = setRelSize(item->relSize());
  *ts << item->fillStyle() << " " << item->width() << " " << item->height()
      << " " << item->dir() << " " << item->x0() << " " << item->y0()
      << " " << item->symbol() << " ellcm\n";
  setRelSize(sav);
}

void PSGraph::drawErrorBars(const double* x, const double* y,
                            const double* errors, double fx, double fy, int n,
                            double xOff, double yOff, bool clip)
{
  int ie = 0;
  int ia;
  bool xNorm = (fx > 1e33) || (fx < 1e-33);
  bool yNorm = (fy > 1e33) || (fy < 1e-33);
  do {
    ia = qMin(n - 1, ie + 127);
    *ts << "\n";
    int fail = 0;
    for (int i = ia; i >= ie; i--) {
      double xa = xOff + x[i];
      if (xNorm)
        xa *= fx;
      double ya = yOff + y[i];
      if (yNorm)
        ya *= fy;
      double ea = yNorm ? (fy * errors[i]) : errors[i];
      if (FrameItem::valid(xa, m_logX) && FrameItem::valid(ya, m_logY) &&
          FrameItem::valid(ea, m_logY))
        *ts << xa << " " << ya <<  " " << ea << "\n";
      else
        fail++;
    }
    *ts << (xNorm ? 1 : fx) << " " << (yNorm ? 1 : fy) << " "
        << (ia - ie - fail + 1) << (clip ? " T" : " F") << " plerry\n";
    ie = ia + 1;
  } while (ia != (n - 1));
}

void PSGraph::drawFrame(FrameItem* item)
{
  if (item->width() && item->height()) {
    *ts << m_size.width() << " " << m_size.height() << " format\n";
    setRelSize(item->relSize());
    window(item->x0(), item->width(), item->y0(), item->height());
    scale(item->xMin(), item->xMax(), item->yMin(), item->yMax(), item->logX(),
          item->logY());
    *ts << "/nx " << item->numDigX() << " d\n/ny " << item->numDigY()
        << " d\n";
    setColorFrame(item->colorFrame());
    setColorGrid(item->colorGrid());
    *ts << item->xTic() << " " << item->yTic() << " " << item->minTicX() << " "
        << item->minTicY() << " " << item->gridMode() << " " << item->xOffset()
        << " " << item->yOffset() << (item->xBottom() ? " T" : " F")
        << (item->xTop() ? " T" : " F") << (item->yLeft() ? " T" : " F")
        << (item->yRight() ? " T" : " F") << " raster\n";
    if (item->xBottom())
      labelX(item->textX(), item->expX(), false);
    if (item->xTop())
      labelX(item->textX(), item->expX(), true);
    if (item->yLeft())
      labelY(item->textY(), item->expY(), false);
    if (item->yRight())
      labelY(item->textY(), item->expY(), true);
    psStr(item->header());
    *ts << (item->xTop() ? " T" : " F") << " LH\n";
  }
}

void PSGraph::drawFunc(FunItem* item)
{
  if (item->funcY && (item->symbol() != -17)) {
    setProperties(item);
    double sav = m_relSize;
    setRelSize(sav * item->relSize());
    int n = item->calcTable(m_logX);
    drawArray(item->tValues(), item->yValues(), item->normX(), item->normY(),
              n, item->fillStyle(), item->xOffset(), item->yOffset());
    setRelSize(sav);
  }
}

void PSGraph::drawFunc3D(Fun3DItem* item)
{
  item->calcTable();
  draw3D(item);
}

void PSGraph::drawImage(ImageItem* item)
{
  QString fn;
  double x1, x2, y1, y2;
  if (ImageItem::boundingBox(item->url(), fn, &x1, &y1, &x2, &y2)) {
    *ts << "save /showpage { } d 0 setlinecap 1 setlinewidth 0 setlinejoin\n"
        << "[ ] 0 setdash\n" << item->x0() << " sx mul " << item->y0()
        << " sx mul TR " << item->scaleX() << " " << item->scaleY() << " "
        << item->dir() << " rotate scale\n" << -x1 << " " << -y1 << " TR "
        << x1 << " " << y1 << " " << x2 << " " << y2
        << " re clip 1 setgray fill 0 setgray\n"
        << "%%BeginDocument: " << item->url().path() << "\n";
    ts->flush();
    QFile f(fn);
    if (f.open(QIODevice::ReadOnly)) {
      *ts << f.readAll();
      ts->flush();
      f.close();
    }
    if (!item->url().isLocalFile())
      QFile::remove(fn);
    *ts << "\n%%EndDocument\nrestore\n";
  }
}

void PSGraph::drawLegend(LegendItem* item)
{
  setColorData(item->color());
  setSymbol(item->symbol());
  int n;
  double xa[2], ya[2];
  if (m_symbol > 0) {
    cm2r(item->x0() - 0.4 * item->textX(), item->y0(), 1, 1, &xa[0], &ya[0]);
    cm2r(item->x0() + 0.4 * item->textX(), item->y0(), 1, 1, &xa[1], &ya[1]);
    n = 2;
    resetDashOffset();
  } else {
    cm2r(item->x0(), item->y0(), 1, 1, &xa[0], &ya[0]);
    n = 1;
  }
  double sav = m_relSize;
  setRelSize(sav * item->symbolSize());
  drawArray(xa, ya, 1, 1, n, KplItem::NoBrush, 0, 0, false);
  resetDashOffset();
  setColorFrame(item->textColor());
  setRelSize(item->relSize());
  drawText(item->text(), item->x0() + item->textX(),
           item->y0() + item->textY());
  setRelSize(sav);
}

void PSGraph::drawLine(LineItem* item)
{
  setColorData(item->color());
  setSymbol(item->symbol());
  double xa[2], ya[2];
  cm2r(item->x0(), item->y0(), 1, 1, &xa[0], &ya[0]);
  double a = qDegreesToRadians(item->dir());
  double x1 = item->x0() + item->length() * cos(a);
  double y1 = item->y0() + item->length() * sin(a);
  cm2r(x1, y1, 1, 1, &xa[1], &ya[1]);
  double sav = setRelSize(item->relSize());
  drawArray(xa, ya, 1, 1, 2, KplItem::NoBrush, 0, 0, false);
  setRelSize(sav);
}

void PSGraph::drawParFunc(ParFunItem* item)
{
  if (item->funcX && item->funcY && (item->symbol() != -17)) {
    setProperties(item);
    double sav = m_relSize;
    setRelSize(sav * item->relSize());
    int n = item->calcTable();
    drawArray(item->xValues(), item->yValues(), item->normX(), item->normY(),
              n, item->fillStyle(), item->xOffset(), item->yOffset());
    setRelSize(sav);
  }
}

void PSGraph::drawRect(RectItem* item)
{
  setColorData(item->color());
  double sav = setRelSize(item->relSize());
  *ts << item->fillStyle() << " " << item->width() << " "
      << item->height() << " " << item->dir() << " " << item->x0() << " "
      << item->y0() << " " << item->symbol() << " rectcm\n";
  setRelSize(sav);
}

void PSGraph::drawScaleBar(ScaleBarItem* item)
{
  setColorFrame(item->color());
  double sav = setRelSize(item->relSize());
  *ts << item->length() << " " << item->vertical() << " " << item->x0() << " "
      << item->y0() << " scalebar\n";
  if (item->vertical())
    drawText(item->text(), item->x0() + 0.01 * m_width,
             item->y0() + 0.5 * item->length() - 0.012 * m_relSize * m_width);
  else
    drawText(item->text(), item->x0() + 0.5 * item->length(),
             item->y0() + (2 + m_relSize) * 0.005 * m_width,
             TextItem::AlignCentered);
  setRelSize(sav);
}

void PSGraph::drawSpline(SplineItem* item)
{
  if (item->numKnots()) {
    setProperties(item);
    double sav = m_relSize;
    setRelSize(sav * item->relSize());
    int n = item->calcTable(m_logX);
    drawArray(item->xValues(), item->yValues(), item->normX(), item->normY(),
              n, item->fillStyle(), item->xOffset(), item->yOffset());
    setRelSize(sav);
  }
}

void PSGraph::drawText(TextItem* item)
{
  setColorFrame(item->color());
  setDir(item->dir());
  double sav = setRelSize(item->relSize());
  drawText(item->text(), item->x0(), item->y0(), item->alignment());
  setRelSize(sav);
  setDir(0);
}

void PSGraph::labelX(const QString& text, int exp, bool xTop)
{
  psStr(text);
  *ts << " " << exp << (xTop ? " T" : " F") << " LX\n";
}

void PSGraph::labelY(const QString& text, int exp, bool yRight)
{
  psStr(text);
  *ts << " " << exp << (yRight ? " T" : " F") << " LY\n";
}

void PSGraph::prolog(bool landscape, bool eps, const QPointF& topRight,
                     const QPointF& bottomLeft)
{
  QSizeF sp = pts * m_size;
  QSizeF strp = topRight.isNull() ?
    sp : (pts * QSizeF(topRight.x(), topRight.y()));
  QSizeF sblp = pts * QSizeF(bottomLeft.x(), bottomLeft.y());
  if (landscape) {
    sp.transpose();
    strp.transpose();
    sblp.transpose();
  }
  *ts << "%!PS-Adobe-3.0";
  if (eps)
    *ts << " EPSF-3.0";
  *ts << "\n"
    "%%BoundingBox: "
    << floor(sblp.width()) << " " << floor(sblp.height()) << " "
    << ceil(strp.width()) << " " << ceil(strp.height()) << "\n"
    "%%Creator: Kpl\n"
    "%%CreationDate: " << QDateTime::currentDateTime().toString() << "\n"
    "%%Orientation: " << (landscape ? "Landscape" : "Portrait") << "\n"
    "%%Pages: 1\n\n"
    "%User code starts at line 375\n\n"
    << "/d /def load def\n"
    "/D {bind d} bind d\n"
    "/ex {exch} D\n"
    "/ed {ex d} D\n"
    "/LT {lineto} D\n"
    "/MT {moveto} D\n"
    "/S {stroke} D\n"
    "/CP {closepath} D\n"
    "/cp {currentpoint} D\n"
    "/RL {rlineto} D\n"
    "/RM {rmoveto} D\n"
    "/NP {newpath} D\n"
    "/TR {translate} D\n"
    "/R {rotate} D\n"
    "/i {index} D\n"
    "/IE {ifelse} D\n"
    "/FF {findfont} D\n"
    "/LE {length} D\n"
    "/p {pop} D\n"
    "/sg {string} D\n"
    "/SH {show} D\n"
    "/r {repeat} D\n"
    "/L {loop} D\n"
    "/gs {gsave} D\n"
    "/gr {grestore} D\n"
    "/T {true} D\n"
    "/F {false} D\n"
    "/xmi 0 d\n"
    "/ymi 0 d\n"
    "/xma "
    << m_size.width() << " d\n"
    "/yma "
    << m_size.height() << " d\n"
    "/sx 28.346456 d\n"
    "/xs sx d\n"
    "/ys xs d\n"
    "/mix 0 d\n"
    "/miy 0 d\n"
    "/max xs xma mul d\n"
    "/may ys yma mul d\n"
    "/cf 0 d\n"
    "/cg 0 d\n"
    "/cd 0 d\n"
    "/di 0 d\n"
    "/sy 0 d\n"
    "/dp 0 d\n"
    "/dc 0 d\n"
    "/rs 1 d\n"
    "/nx -1 d\n"
    "/ny -1 d\n"
    "/lx F d\n"
    "/ly F d\n\n"
    << "/EF {\n"
    " FF dup LE dict begin {\n"
    "  1 i dup /FID eq ex /Encoding eq or {p p} {d} IE\n"
    " } forall /Encoding ISOLatin1Encoding d currentdict end definefont\n"
    "} D\n\n"
    << "/sc {\n"
    " /s2 ed /s1 ed s1 LE dup s2 LE add sg /s3 ed s3 ex\n"
    " s2 putinterval s3 0 s1 putinterval s3\n"
    "} D\n\n"
    << "/rr {ymi sub ys mul miy add ex xmi sub xs mul mix add ex} D\n\n"
    << "/mv {NP rr MT} D\n\n"
    << "/re {/y2 ed /x2 ed /y1 ed /x1 ed"
    " NP x1 y1 MT x2 y1 LT x2 y2 LT x1 y2 LT CP} D\n\n"
    << "/SC {dup stringwidth ex 2 div neg ex RM cp"
    " 3 2 roll SH MT} D\n\n"
    << "/SR {dup stringwidth ex neg ex RM cp 3 2 roll SH MT} D\n\n"
    << "/rl {max mix sub mul rs mul} D\n\n"
    << "/sw {rl setlinewidth} D\n\n"
    << "/sf {rl scalefont setfont} D\n\n"
    << "/mk {/sh 0.01 rl d sh mul ex sh mul ex RM {"
    "sh mul ex sh mul ex RL} r} D\n\n"
    << "/m0 {NP rr 0.01 rl 0 360 arc} D\n\n"
    "/m1 {NP rr 0.002 rl 0 360 arc} D\n\n"
    "/m2 {mv -1 1 1 1 1 -1 -1 -1 4 0 1 mk} D\n\n"
    "/m3 {mv 0 2 2 0 0 -2 -2 0 4 1 1 mk} D\n\n"
    "/m4 {mv -1 1.732 2 0 -1 -1.732 3 0 0.866 mk} D\n\n"
    "/m5 {mv 2 2 -2 0 2 -2 -2 0 4 1 1 mk} D\n\n"
    "/m6 {mv 0 1 0.5 0.5 1 0 0.5 -0.5 0 -1 -0.5 -0.5 -1 0 -0.5"
    " 0.5 8 1 0.5 mk} D\n\n"
    "/m7 {mv 2 -2 -1 1 1 1 -2 -2 4 1 1 mk} D\n\n"
    "/m8 {mv 2 0 -1 0 0 1 0 -2 4 0 1 mk} D\n\n"
    "/m9 {mv -1 1 -1 -1 1 1 0 0.866 4 0 -0.866 mk} D\n\n"
    "/m10 {mv 1 1 -2 0 1 -1 0 2 4 0 -1 mk} D\n\n"
    << "/Lt {\n 1 sub dup 0 lt {p 0} if 9 mod dup\n"
    " [8 7 2 1 9 8 7 4 1] ex get /p0 ed"
    " [8 0 0 0 0 1 2 4 1] ex get /p1 ed\n"
    " /f 0.003125 rl d 4 array dup 0 p0 f mul put dup 1 16 p0 sub p1 sub\n"
    " f mul p1 0 eq {put dup dup 0 get 2 ex} {0.5 mul put dup 2 p1 f mul}\n"
    " IE put dup dup 1 get 3 ex put dp setdash\n"
    "} D\n\n"
    << "/er {\n fy mul ys mul /dy ed fy mul /y ed fx mul /x ed ly\n"
    " {y 0 gt {/dy dy y div 0.4343 mul d} {/dy 0 d} IE} if\n"
    " x 0 gt lx not or dy sh ge and {\n"
    "  x lx {log} if y ly {log} if mv 0 sh RM 0 dy sh sub\n"
    "  RL 0.5 sh mul 0 RM sh neg 0 RL 0 -2 dy mul RM sh\n"
    "  0 RL -0.5 sh mul 0 RM 0 dy sh sub RL S\n"
    " } if\n"
    "} D\n\n"
    << "/format {\n /yma ed /xma ed /xmi 0 d /ymi 0 d"
    " /xs sx d /ys xs d /mix 0 d /miy 0 d /max xs\n"
    " xma mul d /may ys yma mul d\n"
    "} D\n\n"
    << "/window {\n /yma ed /ymi ed /xma ed /xmi ed"
    " /xma xma xmi add d /yma yma ymi add d\n"
    " /mix xs xmi mul d /miy ys ymi mul d"
    " /max xs xma mul d /may ys yma mul d\n"
    " 0.00295 sw"
    " /l1 FF 0.033333 sf\n"
    "} D\n\n"
    << "/Sc {\n /ly ed /lx ed /yma ed /ymi ed /xma ed /xmi ed\n"
    " lx {/xmi xmi log d /xma xma log d} if"
    " ly {/ymi ymi log d /yma yma log d} if\n"
    " /xs max mix sub xma xmi sub div d"
    " /ys may miy sub yma ymi sub div d\n"
    "} D\n\n"
    << "/sC {\n"
    " /col ed col 16#ff and 16#ff div col -8 bitshift 16#ff and 16#ff div\n"
    " col -16 bitshift 16#ff div setrgbcolor\n"
    "} D\n\n"
    << "/frame {mix miy max may re S} D\n\n"
    "/cl {mix dc add miy dc add max dc sub may dc sub re clip} D\n\n"
    << "/pn {\n"
    " /nD ed /v ed v 0 lt {/v v neg d (-)} {()} IE /v nD 0 gt\n"
    " {10 nD exp dup v mul round dup cvi /iv ed ex div}\n"
    " {v round dup cvi /iv ed} IE d /ie v 10 lt {0} {v log cvi}\n"
    " IE d /iend nD 0 gt {nD neg} {0} IE d"
    " /id 10 ie iend sub exp round cvi d {\n"
    "  iv dup id mod 10 mul /iv ed id idiv 256 sg cvs\n"
    "  sc ie 0 eq {nD 0 ge {(.) sc} if} if ie iend eq"
    " {exit} if /ie ie 1 sub d\n"
    " } L\n"
    "} D\n\n"
    << "/il {\n"
    " log /lf ed /xl ed /ma xl dup dup abs 1.0e-5 mul add cvi sub d\n"
    " ma abs 1.0e-6 lt {/ma 0 d} if ma 0 lt {/ma ma 1 add d}\n"
    " if xl ma 0.1 lt {lf add} {ma 0.4 lt {2.5 log add}\n"
    "  {ma 0.6 lt {10 3 div log add} {2 log add} IE} IE} IE\n"
    "} D\n\n"
    << "/raster {\n"
    " /yR ed /yL ed /xT ed /xB ed /yO ed /xO ed /mo ed\n"
    " /iy ed /ix ed /yt ed /xt ed mo -1 ne {mo 0 ne {\n"
    "  lx {xt 2.5 lt {/xt 2 d} {xt 5.5 lt {/xt 3 d}\n"
    "   {xt 99 lt {/xt 10 d} if} IE} IE} if\n"
    "  ly {yt 2.5 lt {/yt 2 d} {yt 5.5 lt {/yt 3 d}\n"
    "   {yt 99 lt {/yt 10 d} if} IE} IE} if\n"
    "  cg sC 0.00208 sw /dx max mix sub d /dy may miy sub d\n"
    "  mo 2 le {/dx dx 0.02 mul d /dy dx d} if\n"
    "  mo 2 le ix 1 gt and xO 0 gt and lx not and {\n"
    "   /dd xt ix div d /xm xmi xO add dd sub d {xm xmi le {exit} if xm xma lt {\n"
    "    xB {xm ymi mv 0 dy 0.5 mul RL S} if\n"
    "    xT {xm yma mv 0 dy -0.5 mul RL S} if} if\n"
    "    /xm xm dd sub d} L} if\n"
    "  /x xmi xO add d {lx\n"
    "   {/xa x d /x x xt il d\n"
    "    x xma lt {xB {x ymi mv 0 dy RL S} if\n"
    "    xT {x yma mv 0 dy neg RL S} if} if ix 1 gt {\n"
    "     /xd 10 xa ma sub exp d {/xa 10 xa exp xd add log d\n"
    "      xa x ge xa xma ge or {exit} if\n"
    "      xB {xa ymi mv 0 dy mo 2 le {0.5 mul} if RL S} if\n"
    "      xT {xa yma mv 0 dy neg mo 2 le {0.5 mul} if RL S} if\n"
    "     } L\n"
    "    } if} {mo 2 le ix 1 gt and {\n"
    "     /xm x d ix 1 sub {\n"
    "      /xm xt ix div xm add d xm xma ge {exit} if\n"
    "      xB {xm ymi mv 0 dy 0.5 mul RL S} if"
    " xT {xm yma mv 0 dy -0.5 mul RL S} if\n"
    "     } r\n"
    "    } if x xmi gt x xma lt and {\n"
    "     xB {x ymi mv 0 dy RL S} if\n"
    "     xT {x yma mv 0 dy neg RL S} if\n"
    "    } if /x x xt add d\n"
    "   } IE x xma ge {exit} if} L\n"
    "  mo 2 le iy 1 gt and yO 0 gt and ly not and {\n"
    "   /dd yt iy div d /ym ymi yO add dd sub d {\n"
    "    ym ymi le {exit} if ym yma lt {"
    "     yL {xmi ym mv dx 0.5 mul 0 RL S} if\n"
    "     yR {xma ym mv dx -0.5 mul 0 RL S} if} if\n"
    "    /ym ym dd sub d} L} if\n"
    "   /y ymi yO add d {ly {/ya y d /y y yt il d\n"
    "    y yma lt {yL {xmi y mv dx 0 RL S} if\n"
    "    yR {xma y mv dx neg 0 RL S} if} if iy 1 gt {\n"
    "     /yd 10 ya ma sub exp d {/ya 10 ya exp yd add log d\n"
    "      ya y ge ya yma ge or {exit} if\n"
    "      yL {xmi ya mv dx mo 2 le {0.5 mul} if 0 RL S} if\n"
    "      yR {xma ya mv dx neg mo 2 le {0.5 mul} if 0 RL S} if\n"
    "     } L\n"
    "    } if} {mo 2 le iy 1 gt and {\n"
    "     /ym y d iy 1 sub {\n"
    "      /ym yt iy div ym add d ym yma ge {exit} if\n"
    "      yL {xmi ym mv dx 0.5 mul 0 RL S} if"
    " yR {xma ym mv dx -0.5 mul 0 RL S} if\n"
    "     } r} if y ymi gt y yma lt and {\n"
    "     yL {xmi y mv dx 0 RL S} if"
    " yR {xma y mv dx neg 0 RL S} if\n"
    "    } if /y y yt add d\n"
    "   } IE y yma ge {exit} if} L mo 2 eq mo 4 eq or {\n"
    "   xB {/yB ymi xma xmi sub rs mul xs mul ys div 0.05 mul sub d} if\n"
    "   xT {/yT yma xma xmi sub rs mul xs mul ys div 0.025 mul neg sub d} if\n"
    "   /e xma 2.99 gt xmi -2.99 lt or xt 9.9 gt and\n"
    "   lx and d /nd nx d /x xmi xO add d /xe xma xmi sub\n"
    "   1e-3 mul xma add d /tol xt 1e-6 mul d {x xe gt {exit} if\n"
    "    x xma gt {/x xma d} if e {(10#nu) x round cvi 4 sg cvs sc\n"
    "     xB {dup x yB rr 2 text} if xT {x yT rr 2 text} {p} IE}\n"
    "     {lx {x 0 lt {/nd 1 x 0.01 add cvi sub d} {/nd -1 d} IE}\n"
    "     if x lx {10 ex exp} {x abs tol lt {p 0} if} IE nd pn cf sC\n"
    "     xB {dup x yB mv SC} if xT {x yT mv SC} {p} IE}\n"
    "    IE lx {/x x xt il d} {/x x xt add d} IE\n"
    "   } L yL {/xL xmi xma xmi sub rs mul 0.025 mul sub d} if\n"
    "   yR {/xR xma xma xmi sub rs mul -0.025 mul sub d} if\n"
    "   /dy xma xmi sub 0.0125 mul rs mul xs mul ys div d\n"
    "   /e yma 2.99 gt ymi -2.99 lt or yt 9.9 gt and ly and d\n"
    "   /nd ny d /y ymi yO add d /ye yma ymi sub 1e-3 mul\n"
    "   yma add d /tol yt 1e-6 mul d {y ye gt {exit} if\n"
    "    y yma gt {/y yma d} if\n"
    "    e {(10#nu) y round cvi 4 sg cvs sc\n"
    "     yL {dup xL y dy sub rr 3 text} if"
    " yR {xR y dy sub rr 1 text} {p} IE}\n"
    "     {ly {y 0 lt {/nd 1 y 0.01 add cvi sub d} {/nd -1 d} IE}\n"
    "     if y ly {10 ex exp} {y abs tol lt {p 0} if} IE nd pn cf sC\n"
    "     yL {dup xL y dy sub mv SR} if yR {xR y dy sub mv SH} {p} IE\n"
    "    } IE ly {/y y yt il d} {/y y yt add d} IE\n"
    "   } L\n"
    "  } if\n"
    " } if cf sC 0.00417 sw /dc currentlinewidth 0.5 mul d"
    " frame cd sC 0.00295 sw} if\n"
    "} D\n\n"
    << "/text {\n"
    " /io ed gs TR di R 0 0 MT /str ed"
    " /l1 FF 0.033333 sf cf sC str (#) search {\n"
    "  p p p /dx 0 d io 1 eq {1} {0} IE 1 1 {\n"
    "   /j ed /ilev 0 d /s str d"
    " j 1 eq {dx io 2 eq {0.5 mul} if neg 0 RM} if {\n"
    "    s (#) search ex /s1 ed {p /c ed} {/c 0 sg d} IE\n"
    "    j 0 eq {s1 stringwidth p dx add /dx ed} {\n"
    "     ilev 0 eq {s1 SH} {\n"
    "      ilev 1 eq {0.014} {-0.0045} IE rl dup 0 ex RM s1 SH neg 0 ex RM\n"
    "     } IE } IE c LE 2 lt {exit} {\n"
    "     c 0 get dup 96 gt {32 sub} if 1 sg dup 0 4 -1 roll put\n"
    "     (NHBOSZI) ex search {\n"
    "      LE [/l1 /l1 /l1B /l1O /Symbol /Dingbats /l1I]\n"
    "      ex get FF c 1 get dup 96 gt {32 sub} if 1 sg dup\n"
    "      0 4 -1 roll put (NUD) ex search {LE /ilev ed p} if\n"
    "      p ilev 0 eq {0.033333} {0.022222} IE sf p\n"
    "     } if p /s c 2 c LE 2 sub getinterval d\n"
    "    } IE s LE 0 eq {exit} if\n"
    "   } L /l1 FF 0.033333 sf\n"
    "  } for\n"
    " } {io 1 eq {SH} {io 2 eq {SC} {SR} IE} IE} IE gr\n"
    "} D\n\n"
    << "/arrow {\n"
    " /s ed /b ed gs sx mul ex sx mul ex TR NP 0 0 MT R -0.03333 rl dup dup\n"
    " dup 0.2 mul dup neg 3 1 roll LT LT CP fill ex sx mul b {\n"
    "  dup neg 0 TR 180 R NP 0 0 MT ex dup dup 0.2 mul dup neg 3 1 roll\n"
    "  LT LT CP fill -180 R dup 0 TR\n"
    " } {ex p} IE dup 0.03333 b {dup add} if rl ge {\n"
    "  0.00417 sw b {0.03333 rl sub} if neg 0 MT -0.03333 rl 0 LT s Lt S\n"
    " } {p} IE gr\n"
    "} D\n\n"
    << "/arccm {\n"
    " /s ed gs sx mul ex sx mul ex TR R /h ed h div\n"
    " /a ed /wl ed /wa ed 6 array currentmatrix a 1 scale NP 0 0 h\n 0.5 mul"
    " sx mul wa wl wl 0 ge {wa add} {add wa} IE arc setmatrix s Lt S gr\n"
    "} D\n\n"
    << "/rectcm {\n"
    " /s ed gs sx mul ex sx mul ex TR R sx mul ex\n"
    " sx mul ex 0 0 re dup 0 gt {bf} {p s Lt S} IE gr\n"
    "} D\n\n"
    << "/ellcm {\n"
    " /s ed gs sx mul ex sx mul ex TR R /h ed h div\n"
    " /a ed 6 array currentmatrix a 1 scale NP 0 0 h 0.5 mul sx\n"
    " mul 0 360 arc setmatrix dup 0 gt {bf} {p s Lt S} IE gr\n"
    "} D\n\n"
    << "/scalebar {\n"
    " gs cf sC sx mul ex sx mul ex TR 0 ne {90 R} if 0.00417 sw\n"
    " /a max mix sub rs mul 0.005 mul d NP 0 a MT 0 a neg LT S NP 0 0 MT\n"
    " sx mul 0 LT cp S NP MT 0 a RM 0 a -2 mul RL S gr\n"
    "} D\n\n"
    << "/LH {{0.16975} {0.025} IE rl may add mix max add 0.5 mul ex"
    " 2 text} D\n\n"
    << "/XY {\n"
    " /iax ed /op ed /iex ed /str ed str LE iex or 0 ne {\n"
    "  iex 0 eq {str} {(10#nu) iex 4 sg cvs sc ( #nn) sc str\n"
    "   sc} IE iax 2 eq {/l 3 ly {yma 2.99 gt ymi -2.99 lt\n"
    "   or {ymi 0 lt {4} {3} IE} {yma 1.99 ymi sub 2 copy lt\n"
    "   {ex} if p cvi 1 add} IE} {ymi ny pn LE\n"
    "   yma ny pn LE 2 copy lt {ex} if p} IE add d\n"
    "   op {max} {mix} IE l 0.01875 rl mul op\n"
    "   {neg} if sub miy may add 0.5 mul /di 90 op {neg} if d\n"
    "  } {mix max add 0.5 mul op {may -0.085} {miy 0.11} IE\n"
    "  rl sub} IE 2 text /di 0 d\n"
    " } if\n"
    "} D\n\n"
    << "/LX {1 XY} D\n\n"
    << "/LY {2 XY} D\n\n"
    << "/plarry {\n"
    " gs {cl} if 3 1 roll /fy ed /fx ed sy 0 gt {\n"
    "  sy Lt 3 1 roll fy mul /ya ed fx mul /xa ed /p1 F d\n"
    "  xa 0 gt lx not or ya 0 gt ly not or and {\n"
    "   xa lx {log} if ya ly {log} if NP rr /y ed /x ed x y MT /p1 T d\n"
    "  } if 1 sub {\n"
    "   fy mul /ya ed fx mul /xa ed"
    " xa 0 gt lx not or ya 0 gt ly not or and {\n"
    "    xa lx {log} if ya ly {log} if rr dup y sub dup mul\n"
    "    2 i x sub dup mul add sqrt dp add /dp ed dup /y ed\n"
    "    /x 2 i d p1 {LT} {MT /p1 T d} IE\n"
    "   } if\n"
    "  } r S /dp dp dup 0.05 rl div floor 0.05 rl mul sub d [] dp setdash\n"
    " } {\n"
    "  /m [/m0 /m1 /m2 /m3 /m4 /m5 /m6 /m7 /m8 /m9 /m10 /m0 /m2 /m3 /m4\n"
    "  /m5 /m6] sy neg 17 mod get load d {\n"
    "   sy -17 eq {p p} {fy mul /ya ed fx mul /xa ed xa 0 gt\n"
    "   lx not or ya 0 gt ly not or and {xa lx {log} if ya ly\n"
    "   {log} if m sy neg 17 mod dup 1 eq ex 10 gt or {fill} {S} IE} if}"
    " IE\n  } r\n"
    " } IE gr\n"
    "} D\n\n"
    << "/plerry {gs {cl} if 3 1 roll /fy ed /fx ed /sh 0.01 rl d"
    " {er} r gr} D\n\n"
    "/bf {\n"
    " /b ed b 1 eq {fill} {clip pathbbox 2 i sub /h ed\n"
    " 2 i sub /w ed TR 0.2 setlinewidth NP\n"
    " b 2 eq b 4 eq or {0 2.834 h {dup 0 ex MT w ex LT} for} if\n"
    " b 3 eq b 4 eq or {0 2.834 w {dup 0 MT h LT} for} if\n"
    " b 5 eq b 7 eq or {w h gt {0 4 w h add "
    "{dup h MT h sub 0 LT} for}\n"
    " {0 4 w h add {dup w ex MT w sub 0 ex LT} for} IE} if\n"
    " b 6 eq b 7 eq or {w h gt {0 4 w h add "
    "{dup 0 MT h sub h LT} for}\n"
    " {0 4 w h add {dup 0 ex MT w sub w ex LT} for} IE} if S} IE\n"
    "} D\n\n"
    << "%%Page: 1 1\n\n"
    "gs\n"
    "/l1 /Helvetica EF 0.033333 sf\n"
    "/l1B /Helvetica-Bold EF p\n"
    "/l1O /Helvetica-Oblique EF p\n"
    "/l1I /Helvetica-BoldOblique EF p\n"
    "1 setlinecap 1 setlinejoin\n\n"
    "0.00295 sw\n\n%User code starts here\n\n";
  if (landscape)
    *ts << qRound(sp.width()) << " 0 TR 90 R\n";
}

void PSGraph::psStr(const QString& text)
{
  *ts << "(";
  for (int i = 0; i < text.size(); ++i) {
    QChar c = text[i];
    unsigned char uc = (unsigned char) c.toLatin1();
    if (uc > 127)
      *ts << "\\" << QString::number(uc, 8);
    else
      switch (uc) {
        case 0:
          *ts << (c.isPrint() ? "\\040" : "\\000");
          break;
        case 40:
        case 41:
        case 92:
          *ts << "\\";
        default:
          *ts << c;
      }
  }
  *ts << ")";
}

void PSGraph::resetDashOffset()
{
  *ts << "/dp 0 d\n";
}

void PSGraph::scale(double xMin, double xMax, double yMin, double yMax,
                    bool logX, bool logY)
{
  m_logX = logX;
  m_logY = logY;
  if (m_logX) {
    m_xMin = log10(xMin);
    m_xMax = log10(xMax);
  } else {
    m_xMin = xMin;
    m_xMax = xMax;
  }
  if (m_logY) {
    m_yMin = log10(yMin);
    m_yMax = log10(yMax);
  } else {
    m_yMin = yMin;
    m_yMax = yMax;
  }
  *ts << xMin << " " << xMax << " " << yMin << " " << yMax
      << (m_logX ? " T" : " F") << (m_logY ? " T" : " F") << " Sc\n";
}

void PSGraph::setColorFrame(unsigned icol)
{
  *ts << QString("/cf 16#%1 d\n").arg(KplItem::rgbQt1(icol), 0, 16);
}

void PSGraph::setColorGrid(unsigned icol)
{
  *ts << QString("/cg 16#%1 d\n").arg(KplItem::rgbQt1(icol), 0, 16);
}

void PSGraph::setProperties(ScaledItem* item)
{
  setColorData(item->color());
  setSymbol(item->symbol());
}

void PSGraph::setSymbol(KplItem::Symbol is)
{
  m_symbol = is;
  *ts <<"/sy " << is << " d\n";
}

void PSGraph::window(double x0, double width, double y0, double height)
{
  m_x0 = x0;
  m_xMax = m_width = width;
  m_y0 = y0;
  m_yMax = m_height = height;
  m_xMin = m_yMin = 0;
  *ts << m_x0 << " " << m_width << " " << m_y0 << " " << m_height
      << " window\n";
}
