/***************************************************************************
                          rectitem.h  -  description
                             -------------------
    begin                : Sat Apr 28 2001
    copyright            : (C) 2017 by Werner Stille
    email                : kpl@werner-stille.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef RECTITEM_H
#define RECTITEM_H

#include <QGraphicsRectItem>
#include "decoitem.h"

/**
  * Rectangle item class. Class for rectangle decoration items.
  * @author Werner Stille
  */
class Q_DECL_EXPORT RectItem : public DecoItem, public QGraphicsRectItem
{
public:
  /** Specific QGraphicsItem type number for function type() */
  enum {Type = UserType + Rectangle
    /**< Specific QGraphicsItem type number of RectItem for function type(). */
  };
  /**
    * Constructor. Initializes corresponding to arguments.
    * @param x0 x position in cm.
    * @param y0 y position in cm.
    * @param color color for rectangle.
    * @param parent pointer to parent item.
    * @param width width in cm.
    * @param height height in cm.
    * @param visible true for visible item.
    * @param fillStyle fill style.
    * @param symbol line type.
    * @param relSize relative line width.
    * @param dir rotation angle in degree.
    */
  RectItem(double x0 = 0, double y0 = 0, unsigned color = 0,
           QGraphicsItem* parent = 0, double width = 3, double height = 2,
           bool visible = true, FillStyle fillStyle = NoBrush,
           Symbol symbol = SolidLine, double relSize = 1, double dir = 0);
  /**
    * Copy constructor.
    * @param a source item.
    */
  RectItem(const RectItem& a);
  /**
    * Constructor. Initializes corresponding to plot file.
    * @param plo pointer to KConfigGroup object containing item data.
    * @param color default color for rectangle.
    * @param parent pointer to parent item.
    */
 RectItem(const KConfigGroup* plo, unsigned color = 0,
          QGraphicsItem* parent = 0);
  /** Destructor. */
  virtual ~RectItem();
  /** Returns rotation angle in deg. */
  inline double dir() const { return m_dir; };
  /** Returns fill style. */
  inline FillStyle fillStyle() const { return m_fillStyle; };
  /** Returns height in cm. */
  inline double height() const { return m_height; };
  /** Returns decoration pixmap for item list. */
  QPixmap pixmap() const;
  /**
    * Sets rotation angle.
    * @param dir rotation angle in deg.
    */
  inline void setDir(double dir) { m_dir = dir; };
  /**
    * Sets fill style.
    * @param style fill style.
    */
  inline void setFillStyle(FillStyle style) { m_fillStyle = style; };
  /**
    * Sets height.
    * @param height height in cm.
    */
  inline void setHeight( double height) { m_height = height; };
  /** Sets pen width. */
  void setPenWidth();
  /**
    * Sets properties according to transformation matrix.
    * @param m transformation matrix.
    */
  void setProperties(const QTransform& m);
  /**
    * Sets line type.
    * @param symbol line type.
    */
  inline void setSymbol(Symbol symbol) { m_symbol = symbol; };
  /**
    * Sets width.
    * @param width width in cm.
    */
  inline void setWidth( double width) { m_width = width; };
  /** Returns line type. */
  inline Symbol symbol() const { return m_symbol; };
  /** Returns specific QGraphicsItem type number. */
  virtual int type() const { return Type; };
  /** Applies changes and forces redrawing. */
  void update();
  /** Returns width in cm. */
  inline double width() const { return m_width; };
  /**
   * Writes into plot file.
   * @param plo KConfigGroup object.
   */
  void writePlo(KConfigGroup* plo) const;

protected:
  /** Updates position. */
  virtual void mouseReleaseEvent(QGraphicsSceneMouseEvent* event);
  /** Sets properties according to transformation matrix. */
  void setProperties();
  FillStyle m_fillStyle;
  double m_width, m_height, m_dir;
  Symbol m_symbol;
};

#endif
