//
// C++ Implementation: scaleditem
//
// Description: 
//
//
// Author: Werner Stille <kpl@werner-stille.de>, (C) 2017
//
// Copyright: See COPYING file that comes with this distribution
//
//
#include <QLineF>
#include <QPainter>
#include <QPolygonF>
#include <KConfigGroup>
#include "frameitem.h"
#include "kplscene.h"
#include "scaleditem.h"

ScaledItem::ScaledItem() :
 m_fillStyle(NoBrush), m_symbol(OpenCircle), m_color(0), m_normX(1),
 m_normY(1), m_xOffset(0), m_yOffset(0), myDashOffset(0)
{
  setFlags(QGraphicsItem::ItemIsSelectable);
}

ScaledItem::ScaledItem(const ScaledItem& a) :
 KplItem(a), QGraphicsItem(), m_fillStyle(a.m_fillStyle), m_symbol(a.m_symbol),
 m_color(a.m_color), m_normX(a.m_normX), m_normY(a.m_normY),
 m_xOffset(a.m_xOffset), m_yOffset(a.m_yOffset), myDashOffset(a.myDashOffset),
 m_matrix(a.m_matrix)
{
  setZValue(a.zValue());
  setFlags(QGraphicsItem::ItemIsSelectable);
  setSelected(a.isSelected());
  setVisible(a.isVisible());
}

ScaledItem::ScaledItem(const KConfigGroup* plo, Symbol symbol, unsigned col,
                       QGraphicsItem* parent) :
QGraphicsItem(parent), myDashOffset(0)
{
  readPlo(plo, symbol, col);
}

ScaledItem::ScaledItem(FillStyle fillStyle, Symbol symbol, unsigned col,
                       double fx, double fy, double xOffset, double yOffset,
                       double relSize, QGraphicsItem* parent) :
 KplItem(relSize), QGraphicsItem(parent), m_fillStyle(fillStyle),
 m_symbol(symbol), m_color(col), m_normX(fx), m_normY(fy), m_xOffset(xOffset),
 m_yOffset(yOffset), myDashOffset(0)
{
  setFlags(QGraphicsItem::ItemIsSelectable);
}

ScaledItem::~ScaledItem()
{
}

void ScaledItem::exponent(double a, int& iExp, double& f)
{
  if ((iExp = -int(floor(log10(1.01 * a + 1e-301)))) > 0)
    iExp++;
  iExp = 3 * (iExp / 3);
  f = pow(10, iExp);
}

bool ScaledItem::limits(double& xMin, double& xMax, const double* values,
                        int n)
{
  xMin = 1e300;
  xMax = -xMin;
  for (int i = 0; i < n; i++)
    if (FrameItem::valid(values[i])) {
      xMin = qMin(xMin, values[i]);
      xMax = qMax(xMax, values[i]);
    }
  return (xMax > xMin);
}

QPixmap ScaledItem::pixmap() const
{
  return KplItem::pixmap(m_pen, m_fillStyle ? FilledSquare : m_symbol);
}

void ScaledItem::readPlo(const KConfigGroup* plo, Symbol symbol,
                         unsigned color)
{
  KplItem::readPlo(plo);
  m_fillStyle = FillStyle(plo->readEntry("fillstyle", 0));
  m_symbol = Symbol(plo->readEntry("symb", int(symbol)));
  QString s = plo->readEntry("color", "");
  m_color = s.isEmpty() ? color : rgb(s);
  m_normX = plo->readEntry("fx", 1.0);
  m_normY = plo->readEntry("fy", 1.0);
  m_xOffset = plo->readEntry("x", 0.0);
  m_yOffset = plo->readEntry("y", 0.0);
  if (!plo->readEntry("active", true))
    hide();
  setFlags(QGraphicsItem::ItemIsSelectable);
}

void ScaledItem::setProperties(const QTransform& m)
{
  m_matrix = m;
  setProperties();
}

void ScaledItem::setSize()
{
  prepareGeometryChange();
  QGraphicsItem* parent = parentItem();
  double r = (parent && parent->isVisible() &&
              parent->type() == FrameItem::Type) ?
              qgraphicsitem_cast<FrameItem*>(parent)->relSize() : 1;
  m_pen.setWidthF(r * penWidth(this));
  double dx = 0.01 * r * m_relSize * parentWidth(this);
  double dy = dydx * dx;
  symbRect.setCoords(-dx, dy, dx, -dy);
  bPath = QPainterPath();
  bRect = QRectF();
}

void ScaledItem::update()
{
  setProperties();
  QGraphicsItem::update();
}

void ScaledItem::updateGeometry()
{
  prepareGeometryChange();
}

void ScaledItem::writePlo(KConfigGroup* plo, bool visible) const
{
  KplItem::writePlo(plo, visible);
  plo->writeEntry("fillstyle", int(m_fillStyle));
  plo->writeEntry("symb", int(m_symbol));
  plo->writeEntry("color", QString::number(rgbQt1(m_color), 16));
  plo->writeEntry("fx", m_normX);
  plo->writeEntry("fy", m_normY);
  plo->writeEntry("x", m_xOffset);
  plo->writeEntry("y", m_yOffset);
}

double ScaledItem::length(const double* x, const double* y, int n) const
{
  double len = 0;
  QPointF p1, p2;
  const FrameItem* frame = qgraphicsitem_cast<FrameItem*>(parentItem());
  const KplScene* sc = (frame && frame->isVisible()) ?
                        0 : static_cast<KplScene*>(scene());
  bool first = true;
  for (int i = 0; i < n; ++i) {
    double xa = m_normX * (m_xOffset + x[i]);
    double ya = m_normY * (m_yOffset + y[i]);
    if (sc ? (FrameItem::valid(xa) && FrameItem::valid(ya)) :
             frame->valid(xa, ya)) {
      QPointF pu(xa, ya);
      p2 = sc ? sc->map(pu) : frame->mapFromUser(pu);
      if (first)
        first = false;
      else
        len += QLineF(p1, p2).length();
      p1 = p2;
    }
  }
  return len;
}

void ScaledItem::plotArray(QPainter* painter, const double* x, const double* y,
                           int n) const
{
  const FrameItem* frame = qgraphicsitem_cast<FrameItem*>(parentItem());
  const KplScene* sc = (frame && frame->isVisible()) ?
                        0 : static_cast<KplScene*>(scene());
  if (!(frame || sc))
    return;
  QPolygonF ap;
  if (m_fillStyle || (m_symbol > OpenCircle))
    for (int i = 0; i < n; ++i) {
      double xa = m_normX * (m_xOffset + x[i]);
      double ya = m_normY * (m_yOffset + y[i]);
      if (sc ? (FrameItem::valid(xa) && FrameItem::valid(ya)) :
               frame->valid(xa, ya)) {
        QPointF pu(xa, ya);
        ap << (sc ? sc->map(pu) : frame->mapFromUser(pu));
      }
    }
  if (painter && (n > 30001))
// TODO: drawPolyline performance problem
    painter->setRenderHint(QPainter::Antialiasing, false);
  if (m_fillStyle) {
    if (painter)
      painter->drawPolygon(ap);
    else {
      bPath.addPolygon(ap);
      bRect = ap.boundingRect();
    }
  } else {
    if (m_symbol > OpenCircle) {
      if (painter)
        painter->drawPolyline(ap);
      else {
        QPainterPath pp;
        pp.addPolygon(ap);
        QPainterPathStroker pps;
        pps.setWidth(6);
        bPath = pps.createStroke(pp);
        double d = 0.5 * m_pen.widthF();
        bRect = ap.boundingRect().adjusted(-d, -d, d, d);
      }
    } else {
      QRectF r = symbRect;
      int k = (-m_symbol) % Invisible;
      if ((k > -Umbrella) || (k == -Dot)) {
        if (painter) {
          painter->setBrush(m_pen.color());
          painter->setPen(Qt::NoPen);
        }
      } else
        if (!painter) {
          double d = 0.5 * m_pen.widthF();
          r.adjust(-d, -d, d, d);
        }
      for (int i = 0; i < n; ++i) {
        double xa = m_normX * (m_xOffset + x[i]);
        double ya = m_normY * (m_yOffset + y[i]);
        if (sc ? (FrameItem::valid(xa) && FrameItem::valid(ya)) :
                 frame->valid(xa, ya)) {
          QPointF pu(xa, ya);
          QPointF p = sc ? sc->map(pu) : frame->mapFromUser(pu);
          if (painter) {
            painter->save();
            painter->translate(p);
            paintSymbol(painter, symbRect, m_symbol);
            painter->restore();
          } else {
            QRectF r2 = r.translated(p);
            bPath.addRect(r2);
            if (bRect.isNull())
              bRect = r2;
            else
              bRect |= r2;
          }
        }
      }
    }
  }
}

void ScaledItem::setProperties()
{
  dydx = m_matrix.m22() / m_matrix.m11();
  m_pen = createPen(this, m_color, DataPen, qMax(SolidLine, m_symbol));
  m_pen.setDashOffset(myDashOffset);
  setSize();
}
