/***************************************************************************
                          splineitem.h  -  description
                             -------------------
    begin                : Wed Feb 27 2002
    copyright            : (C) 2017 by Werner Stille
    email                : kpl@werner-stille.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef SPLINEITEM_H
#define SPLINEITEM_H

#include "funitem.h"

/**
  * Spline item class. Class for smoothing spline items for plotting spline
  * interpolations, derivatives or integrals by lines or markers.
  * @author Werner Stille
  */
class Q_DECL_EXPORT SplineItem : public ScaledItem
{
public:
  /** Specific QGraphicsItem type number for function type() */
  enum {
    Type = UserType + Spline
    /**<
      * Specific QGraphicsItem type number of SplineItem for function
      * type().
      */
  };
  /**
    * Constructor. Initializes corresponding to arguments.
    * @param degree degree of spline.
    * @param derivOrder order of derivative.
    * @param xLow lower x limit for spline integral.
    * @param color color for plot.
    * @param symbol symbol type.
    * @param fx normalization factor for x values.
    * @param fy normalization factor for y values.
    * @param parent pointer to parent item.
    * @param xMin minimum argument value.
    * @param xMax maximum argument value.
    * @param deltaX argument step size.
    * @param visible true for visible item.
    * @param fillStyle > 0 for filled polygon representation.
    * @param xOffset offset for x values.
    * @param yOffset offset for x values.
    * @param relSize relative size of lines or symbols.
    */
  SplineItem(int degree = 3, int derivOrder = 0, double xLow = 0,
             unsigned color = 0, Symbol symbol = SolidLine, double fx = 1,
             double fy = 1, QGraphicsItem* parent = 0, double xMin = 0,
             double xMax = 0, double deltaX = 0, bool visible = true,
             FillStyle fillStyle = NoBrush, double xOffset = 0,
             double yOffset = 0, double relSize = 1);
  /** Copy constructor. */
  SplineItem(const SplineItem& a);
  /**
    * Constructor. Initializes corresponding to plot file.
    * @param plo pointer to KConfigGroup object containing item data.
    * @param color default color for function plot.
    * @param symbol default symbol type.
    * @param parent pointer to parent item.
    */
  SplineItem(const KConfigGroup* plo, unsigned color = 0,
             Symbol symbol = SolidLine, QGraphicsItem* parent = 0);
  /** Destructor. */
  virtual ~SplineItem();
  /** Calculates bounding rectangle. */
  virtual QRectF boundingRect() const;
  /**
   * Calculates spline value table.
   * @param logX true for logarithmic steps
   * @return number of rows.
   */
  int calcTable(bool logX) const;
  /**
   * Calculates spline value table.
   * @param x array with x values.
   * @param values array for y values.
   * @param n number of points.
   * @param derivOrder order of derivative.
   * @param xLow lower x limit in case of spline integral.
   * @return error code.
   */
  int calcTable(const double* x, double* values, int n = 1, int derivOrder = 0,
                 double xLow = 0) const;
  /**
   * Copies spline data.
   * @param a pointer to source item.
   */
  void copySpline(const SplineItem* a);
  /** Returns degree of spline. */
  inline int degree() const { return m_degree; };
  /** Returns argument step size. */
  inline double deltaX() const { return m_deltaX; };
  /** Returns order of derivative. */
  inline int derivOrder() const { return m_derivOrder; };
  /**
   * Calculates normalization factors.
   * @param expX reference to logarithm of x normalization factor.
   * @param expY reference to logarithm of x normalization factor.
   * @param fx reference to x normalization factor.
   * @param fy reference to y normalization factor.
   * @param logX true for logarithmic x axis.
   */
  void exponent(int& expX, int& expY, double& fx, double& fy, bool logX) const;
  /**
   * Exports spline value table to file.
   * @param stream stream for output file.
   * @param separator column separator character.
   * @param format C format character.
   * @param precision number of digits.
   */
  void exportTable(QTextStream& stream, char separator, char format,
                   int precision) const;
  /**
   * Performs smoothing spline fit.
   * @param ai pointer to array item.
   * @param xMin lower boundary of approximation interval.
   * @param xMax upper boundary of approximation interval.
   * @param chiSquare chi-square.
   * @param numKnots number of knots.
   * @param smooth smoothing factor.
   * @param err pointer to structure containing the data error settings.
   * @param avgErr average error.
   * @return error code.
   */
  int fit(const ArrayItem* ai, double xMin, double xMax, double* chiSquare = 0,
          int* numKnots = 0, double smooth = 0,
          FunItem::DataErrorStruct* err = 0, double* avgErr = 0);
  /**
   * Performs smoothing spline fit.
   * @param x array with x values.
   * @param y array with y values.
   * @param errors array with error values.
   * @param n number of points.
   * @param xMin lower boundary of approximation interval.
   * @param xMax upper boundary of approximation interval.
   * @param chiSquare chi-square.
   * @param numKnots number of knots.
   * @param smooth smoothing factor.
   * @param err pointer to structure containing the data error settings.
   * @param avgErr average error.
   * @return error code.
   */
  int fit(const double* x, const double* y, const double* errors, int n,
          double xMin, double xMax, double* chiSquare = 0, int* numKnots = 0,
          double smooth = 0, FunItem::DataErrorStruct* err = 0,
          double* avgErr = 0);
  /**
   * Performs interpolating (non-smoothing) spline fit.
   * @param x array with x values.
   * @param y array with y values.
   * @param n number of points.
   * @param xMin lower boundary of approximation interval.
   * @param xMax upper boundary of approximation interval.
   * @return error code.
   */
  int interpolation(const double* x, const double *y, int n, double xMin,
                    double xMax);
  /**
   * Calculates minimum and maximum values.
   * @param xMin reference to minimum of x values.
   * @param xMax reference to maximum of x values.
   * @param yMin reference to minimum of y values.
   * @param yMax reference to maximum of y values.
   * @param logX true for logarithmic x axis.
   */
  void limits(double& xMin, double& xMax, double& yMin, double& yMax,
              bool logX) const;
  /** Returns number of knots. */
  inline int numKnots() const { return m_numKnots; };
  /** Paints item. */
  virtual void paint(QPainter* painter, const QStyleOptionGraphicsItem* option,
                     QWidget* widget = 0);
  /**
   * Finds roots of a cubic spline or its derivatives.
   * @param zero array for roots.
   * @param derivOrder order of derivative.
   * @param offset value of the spline or derivative.
   * @return error code.
   */
  int roots(QVector<double>& zero, int derivOrder = 0,
            double offset = 0) const;
  /** Sets offset for dashed lines. */
  void setDashOffset();
  /**
    * Sets degree of spline.
    * @param degree degree.
    */
  inline void setDegree(int degree) { m_degree = degree; };
  /**
    * Sets argument step size.
    * @param deltaX step size.
    */
  inline void setDeltaX(double deltaX) { m_deltaX = deltaX; };
  /**
    * Sets order of derivative.
    * @param derivOrder order.
    */
  inline void setDerivOrder(int derivOrder) { m_derivOrder = derivOrder; };
  /**
    * Sets lower x limit for spline integral.
    * @param xLow lower x limit.
    */
  inline void setXLow(double xLow) { m_xLow = xLow; };
  /**
    * Sets maximum argument value.
    * @param xMax maximum argument value.
    */
  inline void setXMax(double xMax) { m_xMax = xMax; };
  /**
    * Sets minimum argument value.
    * @param xMin minimum argument value.
    */
  inline void setXMin(double xMin) { m_xMin = xMin; };
  /** Returns the shape of the plot item. */
  virtual QPainterPath shape() const;
  /**
   * Sorts arrays and calculates error column.
   * @param x0 array with x values.
   * @param y0 array with y values.
   * @param errors0 array with error values.
   * @param n number of points.
   * @param err pointer to structure containing the data error settings.
   * @param x array for sorted x values.
   * @param y array for y values.
   * @param errors array for error values.
   * @return error code.
   */
  static int sortArrays(const double* x0, const double* y0,
                        const double* errors0, int& n,
                        FunItem::DataErrorStruct* err, double* x, double* y,
                        double* errors);
  /** Returns specific QGraphicsItem type number. */
  virtual int type() const { return Type; };
  /**
   * Writes into plot file.
   * @param plo KConfigGroup object.
   * @param format C format character.
   * @param precision number of digits.
   */
  void writePlo(KConfigGroup* plo, char format, int precision) const;
  /** Returns lower x limit for spline integral. */
  inline double xLow() const { return m_xLow; };
  /** Returns maximum argument value. */
  inline double xMax() const { return m_xMax; };
  /** Returns minimum argument value. */
  inline double xMin() const { return m_xMin; };
  /** Returns pointer to x function value array. */
  double* xValues() const;
  /** Returns pointer to y function value array. */
  double* yValues() const;

protected:
  /**
   * Paints or calculates bounding area.
   * @param painter QPainter object, 0 for calculating bounding area.
   */
  void paintCond(QPainter* painter) const;
  mutable bool logXo;
  int m_degree, m_numKnots, m_derivOrder;
  mutable int degreeo, numKnotso, derivOrdero;
  double m_xMin, m_xMax, m_deltaX, m_xLow;
  mutable double xMino, xMaxo, deltaXo, xLowo;
  mutable QVector<double> m_xValues, m_yValues, t, c;
};

#endif
