/***************************************************************************
                          textitem.cpp  -  description
                             -------------------
    begin                : Wed Apr 18 2001
    copyright            : (C) 2017 by Werner Stille
    email                : kpl@werner-stille.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <QAbstractTextDocumentLayout>
#include <QPainter>
#include <QTextFrame>
#include <KConfigGroup>
#include "kplscene.h"
#include "textitem.h"

TextItem::TextItem(double x0, double y0, unsigned color, QGraphicsItem* parent,
                   const QString& text, bool visible, Alignment alignment,
                   double relSize, double dir) :
 DecoItem(color, x0, y0, relSize), QGraphicsItem(parent),
 m_alignment(alignment), m_dir(dir), m_text(text)
{
  if (!visible)
    hide();
  setFlags(QGraphicsItem::ItemIsMovable | QGraphicsItem::ItemIsSelectable);
  init();
}

TextItem::TextItem(const TextItem& a) :
 DecoItem(a), QGraphicsItem(), m_alignment(a.m_alignment), m_dir(a.m_dir),
 m_text(a.m_text)
{
  setZValue(a.zValue());
  setFlags(QGraphicsItem::ItemIsMovable | QGraphicsItem::ItemIsSelectable);
  setSelected(a.isSelected());
  setVisible(a.isVisible());
  init();
}

TextItem::TextItem(const KConfigGroup* plo, unsigned color,
                   QGraphicsItem* parent) :
 DecoItem(plo, color), QGraphicsItem(parent)
{
  if (!plo->readEntry("active", true))
    hide();
  m_alignment = Alignment(plo->readEntry("align", int(AlignLeft)));
  m_dir = plo->readEntry("dir", 0.0);
  m_text = plo->readEntry("s", "");
  setFlags(QGraphicsItem::ItemIsMovable | QGraphicsItem::ItemIsSelectable);
  init();
}

TextItem::~TextItem()
{
}

QRectF TextItem::boundingRect() const
{
  QRectF r = m_doc.documentLayout()->blockBoundingRect(m_doc.begin());
  r.moveTopLeft(r.topLeft() -
                QPointF(0.5 * (m_alignment - 1) * r.width(),
                        QFontMetricsF(m_doc.defaultFont()).ascent()));
  return r;
}

void TextItem::paint(QPainter* painter, const QStyleOptionGraphicsItem*,
                     QWidget*)
{
  if (isSelected()) {
    painter->setPen(QPen(Qt::DashLine));
    painter->drawRect(boundingRect());
  }
  painter->translate(boundingRect().topLeft());
  m_doc.drawContents(painter);
}

void TextItem::setProperties(const QTransform& m)
{
  m_matrix = m;
  setProperties();
}

void TextItem::setSize()
{
  QFont f("Helvetica");
  f.setPixelSize(qRound(0.033333 * parentWidth(this) * m_relSize));
  m_doc.setDefaultFont(f);
}

void TextItem::setText(QTextDocument& doc, const QString& text, unsigned color)
{
  QString ht;
  QString s = text;
  if (!text.contains("</"))
    s.replace(QRegExp("<"), "&lt;");
  if (text.contains("</") || (!text.contains('#')))
    ht = s;
  else {
    static const QString face[] = {"Helvetica", "Symbol", "Dingbats"};
    static const QString fontcode = "NHBOSZI";
    static const QString levcode = "NUD";
    static const int iface[7] = {0, 0, 0, 0, 1, 2, 0};
    static const bool weight[7] = {false, false, true, false, false, false,
                                   true};
    static const bool italic[7] = {false, false, false, true, false, false,
                                   true};
    while (s.length()) {
      int l;
      QString c;
      if ((l = s.indexOf('#')) < 0)
        l = s.length();
      else
        c = s.mid(l);
      if (l)
        ht += s.left(l);
      if (!c.isNull()) {
        if (c.length() < 3)
          break;
        s = c.mid(3);
        int ifnt = fontcode.indexOf(c.at(1).toUpper());
        int ilev = levcode.indexOf(c.at(2).toUpper());
        if ((ifnt >= 0) && (ilev >= 0)) {
          int k = iface[ifnt];
          if (k)
            ht = ht + "<font face=" + face[k] + ">";
          if (weight[ifnt])
            ht = ht + "<b>";
          if (italic[ifnt])
            ht = ht + "<i>";
          if (ilev == 1)
            ht = ht + "<sup>";
          if (ilev == 2)
            ht = ht + "<sub>";
          if ((l = s.indexOf('#')) < 0)
            l = s.length();
          if (l) {
            ht += s.left(l);
            s.remove(0, l);
          }
          if (ilev == 1)
            ht = ht + "</sup>";
          if (ilev == 2)
            ht = ht + "</sub>";
          if (italic[ifnt])
            ht = ht + "</i>";
          if (weight[ifnt])
            ht = ht + "</b>";
          if (k)
            ht = ht + "</font>";
        } else {
          if ((l = s.indexOf('#')) < 0)
            l = s.length();
          if (l) {
            ht = ht + "#" + s.left(l);
            s.remove(0, l);
          }
        }
      }
    }
  }
  if (color)
    ht = "<font color=" + QColor(color).name() + ">" + ht + "</font>";
  doc.setHtml(ht);
}

void TextItem::update()
{
  init();
  prepareGeometryChange();
  setProperties();
  QGraphicsItem::update();
}

void TextItem::init()
{
  setText(m_doc, m_text, m_color);
  QTextFrameFormat tf = m_doc.rootFrame()->frameFormat();
  tf.setMargin(0);
  m_doc.rootFrame()->setFrameFormat(tf);
}

void TextItem::mouseReleaseEvent(QGraphicsSceneMouseEvent* event)
{
  QGraphicsItem::mouseReleaseEvent(event);
  ((KplScene*)scene())->updatePos(this);
}

void TextItem::setProperties()
{
  setPos(m_matrix.map(QPointF(m_x0, m_y0)));
  setRotation(-m_dir);
  setSize();
}

void TextItem::writePlo(KConfigGroup* plo) const
{
  plo->writeEntry("Type", "TEXTITEM");
  DecoItem::writePlo(plo, isVisible());
  plo->writeEntry("align", int(m_alignment));
  plo->writeEntry("dir", m_dir);
  plo->writeEntry("s", m_text);
}
