/***************************************************************************
                          utils.cpp  -  description
                             -------------------
    begin                : Sun Jan 7 2001
    copyright            : (C) 2017 by Werner Stille
    email                : kpl@werner-stille.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <QFileInfo>
#include <QLibrary>
#include <QProcess>
#include <QRegExp>
#include <QTemporaryFile>
#include <QUrl>
#include <KIO/Job>
#include <KJobWidgets>
#include <KJobUiDelegate>
#include <KLocalizedString>
#include <KMessageBox>
#include "utils.h"

bool Utils::download(const QUrl& src, QString& target, QWidget* window)
{
  if (src.isLocalFile()) {
    target = src.toLocalFile();
    return QFileInfo(target).isReadable();
  }
  if (target.isEmpty())
    target = tempFileName();
  KIO::Job* job = KIO::file_copy(src, QUrl::fromLocalFile(target), -1,
                                 KIO::Overwrite);
  KJobWidgets::setWindow(job, window);
  bool success = job->exec();
  if (!success)
    job->uiDelegate()->showErrorMessage();
  job->deleteLater();
  return success;
}

QStringList Utils::funcNames(const QString& libPath)
{
  QStringList list;
  QRegExp rx("^[0-9a-f]{8,} T ([^_\\s]+\\S*)");
  QFileInfo fInf(libPath);
  QString path = fInf.absolutePath() + "/" + fInf.baseName();
  QFile f(path + ".def");
  if (f.open(QIODevice::ReadOnly)) {
    while (!f.atEnd())
      if (rx.indexIn(f.readLine()) != -1)
        list << rx.capturedTexts()[1];
    f.close();
  } else {
    QProcess p;
    p.start(QString("nm \"%1\"").arg(libPath));
    p.waitForFinished();
    while (!p.atEnd())
      if (rx.indexIn(p.readLine()) != -1)
        list << rx.capturedTexts()[1];
    p.close();
  }
  if (list.isEmpty()) {
    f.setFileName(path + ".c");
    bool open = f.open(QIODevice::ReadOnly);
    if (!open) {
      f.setFileName(path + ".cpp");
      open = f.open(QIODevice::ReadOnly);
    }
    if (open) {
      QRegExp rx2("^\\s?double\\s+([^_\\s]+\\S*)\\s?\\(\\s?double\\s?");
      while (!f.atEnd())
        if (rx2.indexIn(f.readLine()) != -1)
          list << rx2.capturedTexts()[1];
      f.close();
      list.sort();
    }
  }
  return list;
}

bool Utils::getFuncAddr(const QString& path, const QString& name,
                        QLibrary** lib,
                        QFunctionPointer* fkt)
{
  *lib = 0;
  *fkt = 0;
  if (path.isEmpty() || name.isEmpty())
    KMessageBox::sorry(0, i18n("no library or no function specified!"));
  else {
    *lib = new QLibrary(path);
    if (!(*lib)->load())
      KMessageBox::error(0,  path + "\n" + i18n("library cannot be opened!"));
    else
      if ((*fkt = (*lib)->resolve(name.toLatin1())))
        return true;
      else
        KMessageBox::error(0, path + ": " + name + "\n" +
                           i18n("function name cannot be resolved!"));
    delete *lib;
    *lib = 0;
  }
  return false;
}

bool Utils::move(const QString& src, const QUrl& dest, QWidget* window)
{
  KIO::Job* job = KIO::file_move(url(src), dest, -1, KIO::Overwrite);
  KJobWidgets::setWindow(job, window);
  bool success = job->exec();
  if (!success)
    job->uiDelegate()->showErrorMessage();
  job->deleteLater();
  return success;
}

QString Utils::tempFileName(const QString& suffix)
{
  QTemporaryFile f;
  f.setFileTemplate(f.fileTemplate() + suffix);
  f.open();
  return f.fileName();
}

QUrl Utils::url(const QString& s)
{
  return QUrl::fromUserInput(s, QString(), QUrl::AssumeLocalFile);
}
