//
// C++ Interface: filedialog
//
// Description: 
//
//
// Author: Werner Stille <kpl@werner-stille.de>, (C) 2017
//
// Copyright: See COPYING file that comes with this distribution
//
//
#ifndef FILEDIALOG_H
#define FILEDIALOG_H

#include <qdialog.h>
#include <KFile>

class QComboBox;
class KFileWidget;

/**
  * File dialog class. Provides a dialog for selection of data files.
  * @author Werner Stille
  */
class FileDialog : public QDialog
{
  Q_OBJECT
public:
  /**
   * Constructor.
   * @param startDir starting directory.
   * @param filter file filter.
   * @param idec decimal character: Point or Comma.
   * @param mode mode of dialog as defined in KFile.
   * @param parent pointer to parent widget.
   */
  FileDialog(const QUrl& startDir, const QString& filter,
             ArrayItem::DecimalCharacter idec, KFile::Modes mode,
             QWidget* parent = 0);
  /**
   * Constructor.
   * @param startDir starting directory.
   * @param filter file filter.
   * @param mode mode of dialog as defined in KFile.
   * @param parent pointer to parent widget.
   */
  FileDialog(const QUrl& startDir, const QString& filter,
             KFile::Modes mode, QWidget* parent = 0);
  /** Destructor. */
  ~FileDialog();
  /**
   * Returns chosen decimal character.
   * @return Point or Comma.
   */
  ArrayItem::DecimalCharacter decimalCharacter();
  /**
   * Retrieves URL for reading.
   * @param idec decimal character: Point or Comma.
   * @param startDir starting directory.
   * @param parent pointer to parent widget.
   * @param filter file filter.
   * @return QUrl of chosen file.
   */
  static QUrl getReadUrl(ArrayItem::DecimalCharacter& idec,
                         const QString& startDir = QString::null,
                         QWidget* parent = 0,
                         const QString& filter = "*.dat *.DAT\n*");
  /**
   * Retrieves URLs for reading.
   * @param idec decimal character: Point or Comma.
   * @param parent pointer to parent widget.
   * @param filter file filter.
   * @return QUrl list of chosen files.
   */
  static QList<QUrl> getReadUrls(ArrayItem::DecimalCharacter& idec,
                                 QWidget* parent = 0,
                                 const QString& filter = "*.dat *.DAT\n*");
  /**
   * Retrieves URLs for reading.
   * @param parent pointer to parent widget.
   * @param filter file filter.
   * @return QUrl list of chosen files.
   */
  static QList<QUrl> getReadUrls(QWidget* parent = 0,
                                 const QString& filter = "*.plo *.PLO\n*");
  /**
   * Returns selected URL.
   * @return selected URL.
   */
  QUrl selectedUrl() const;
  /**
   * Returns list of selected URL.
   * @return list of selected URL.
   */
  QList<QUrl> selectedUrls() const;

protected slots:
  /** Reimplemented from QDialog. */
  void accept();
  /** Restores window size. */
  void slotRestoreSize();

protected:
  /**
   * Initializes dialog.
   * @param startDir starting directory.
   * @param filter file filter.
   * @param mode mode of dialog as defined in KFile.
   * @param customWidget pointer to optional custom widget.
   */
  void init(const QUrl& startDir, const QString& filter, KFile::Modes mode,
            QWidget* customWidget = 0);
  KFileWidget* fileWidget;
  QComboBox* comboBox;
};

#endif
