/***************************************************************************
                          funcdlg.cpp  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Sun Apr 25 1999
    copyright            : (C) 2017 by Werner Stille
    email                : kpl@werner-stille.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <QDesktopServices>
#include <QDialogButtonBox>
#include <QGroupBox>
#include <QKeyEvent>
#include <QLabel>
#include <QLayout>
#include <QTimer>
#include <QUndoStack>
#include <KColorButton>
#include <KLocalizedString>
#include <KMessageBox>
#include <KUrlRequester>
#include <KWindowConfig>
#include "brushdlg.h"
#include "choosefuncdlg.h"
#include "doubleedit.h"
#include "doublespinbox.h"
#include "editparameterdlg.h"
#include "funcdlg.h"
#include "kpl.h"
#include "kpldoc.h"
#include "kplsettings.h"
#include "kplutils.h"
#include "parfunitem.h"
#include "rootdlg.h"
#include "spinbox.h"
#include "splineitem.h"
#include "symboldlg.h"
#include "utils.h"

FuncDlg::FuncDlg(QWidget* parent, KplDoc* model, FunItem* fi0,
                 const QPoint& pos) :
 QDialog(parent), m(model), fi(fi0), fit(new FunItem(*fi0))
{
  setWindowTitle(i18n("Function"));
  QVBoxLayout* vbox = new QVBoxLayout(this);
  QGroupBox* g = new QGroupBox("y(x)");
  QGridLayout* grid = new QGridLayout;
  g->setLayout(grid);
  vbox->addWidget(g);
  grid->addWidget(new QLabel(i18n("Library")), 0, 0);
  grid->addWidget(fileName = new KUrlRequester(fi->urlY()), 0, 1);
  fileName->setStartDir(Utils::url(KplSettings::currentDir()));
  fileName->setFilter("*.so\n*");
  fileName->setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::Fixed);
  grid->addWidget(new QLabel(i18n("Function")), 1, 0);
  QHBoxLayout* hbox = new QHBoxLayout;
  grid->addLayout(hbox, 1, 1);
  hbox->addWidget(func = new QLineEdit(fi->nameY()));
  QPushButton* tb = new QPushButton("...");
  tb->setFixedSize(fileName->button()->size());
  hbox->addWidget(tb);
  connect(tb, SIGNAL(clicked()), SLOT(slotFunc()));
  grid->addWidget(new QLabel(i18n("Parameter")), 2, 0);
  grid->addLayout(hbox = new QHBoxLayout, 2, 1);
  QPushButton* b;
  hbox->addWidget(b = new QPushButton(i18n("Edit")));
  connect(b, SIGNAL(clicked()), SLOT(slotEditParameter()));
  hbox->addWidget(b = new QPushButton(i18n("Load")));
  connect(b, SIGNAL(clicked()), SLOT(slotGetParameter()));
  hbox->addWidget(b = new QPushButton(i18n("Save")));
  connect(b, SIGNAL(clicked()), SLOT(slotSaveParameter()));
  vbox->addWidget(g = new QGroupBox(i18n("Range")));
  grid = new QGridLayout;
  g->setLayout(grid);
  grid->addWidget(new QLabel("xmin"), 0, 0);
  char format = KplDoc::presentation();
  int prec = KplSettings::precision();
  grid->addWidget(exMin = new DoubleEdit(fi->tMin(), format, prec), 0, 1);
  int w = exMin->fontMetrics().width(QString::number(-1e-123 / 3, format,
                                                     prec)) + 3;
  exMin->setMinimumWidth(w);
  grid->addItem(new QSpacerItem(20, 10, QSizePolicy::MinimumExpanding), 0, 2);
  grid->addWidget(new QLabel("xmax"), 0, 3);
  grid->addWidget(exMax = new DoubleEdit(fi->tMax(), format, prec), 0, 4);
  exMax->setMinimumWidth(w);
  grid->addWidget(new QLabel(QString::fromUtf8("\316\224x")), 1, 0);
  grid->addWidget(exStep = new DoubleEdit(fi->deltaT(), format, prec), 1, 1);
  exStep->setMinimumWidth(w);
  grid->addWidget(new QLabel(i18n("x shift")), 2, 0);
  QPointF steps = m->scene()->steps(fi->zValue());
  grid->addWidget(ex0 = new DoubleSpinBox(fi->xOffset(), steps.x()), 2, 1);
  grid->addWidget(new QLabel(i18n("y shift")), 2, 3);
  grid->addWidget(ey0 = new DoubleSpinBox(fi->yOffset(), steps.y()), 2, 4);
  grid->addWidget(new QLabel(i18n("x normalization")), 3, 0);
  grid->addWidget(efx = new DoubleEdit(fi->normX()), 3, 1);
  grid->addWidget(new QLabel(i18n("y normalization")), 3, 3);
  grid->addWidget(efy = new DoubleEdit(fi->normY()), 3, 4);
  vbox->addWidget(g = new QGroupBox(i18n("Representation")));
  grid = new QGridLayout;
  g->setLayout(grid);
  grid->addWidget(new QLabel(i18n("Symbol")), 0, 0);
  grid->addWidget(symb = new SpinBox(fi->symbol(), -17, 9, 1), 0, 1);
  int style = fi->fillStyle();
  grid->addWidget(bSymb = new SymbolButton(symb, false, style), 0, 2);
  grid->addItem(new QSpacerItem(20, 10, QSizePolicy::MinimumExpanding), 0, 3);
  grid->addWidget(new QLabel(i18n("Color")), 0, 4);
  grid->addWidget(colData = new KColorButton(fi->color()), 0, 5);
  colData->setMinimumWidth(50);
  grid->addWidget(new QLabel(i18n("Size")), 1, 0);
  grid->addWidget(eRelSize = new SpinBox(qRound(100 * fi->relSize()), 10, 800),
                  1, 1);
  eRelSize->setSuffix(" %");
  eRelSize->setDisabled(style);
  grid->addWidget(new QLabel(i18n("Fill pattern")), 1, 4);
  grid->addWidget(fill = new BrushButton(style), 1, 5);
  move(pos);
  vbox->addWidget(buttonBox = new QDialogButtonBox(QDialogButtonBox::Help |
                                                   QDialogButtonBox::Ok |
                                                   QDialogButtonBox::Apply |
                                                   QDialogButtonBox::Cancel));
  buttonBox->addButton(b = new QPushButton(i18n("&Export")),
                       QDialogButtonBox::ActionRole);
  connect(b, SIGNAL(clicked()), SLOT(slotExport()));
  buttonBox->addButton(b = new QPushButton(i18n("R&oots")),
                       QDialogButtonBox::ActionRole);
  connect(b, SIGNAL(clicked()), SLOT(slotRoots()));
  buttonBox->button(QDialogButtonBox::Apply)->setEnabled(false);
  buttonBox->button(QDialogButtonBox::Ok)->setEnabled(false);
  connect(fill, SIGNAL(changed(int)), SLOT(slotFillChanged(int)));
  connect(ex0, SIGNAL(valueChanged(int)), SLOT(slotApply()));
  connect(ey0, SIGNAL(valueChanged(int)), SLOT(slotApply()));
  connect(efx, SIGNAL(textEdited(const QString&)), SLOT(slotEdited()));
  connect(efy, SIGNAL(textEdited(const QString&)), SLOT(slotEdited()));
  connect(exMin, SIGNAL(textEdited(const QString&)), SLOT(slotEdited()));
  connect(exMax, SIGNAL(textEdited(const QString&)), SLOT(slotEdited()));
  connect(exStep, SIGNAL(textEdited(const QString&)), SLOT(slotEdited()));
  connect(fileName, SIGNAL(textChanged(const QString&)), SLOT(slotEdited()));
  connect(func, SIGNAL(textChanged(const QString&)), SLOT(slotEdited()));
  connect(symb, SIGNAL(valueChanged(int)), SLOT(slotEdited()));
  connect(eRelSize, SIGNAL(valueChanged(int)), SLOT(slotEdited()));
  connect(colData, SIGNAL(changed(const QColor&)), SLOT(slotEdited()));
  connect(buttonBox, SIGNAL(clicked(QAbstractButton*)),
          SLOT(slotButtonClicked(QAbstractButton*)));
  QTimer::singleShot(0, this, SLOT(slotRestoreSize()));
}

FuncDlg::~FuncDlg()
{
  KConfigGroup config = KSharedConfig::openConfig()->group("FuncDlg");
  KWindowConfig::saveWindowSize(windowHandle(), config);
  delete fit;
}

void FuncDlg::apply(bool ok)
{
  FunItem* fiNew = new FunItem(*fit);
  fiNew->setXOffset(ex0->value());
  fiNew->setYOffset(ey0->value());
  fiNew->setNormX(efx->value());
  fiNew->setNormY(efy->value());
  fiNew->setTMin(exMin->value());
  fiNew->setTMax(exMax->value());
  fiNew->setDeltaT(exStep->value());
  if (!fiNew->setFuncY(fileName->url(), func->text())) {
    delete fiNew;
    return;
  }
  fiNew->setColor(colData->color().rgb());
  fiNew->setSymbol(KplItem::Symbol(symb->value()));
  fiNew->setRelSize(0.01 * eRelSize->value());
  fiNew->setFillStyle(KplItem::FillStyle(fill->value()));
  m->undoStack()->push(new EditCommand(fiNew, fi, m));
  fi = fiNew;
  if (ok)
    accept();
  else {
    buttonBox->button(QDialogButtonBox::Apply)->setEnabled(false);
    buttonBox->button(QDialogButtonBox::Ok)->setEnabled(false);
  }
}

bool FuncDlg::getTemp()
{
  fit->setXOffset(ex0->value());
  fit->setYOffset(ey0->value());
  fit->setNormX(efx->value());
  fit->setNormY(efy->value());
  fit->setTMin(exMin->value());
  fit->setTMax(exMax->value());
  fit->setDeltaT(exStep->value());
  if (!fit->setFuncY(fileName->url(), func->text()))
    return false;
  fit->setColor(colData->color().rgb());
  fit->setSymbol(KplItem::Symbol(symb->value()));
  fit->setRelSize(0.01 * eRelSize->value());
  fit->setFillStyle(KplItem::FillStyle(fill->value()));
  return true;
}

void FuncDlg::keyPressEvent(QKeyEvent* e)
{
  if ((e->QInputEvent::modifiers() == Qt::NoModifier) &&
      (e->key() == Qt::Key_F1)) {
    buttonBox->button(QDialogButtonBox::Help)->animateClick();
    e->accept();
  } else
    QDialog::keyPressEvent(e);
}

void FuncDlg::slotApply()
{
  apply();
}

void FuncDlg::slotButtonClicked(QAbstractButton* button)
{
  int id = buttonBox->standardButton(button);
  switch (id) {
    case QDialogButtonBox::Apply:
    case QDialogButtonBox::Ok:
      apply(id == QDialogButtonBox::Ok);
      break;
    case QDialogButtonBox::Help:
      QDesktopServices::openUrl(QUrl(QStringLiteral("help:/kpl/func.html")));
      break;
    case QDialogButtonBox::Cancel:
      reject();
  }
}

void FuncDlg::slotEdited()
{
  buttonBox->button(QDialogButtonBox::Apply)->setEnabled(true);
  buttonBox->button(QDialogButtonBox::Ok)->setEnabled(true);
}

void FuncDlg::slotEditParameter()
{
  EditParameterDlg dlg(this, fit->parY(), fileName->url().path(), func->text(),
                       true);
  connect(&dlg, SIGNAL(applyClicked()), SLOT(slotApply()));
  if (dlg.exec())
    slotEdited();
}

void FuncDlg::slotExport()
{
  if (getTemp()) {
    QUrl url;
    if (KplUtils::getWriteUrl(this, url, KplSettings::currentDir())) {
      QFile f(url.isLocalFile() ? url.path() : Utils::tempFileName());
      bool success = f.open(QIODevice::WriteOnly);
      if (success) {
        QTextStream ts(&f);
        fit->exportTable(ts, KplDoc::separator(), KplDoc::presentation(),
                         KplSettings::precision());
        f.close();
        if (!url.isLocalFile())
          success = Utils::move(f.fileName(), url, this);
        if (success)
          KplUtils::setCurrentDir(url);
      }
      if (!success)
        KMessageBox::error(this, i18n("while trying to open"));
    }
  }
}

void FuncDlg::slotFillChanged(int brush)
{
  symb->setDisabled(brush);
  bSymb->setDisabled(brush);
  eRelSize->setDisabled(brush);
  slotEdited();
}

void FuncDlg::slotFunc()
{
  KplUtils::getFunc(this, fileName->url(), func);
}

void FuncDlg::slotGetParameter()
{
  if (m->getPar(this, fit->parY()))
    slotEdited();
}

void FuncDlg::slotRestoreSize()
{
  KWindowConfig::restoreWindowSize(windowHandle(),
    KSharedConfig::openConfig()->group("FuncDlg"));
}

void FuncDlg::slotRoots()
{
  if (getTemp()) {
    double n = fit->calcTable();
    if ((n > 3) && (fit->tMax() > fit->tMin())) {
      SplineItem spl;
      spl.interpolation(fit->tValues(), fit->yValues(), n, fit->tMin(),
                        fit->tMax());
      RootDlg dlg(this, m, &spl);
      dlg.exec();
    } else
      KMessageBox::sorry(this, i18n("invalid range!"));
  }
}

void FuncDlg::slotSaveParameter()
{
  m->saveFunPar(this, fit->parY());
}

ParFuncDlg::ParFuncDlg(QWidget* parent, KplDoc* model, ParFunItem* fi0,
                       const QPoint& pos) :
 QDialog(parent), m(model), fi(fi0), fit(new ParFunItem(*fi0))
{
  setWindowTitle(i18n("Parametric function"));
  QVBoxLayout* vbox = new QVBoxLayout(this);
  QGroupBox* g = new QGroupBox("x(t)");
  vbox->addWidget(g);
  QGridLayout* grid = new QGridLayout;
  g->setLayout(grid);
  grid->addWidget(new QLabel(i18n("Library")), 0, 0);
  grid->addWidget(fileNameX = new KUrlRequester(fi->urlX()), 0, 1);
  fileNameX->setStartDir(Utils::url(KplSettings::currentDir()));
  fileNameX->setFilter("*.so\n*");
  fileNameX->setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::Fixed);
  grid->addWidget(new QLabel(i18n("Function")), 1, 0);
  QHBoxLayout* hbox = new QHBoxLayout;
  grid->addLayout(hbox, 1, 1);
  hbox->addWidget(funcX = new QLineEdit(fi->nameX()));
  QPushButton* tb;
  hbox->addWidget(tb = new QPushButton("..."));
  tb->setFixedSize(fileNameX->button()->size());
  connect(tb, SIGNAL(clicked()), SLOT(slotFuncX()));
  grid->addWidget(new QLabel(i18n("Parameter")), 2, 0);
  grid->addLayout(hbox = new QHBoxLayout, 2, 1);
  QPushButton* b;
  hbox->addWidget(b = new QPushButton(i18n("Edit")));
  connect(b, SIGNAL(clicked()), SLOT(slotEditParameterX()));
  hbox->addWidget(b = new QPushButton(i18n("Load")));
  connect(b, SIGNAL(clicked()), SLOT(slotGetParameterX()));
  hbox->addWidget(b = new QPushButton(i18n("Save")));
  connect(b, SIGNAL(clicked()), SLOT(slotSaveParameterX()));
  vbox->addWidget(g = new QGroupBox("y(t)"));
  grid = new QGridLayout;
  g->setLayout(grid);
  grid->addWidget(new QLabel(i18n("Library")), 0, 0);
  grid->addLayout(hbox = new QHBoxLayout, 0, 1);
  grid->addWidget(fileNameY = new KUrlRequester(fi->urlY()), 0, 1);
  fileNameY->setStartDir(Utils::url(KplSettings::currentDir()));
  fileNameY->setFilter("*.so\n*");
  fileNameY->setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::Fixed);
  grid->addWidget(new QLabel(i18n("Function")), 1, 0);
  grid->addLayout(hbox = new QHBoxLayout, 1, 1);
  hbox->addWidget(funcY = new QLineEdit(fi->nameY()));
  hbox->addWidget(tb = new QPushButton("..."));
  tb->setFixedSize(fileNameY->button()->size());
  connect(tb, SIGNAL(clicked()), SLOT(slotFuncY()));
  grid->addWidget(new QLabel(i18n("Parameter")), 2, 0);
  grid->addLayout(hbox = new QHBoxLayout, 2, 1);
  hbox->addWidget(b = new QPushButton(i18n("Edit")));
  connect(b, SIGNAL(clicked()), SLOT(slotEditParameterY()));
  hbox->addWidget(b = new QPushButton(i18n("Load")));
  connect(b, SIGNAL(clicked()), SLOT(slotGetParameterY()));
  hbox->addWidget(b = new QPushButton(i18n("Save")));
  connect(b, SIGNAL(clicked()), SLOT(slotSaveParameterY()));
  vbox->addWidget(g = new QGroupBox(i18n("Range")));
  grid = new QGridLayout;
  g->setLayout(grid);
  grid->addWidget(new QLabel("tmin"), 0, 0);
  char format = KplDoc::presentation();
  int prec = KplSettings::precision();
  grid->addWidget(etMin = new DoubleEdit(fi->tMin(), format, prec), 0, 1);
  int w = etMin->fontMetrics().width(QString::number(-1e-123 / 3, format,
                                     prec)) + 3;
  etMin->setMinimumWidth(w);
  grid->addItem(new QSpacerItem(20, 10, QSizePolicy::MinimumExpanding), 0, 2);
  grid->addWidget(new QLabel("tmax"), 0, 3);
  grid->addWidget(etMax = new DoubleEdit(fi->tMax(), format, prec), 0, 4);
  etMax->setMinimumWidth(w);
  grid->addWidget(new QLabel(QString::fromUtf8("\316\224t")), 1, 0);
  grid->addWidget(etStep = new DoubleEdit(fi->deltaT(), format, prec), 1, 1);
  etStep->setMinimumWidth(w);
  grid->addWidget(new QLabel(i18n("x shift")), 2, 0);
  QPointF steps = m->scene()->steps(fi->zValue());
  grid->addWidget(ex0 = new DoubleSpinBox(fi->xOffset(), steps.x()), 2, 1);
  grid->addWidget(new QLabel(i18n("y shift")), 2, 3);
  grid->addWidget(ey0 = new DoubleSpinBox(fi->yOffset(), steps.y()), 2, 4);
  grid->addWidget(new QLabel(i18n("x normalization")), 3, 0);
  grid->addWidget(efx = new DoubleEdit(fi->normX()), 3, 1);
  grid->addWidget(new QLabel(i18n("y normalization")), 3, 3);
  grid->addWidget(efy = new DoubleEdit(fi->normY()), 3, 4);
  vbox->addWidget(g = new QGroupBox(i18n("Representation")));
  g->setLayout(grid = new QGridLayout);
  grid->addWidget(new QLabel(i18n("Symbol")), 0, 0);
  grid->addWidget(symb = new SpinBox(fi->symbol(), -17, 9), 0, 1);
  int style = fi->fillStyle();
  grid->addWidget(bSymb = new SymbolButton(symb, false, style), 0, 2);
  grid->addItem(new QSpacerItem(20, 10, QSizePolicy::MinimumExpanding), 0, 3);
  grid->addWidget(new QLabel(i18n("Color")), 0, 4);
  grid->addWidget(colData = new KColorButton(fi->color()), 0, 5);
  colData->setMinimumWidth(50);
  grid->addWidget(new QLabel(i18n("Size")), 1, 0);
  grid->addWidget(eRelSize = new SpinBox(qRound(100 * fi->relSize()), 10, 800),
                  1, 1);
  eRelSize->setSuffix(" %");
  eRelSize->setDisabled(style);
  grid->addWidget(new QLabel(i18n("Fill pattern")), 1, 4);
  grid->addWidget(fill = new BrushButton(style), 1, 5);
  move(pos);
  vbox->addWidget(buttonBox = new QDialogButtonBox(QDialogButtonBox::Help |
                                                   QDialogButtonBox::Ok |
                                                   QDialogButtonBox::Apply |
                                                   QDialogButtonBox::Cancel));
  buttonBox->addButton(b = new QPushButton(i18n("&Export")),
                       QDialogButtonBox::ActionRole);
  connect(b, SIGNAL(clicked()), SLOT(slotExport()));
  buttonBox->button(QDialogButtonBox::Apply)->setEnabled(false);
  buttonBox->button(QDialogButtonBox::Ok)->setEnabled(false);
  connect(fill, SIGNAL(changed(int)), SLOT(slotFillChanged(int)));
  connect(ex0, SIGNAL(valueChanged(int)), SLOT(slotApply()));
  connect(ey0, SIGNAL(valueChanged(int)), SLOT(slotApply()));
  connect(efx, SIGNAL(textEdited(const QString&)), SLOT(slotEdited()));
  connect(efy, SIGNAL(textEdited(const QString&)), SLOT(slotEdited()));
  connect(etMin, SIGNAL(textEdited(const QString&)), SLOT(slotEdited()));
  connect(etMax, SIGNAL(textEdited(const QString&)), SLOT(slotEdited()));
  connect(etStep, SIGNAL(textEdited(const QString&)), SLOT(slotEdited()));
  connect(fileNameX, SIGNAL(textChanged(const QString&)), SLOT(slotEdited()));
  connect(funcX, SIGNAL(textChanged(const QString&)), SLOT(slotEdited()));
  connect(fileNameY, SIGNAL(textChanged(const QString&)), SLOT(slotEdited()));
  connect(funcY, SIGNAL(textChanged(const QString&)), SLOT(slotEdited()));
  connect(symb, SIGNAL(valueChanged(int)), SLOT(slotEdited()));
  connect(eRelSize, SIGNAL(valueChanged(int)), SLOT(slotEdited()));
  connect(colData, SIGNAL(changed(const QColor&)), SLOT(slotEdited()));
  connect(buttonBox, SIGNAL(clicked(QAbstractButton*)),
          SLOT(slotButtonClicked(QAbstractButton*)));
  QTimer::singleShot(0, this, SLOT(slotRestoreSize()));
}

ParFuncDlg::~ParFuncDlg()
{
  KConfigGroup config = KSharedConfig::openConfig()->group("ParFuncDlg");
  KWindowConfig::saveWindowSize(windowHandle(), config);
  delete fit;
}

bool ParFuncDlg::getTemp()
{
  fit->setXOffset(ex0->value());
  fit->setYOffset(ey0->value());
  fit->setNormX(efx->value());
  fit->setNormY(efy->value());
  fit->setTMin(etMin->value());
  fit->setTMax(etMax->value());
  fit->setDeltaT(etStep->value());
  if (!fit->setFuncX(fileNameX->url(), funcX->text()))
    return false;
  if (!fit->setFuncY(fileNameY->url(), funcY->text()))
    return false;
  fit->setColor(colData->color().rgb());
  fit->setSymbol(KplItem::Symbol(symb->value()));
  fit->setRelSize(0.01 * eRelSize->value());
  fit->setFillStyle(KplItem::FillStyle(fill->value()));
  return true;
}

void ParFuncDlg::keyPressEvent(QKeyEvent* e)
{
  if ((e->QInputEvent::modifiers() == Qt::NoModifier) &&
      (e->key() == Qt::Key_F1)) {
    buttonBox->button(QDialogButtonBox::Help)->animateClick();
    e->accept();
  } else
    QDialog::keyPressEvent(e);
}

void ParFuncDlg::slotApply()
{
  apply();
}

void ParFuncDlg::apply(bool ok)
{
  ParFunItem* fiNew = new ParFunItem(*fit);
  fiNew->setXOffset(ex0->value());
  fiNew->setYOffset(ey0->value());
  fiNew->setNormX(efx->value());
  fiNew->setNormY(efy->value());
  fiNew->setTMin(etMin->value());
  fiNew->setTMax(etMax->value());
  fiNew->setDeltaT(etStep->value());
  if (!fiNew->setFuncX(fileNameX->url(), funcX->text())) {
    delete fiNew;
    return;
  }
  if (!fiNew->setFuncY(fileNameY->url(), funcY->text())) {
    delete fiNew;
    return;
  }
  fiNew->setColor(colData->color().rgb());
  fiNew->setSymbol(KplItem::Symbol(symb->value()));
  fiNew->setRelSize(0.01 * eRelSize->value());
  fiNew->setFillStyle(KplItem::FillStyle(fill->value()));
  m->undoStack()->push(new EditCommand(fiNew, fi, m));
  fi = fiNew;
  if (ok)
    accept();
  else {
    buttonBox->button(QDialogButtonBox::Apply)->setEnabled(false);
    buttonBox->button(QDialogButtonBox::Ok)->setEnabled(false);
  }
}

void ParFuncDlg::slotButtonClicked(QAbstractButton* button)
{
  int id = buttonBox->standardButton(button);
  switch (id) {
    case QDialogButtonBox::Apply:
    case QDialogButtonBox::Ok:
      apply(id == QDialogButtonBox::Ok);
      break;
    case QDialogButtonBox::Help:
      QDesktopServices::openUrl(QUrl(QStringLiteral("help:/kpl/"
                                                    "parfunc.html")));
      break;
    case QDialogButtonBox::Cancel:
      reject();
  }
}

void ParFuncDlg::slotEdited()
{
  buttonBox->button(QDialogButtonBox::Apply)->setEnabled(true);
  buttonBox->button(QDialogButtonBox::Ok)->setEnabled(true);
}

void ParFuncDlg::slotEditParameterX()
{
  EditParameterDlg dlg(this, fit->parX(), fileNameX->url().path(), funcX->text(),
                       true);
  connect(&dlg, SIGNAL(applyClicked()), SLOT(slotApply()));
  if (dlg.exec())
    slotEdited();
}

void ParFuncDlg::slotEditParameterY()
{
  EditParameterDlg dlg(this, fit->parY(), fileNameY->url().path(),
                       funcY->text(), true);
  connect(&dlg, SIGNAL(applyClicked()), SLOT(slotApply()));
  if (dlg.exec())
    slotEdited();
}

void ParFuncDlg::slotExport()
{
  if (getTemp()) {
    QUrl url;
    if (KplUtils::getWriteUrl(this, url, KplSettings::currentDir())) {
      QFile f(url.isLocalFile() ? url.path() : Utils::tempFileName());
      bool success = f.open(QIODevice::WriteOnly);
      if (success) {
        QTextStream ts(&f);
        fit->exportTable(ts, KplDoc::separator(), KplDoc::presentation(),
                         KplSettings::precision());
        f.close();
        if (!url.isLocalFile())
          success = Utils::move(f.fileName(), url, this);
        if (success)
          KplUtils::setCurrentDir(url);
      }
      if (!success)
        KMessageBox::error(this, i18n("while trying to open"));
    }
  }
}

void ParFuncDlg::slotFillChanged(int brush)
{
  symb->setDisabled(brush);
  bSymb->setDisabled(brush);
  eRelSize->setDisabled(brush);
  slotEdited();
}

void ParFuncDlg::slotFuncX()
{
  KplUtils::getFunc(this, fileNameX->url(), funcX);
}

void ParFuncDlg::slotFuncY()
{
  KplUtils::getFunc(this, fileNameY->url(), funcY);
}

void ParFuncDlg::slotGetParameterX()
{
  if (m->getPar(this, fit->parX()))
    slotEdited();
}

void ParFuncDlg::slotGetParameterY()
{
  if (m->getPar(this, fit->parY()))
    slotEdited();
}

void ParFuncDlg::slotRestoreSize()
{
  KWindowConfig::restoreWindowSize(windowHandle(),
    KSharedConfig::openConfig()->group("ParFuncDlg"));
}

void ParFuncDlg::slotSaveParameterX()
{
  m->saveFunPar(this, fit->parX());
}

void ParFuncDlg::slotSaveParameterY()
{
  m->saveFunPar(this, fit->parY());
}
