/***************************************************************************
                          kpldoc.cpp  -  description
                             -------------------

    This file is a part of kpl - a program for graphical presentation of
    data sets and functions.

    begin                : Sat Apr 24 15:14:00 MEST 1999
    copyright            : (C) 2017 by Werner Stille
    email                : kpl@werner-stille.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <QApplication>
#include <QClipboard>
#include <QFileInfo>
#include <QGraphicsItem>
#include <QMimeData>
#include <KLocalizedString>
#include <KMessageBox>
#include <KNotification>
#include "arcdlg.h"
#include "arcitem.h"
#include "array3ditem.h"
#include "arraydlg.h"
#include "arrowdlg.h"
#include "arrowitem.h"
#include "ellipsedlg.h"
#include "ellipseitem.h"
#include "filedialog.h"
#include "fitdlg.h"
#include "framedlg.h"
#include "frameitem.h"
#include "fun3ditem.h"
#include "funcdlg.h"
#include "imagedlg.h"
#include "imageitem.h"
#include "kpl.h"
#include "kpldoc.h"
#include "kplsettings.h"
#include "kplutils.h"
#include "legenddlg.h"
#include "legenditem.h"
#include "linedlg.h"
#include "lineitem.h"
#include "parfunitem.h"
#include "rectdlg.h"
#include "rectitem.h"
#include "scalebardlg.h"
#include "scalebaritem.h"
#include "splfitdlg.h"
#include "splinedlg.h"
#include "splineitem.h"
#include "textdlg.h"
#include "textitem.h"
#include "threeddlg.h"
#include "utils.h"

const double KplDoc::Snap = 0.05;

/**
 * Applies changes to item and forces redrawing.
 * @param item pointer to plot item.
 */
static void update(QGraphicsItem* item);
/**
 * Update child items of a plot item.
 * @param item pointer to plot item.
 */
static void updateChildren(QGraphicsItem* item);

void update(QGraphicsItem* item)
{
  if (item)
    switch (item->type()) {
      case ArcItem::Type:
        static_cast<KplScene*>(item->scene())->setDashOffsets();
        break;
      case ArrayItem::Type:
        qgraphicsitem_cast<ArrayItem*>(item)->update();
        break;
      case Array3DItem::Type:
        qgraphicsitem_cast<Array3DItem*>(item)->update();
        updateChildren(item);
        break;
      case ArrowItem::Type:
        qgraphicsitem_cast<ArrowItem*>(item)->update();
        break;
      case EllipseItem::Type:
        qgraphicsitem_cast<EllipseItem*>(item)->update();
        break;
      case FrameItem::Type:
        qgraphicsitem_cast<FrameItem*>(item)->update();
        updateChildren(item);
        break;
      case FunItem::Type:
        qgraphicsitem_cast<FunItem*>(item)->update();
        break;
      case Fun3DItem::Type:
        qgraphicsitem_cast<Fun3DItem*>(item)->update();
        updateChildren(item);
        break;
      case ImageItem::Type:
        qgraphicsitem_cast<ImageItem*>(item)->update();
        break;
      case LegendItem::Type:
        qgraphicsitem_cast<LegendItem*>(item)->update();
        break;
      case LineItem::Type:
        static_cast<KplScene*>(item->scene())->setDashOffsets();
        break;
      case ParFunItem::Type:
        qgraphicsitem_cast<ParFunItem*>(item)->update();
        break;
      case RectItem::Type:
        qgraphicsitem_cast<RectItem*>(item)->update();
        break;
      case ScaleBarItem::Type:
        qgraphicsitem_cast<ScaleBarItem*>(item)->update();
        break;
      case SplineItem::Type:
        qgraphicsitem_cast<SplineItem*>(item)->update();
        break;
      case TextItem::Type:
        qgraphicsitem_cast<TextItem*>(item)->update();
        break;
    }
}

void updateChildren(QGraphicsItem* item)
{
  foreach (QGraphicsItem* child, item->childItems())
    switch (child->type()) {
      case ArcItem::Type:
        qgraphicsitem_cast<ArcItem*>(child)->setPenWidth();
        break;
      case ArrowItem::Type:
        qgraphicsitem_cast<ArrowItem*>(child)->setPenWidth();
        break;
      case ArrayItem::Type:
        qgraphicsitem_cast<ArrayItem*>(child)->setSize();
        break;
      case EllipseItem::Type:
        qgraphicsitem_cast<EllipseItem*>(child)->setPenWidth();
        break;
      case FunItem::Type:
        qgraphicsitem_cast<FunItem*>(child)->setSize();
        break;
      case LegendItem::Type:
        qgraphicsitem_cast<LegendItem*>(child)->setSize();
        break;
      case LineItem::Type:
        qgraphicsitem_cast<LineItem*>(child)->setPenWidth();
        break;
      case ParFunItem::Type:
        qgraphicsitem_cast<ParFunItem*>(child)->setSize();
        break;
      case RectItem::Type:
        qgraphicsitem_cast<RectItem*>(child)->setPenWidth();
        break;
      case ScaleBarItem::Type:
        qgraphicsitem_cast<ScaleBarItem*>(child)->setSize();
        break;
      case SplineItem::Type:
        qgraphicsitem_cast<SplineItem*>(child)->setSize();
        break;
      case TextItem::Type:
        qgraphicsitem_cast<TextItem*>(child)->setSize();
    }
}

DeleteCommand::DeleteCommand(QGraphicsItem* it, KplDoc* d,
                             QUndoCommand* parent) :
 QUndoCommand(parent), item(it), parentItem(it->parentItem()), newFrame(0),
 childItems(it->childItems()), doc(d)
{
  if (childItems.count()) {
    QList<QGraphicsItem*> list = doc->scene()->items(Qt::AscendingOrder);
    int i = list.indexOf(item) - 1;
    while (i >= 0) {
      QGraphicsItem* item2 = list.at(i--);
      if (KplScene::framelike(item2->type()) && item2->isVisible()) {
        newFrame = item2;
        break;
      }
    }
  }
}

DeleteCommand::~DeleteCommand()
{
}

void DeleteCommand::redo()
{
  int z = qRound(item->zValue());
  foreach (QGraphicsItem* item2, item->childItems()) {
    item2->setParentItem(newFrame);
    ::update(item2);
  }
  doc->scene()->removeItem(item);
  foreach (QGraphicsItem* item2, doc->scene()->QGraphicsScene::items()) {
    int z2 = qRound(item2->zValue());
    if (z2 > z)
      item2->setZValue(z2 - 1);
  }
  doc->scene()->setDashOffsets();
  doc->reset();
}

void DeleteCommand::undo()
{
  int z = qRound(item->zValue());
  foreach (QGraphicsItem* item2, doc->scene()->QGraphicsScene::items()) {
    int z2 = qRound(item2->zValue());
    if (z2 >= z)
      item2->setZValue(z2 + 1);
  }
  if (parentItem)
    item->setParentItem(parentItem);
  else
    doc->scene()->addItem(item);
  foreach (QGraphicsItem* item2, childItems)
    item2->setParentItem(item);
  ::update(item);
  doc->scene()->setDashOffsets();
  doc->reset();
}

EditCommand::EditCommand(QGraphicsItem* it, QGraphicsItem* old, KplDoc* d,
                         QUndoCommand* parent) :
 QUndoCommand(parent), item(it), oldItem(old), parentItem(old->parentItem()),
 doc(d)
{
}

EditCommand::~EditCommand()
{
  if (!item->scene())
    delete item;
}

void EditCommand::redo()
{
  swapItems(item, oldItem);
}

void EditCommand::undo()
{
  swapItems(oldItem, item);
}

void EditCommand::swapItems(QGraphicsItem* n, QGraphicsItem* o)
{
  foreach (QGraphicsItem* item2, o->childItems())
    item2->setParentItem(n);
  doc->scene()->removeItem(o);
  if (parentItem)
    n->setParentItem(parentItem);
  else
    doc->scene()->addItem(n);
  ::update(n);
  doc->reset();
}

MoveCommand::MoveCommand(int iSource, int iDest, KplDoc* d) :
 is(iSource), id(iDest), doc(d)
{
}

MoveCommand::~MoveCommand()
{
}

void MoveCommand::redo()
{
  swapItems(is, id);
}

void MoveCommand::undo()
{
  swapItems(id, is);
}

void MoveCommand::swapItems(int i1, int i2)
{
  QGraphicsItem* item;
  QGraphicsItem* frame = 0;
  QList<QGraphicsItem*> list = doc->scene()->items(Qt::AscendingOrder);
  for (int i = 0; i < list.count(); ++i) {
    bool upd = false;
    item = list.at(i);
    if (i == i2) {
      item = list.at(i1);
      item->setZValue(i2);
      upd = true;
    } else {
      if (i1 < i2) {
        if ((i >= i1) && (i < i2)) {
          item = list.at(i + 1);
          item->setZValue(i);
          upd = true;
        }
      } else {
        if ((i > i2) && (i <= i1)) {
          item = list.at(i - 1);
          item->setZValue(i);
          upd = true;
        }
      }
    }
    bool framelike = KplScene::framelike(item->type());
    if ((!framelike) && (item->parentItem() != frame)) {
      item->setParentItem(frame);
      upd = true;
    }
    if (upd)
      ::update(item);
    if (framelike && item->isVisible())
      frame = item;
  }
  doc->reset();
}

NewCommand::NewCommand(QGraphicsItem* it, KplDoc* d, QUndoCommand* parent) :
 QUndoCommand(parent), item(it), parentItem(it->parentItem()), doc(d)
{
  if (KplScene::framelike(item->type())) {
    int z = qRound(item->zValue());
    foreach (QGraphicsItem* item2, doc->scene()->items(Qt::AscendingOrder))
      if (item2->zValue() >= z) {
        if (KplScene::framelike(item2->type()) && item2->isVisible())
          break;
        childItems << item2;
        oldParents << item2->parentItem();
      }
  }
}

NewCommand::~NewCommand()
{
 if (!item->scene())
   delete item;
}

void NewCommand::redo()
{
  int z = qRound(item->zValue());
  foreach (QGraphicsItem* item2, doc->scene()->QGraphicsScene::items())
    if (item2 != item) {
      int z2 = qRound(item2->zValue());
      if (z2 >= z)
        item2->setZValue(z2 + 1);
  }
  if (parentItem)
    item->setParentItem(parentItem);
  else
    if (item->scene() != doc->scene())
      doc->scene()->addItem(item);
  if (KplScene::framelike(item->type()))
    doc->scene()->setProperties(item);
  foreach (QGraphicsItem* item2, childItems)
    item2->setParentItem(item);
  ::update(item);
  doc->reset();
}

void NewCommand::undo()
{
  for (int i = 0; i < childItems.count(); ++i) {
    QGraphicsItem* item2 = childItems[i];
    item2->setParentItem(oldParents.at(i));
    ::update(item2);
  }
  doc->scene()->removeItem(item);
  int z = qRound(item->zValue());
  foreach (QGraphicsItem* item2, doc->scene()->QGraphicsScene::items()) {
    int z2 = qRound(item2->zValue());
    if (z2 > z)
      item2->setZValue(z2 - 1);
  }
  doc->reset();
}

ToggleActiveCommand::ToggleActiveCommand(QGraphicsItem* it, KplDoc* d) :
 item(it), newFrame(0), doc(d)
{
  if (KplScene::framelike(item->type())) {
    int z = item->zValue();
    foreach (QGraphicsItem* item2, doc->scene()->items(Qt::AscendingOrder)) {
      int z2 = item2->zValue();
      bool framelike = KplScene::framelike(item2->type());
      if (framelike && item2->isVisible()) {
        if (z2 < z)
          newFrame = item2;
        else
          if (z2 > z)
            break;
      } else
        if ((z2 > z) && (!framelike))
          childItems << item2;
    }
  }
}

ToggleActiveCommand::~ToggleActiveCommand()
{
}

void ToggleActiveCommand::redo()
{
  bool visible = !item->isVisible();
  switch (item->type()) {
    case ArrayItem::Type:
      qgraphicsitem_cast<ArrayItem*>(item)->updateGeometry();
      break;
    case FunItem::Type:
      qgraphicsitem_cast<FunItem*>(item)->updateGeometry();
      break;
    case ParFunItem::Type:
      qgraphicsitem_cast<ParFunItem*>(item)->updateGeometry();
      break;
    case SplineItem::Type:
      qgraphicsitem_cast<SplineItem*>(item)->updateGeometry();
  }
  foreach (QGraphicsItem* item2, childItems) {
    item2->setParentItem(visible ? item : newFrame);
    if (!visible)
      ::update(item2);
  }
  item->setVisible(visible);
  ::update(item);
  ::update(newFrame);
  doc->scene()->setDashOffsets();
  doc->reset();
}

void ToggleActiveCommand::undo()
{
  redo();
}

ToggleCollapsedCommand::ToggleCollapsedCommand(FrameItem* it, KplDoc* d) :
 item(it), doc(d)
{
}

ToggleCollapsedCommand::~ToggleCollapsedCommand()
{
}

void ToggleCollapsedCommand::redo()
{
  item->toggleCollapsed();
  doc->reset();
}

void ToggleCollapsedCommand::undo()
{
  redo();
}

KplDoc::KplDoc(QObject* parent) :
 QAbstractItemModel(parent), defSel(false),
 app((QWidget*) parent), m_fileType(Unknown),
 m_undoStack(new QUndoStack(this)), m_chiSquare(0)
{
  m_scene = new KplScene(QSizeF(KplSettings::formatX(),
                                KplSettings::formatY()), this);
  m_fitNonLinear = KplSettings::fitNonLinear();
  m_fitMaxIterations = KplSettings::fitMaxIterations();
  m_fitTolerance = KplSettings::fitTolerance();
  memset(m_pLast, 0, sizeof(m_pLast));
  memset(m_pErr, 0, sizeof(m_pErr));
  FunItem::stringToPar(KplSettings::fitParam(), m_pFit);
  readFitPar(m_fitPar);
  m_undoStack->setUndoLimit(1000);
  connect(m_undoStack, SIGNAL(cleanChanged(bool)), app, SLOT(updWin()));
  connect(m_scene, SIGNAL(positionChanged(QGraphicsItem*)),
          SLOT(updatePos(QGraphicsItem*)));
  connect(m_scene, SIGNAL(changed(const QList<QRectF> &)), SLOT(notify()));
}

KplDoc::~KplDoc()
{
}

int KplDoc::columnCount(const QModelIndex&) const
{
  return 3;
}

void KplDoc::copyItems(QModelIndexList indexes, bool cut)
{
  QMimeData* md = mimeData(indexes);
  if (md) {
    QApplication::clipboard()->setMimeData(md);
    if (cut)
      removeItems(indexes);
  }
}

QVariant KplDoc::data(const QModelIndex& index, int role) const
{
  if (index.isValid()) {
    QGraphicsItem* item = static_cast<QGraphicsItem*>(index.internalPointer());
    if (role == Qt::DisplayRole)
      switch (index.column()) {
        case 1:
          return typeName(item);
        case 2:
          return settings(item);
      }
    if (role == Qt::DecorationRole  && (index.column() == 2))
      switch (item->type()) {
        case ArcItem::Type:
          return qgraphicsitem_cast<ArcItem*>(item)->pixmap();
        case ArrayItem::Type:
          return qgraphicsitem_cast<ArrayItem*>(item)->pixmap();
        case ArrowItem::Type:
          return qgraphicsitem_cast<ArrowItem*>(item)->pixmap();
        case EllipseItem::Type:
          return qgraphicsitem_cast<EllipseItem*>(item)->pixmap();
        case FunItem::Type:
          return qgraphicsitem_cast<FunItem*>(item)->pixmap();
        case LegendItem::Type:
          return qgraphicsitem_cast<LegendItem*>(item)->pixmap();
        case LineItem::Type:
          return qgraphicsitem_cast<LineItem*>(item)->pixmap();
        case ParFunItem::Type:
          return qgraphicsitem_cast<ParFunItem*>(item)->pixmap();
        case RectItem::Type:
          return qgraphicsitem_cast<RectItem*>(item)->pixmap();
        case ScaleBarItem::Type:
          return qgraphicsitem_cast<ScaleBarItem*>(item)->pixmap();
        case SplineItem::Type:
          return qgraphicsitem_cast<SplineItem*>(item)->pixmap();
      }
    if (role == Qt::CheckStateRole && !index.column())
      return item->isVisible() ? Qt::Checked : Qt::Unchecked;
  }
  return QVariant();
}

bool KplDoc::dropMimeData(const QMimeData* data, Qt::DropAction action,
                          int, int, const QModelIndex& parent)
{
  QUrl url = data->urls().first();
  if (action == Qt::MoveAction)
    removeItems(mimeDataIndexes);
  if ((action == Qt::MoveAction) || (action == Qt::CopyAction))
    insertItems(url, (parent.isValid()) ?
      static_cast<QGraphicsItem*>(parent.internalPointer())->zValue() : -1);
  return true;
}

void KplDoc::editItem(QModelIndexList indexes, QPoint pos)
{
  if (!indexes.isEmpty()) {
    QGraphicsItem* item =
      static_cast<QGraphicsItem*>(indexes.first().internalPointer());
    switch (item->type()) {
      case ArcItem::Type: {
        ArcDlg(app, this, qgraphicsitem_cast<ArcItem*>(item), pos).exec();
        break;
      }
      case ArrayItem::Type: {
        ArrayDlg(app, this, qgraphicsitem_cast<ArrayItem*>(item), pos).exec();
        break;
      }
      case Array3DItem::Type: {
        Array3DDlg(app, this, qgraphicsitem_cast<Array3DItem*>(item), pos).exec();
        break;
      }
      case ArrowItem::Type: {
        ArrowDlg(app, this, qgraphicsitem_cast<ArrowItem*>(item), pos).exec();
        break;
      }
      case EllipseItem::Type: {
        EllipseDlg(app, this, qgraphicsitem_cast<EllipseItem*>(item),
                   pos).exec();
        break;
      }
      case FrameItem::Type: {
        FrameDlg(app, this, qgraphicsitem_cast<FrameItem*>(item), pos).exec();
        break;
      }
      case FunItem::Type: {
        FuncDlg(app, this, qgraphicsitem_cast<FunItem*>(item), pos).exec();
        break;
      }
      case Fun3DItem::Type: {
        Fun3DDlg(app, this, qgraphicsitem_cast<Fun3DItem*>(item), pos).exec();
        break;
      }
      case ImageItem::Type: {
        ImageDlg(app, this, qgraphicsitem_cast<ImageItem*>(item), pos).exec();
        break;
      }
      case LegendItem::Type: {
        LegendDlg(app, this, qgraphicsitem_cast<LegendItem*>(item),
                  pos).exec();
        break;
      }
      case LineItem::Type: {
        LineDlg(app, this, qgraphicsitem_cast<LineItem*>(item), pos).exec();
        break;
      }
      case ParFunItem::Type: {
        ParFuncDlg(app, this, qgraphicsitem_cast<ParFunItem*>(item),
                   pos).exec();
        break;
      }
      case RectItem::Type: {
        RectDlg(app, this, qgraphicsitem_cast<RectItem*>(item), pos).exec();
        break;
      }
      case ScaleBarItem::Type: {
        ScaleBarDlg(app, this, qgraphicsitem_cast<ScaleBarItem*>(item),
                    pos).exec();
        break;
      }
      case SplineItem::Type: {
        SplineDlg(app, this, qgraphicsitem_cast<SplineItem*>(item),
                  pos).exec();
        break;
      }
      case TextItem::Type: {
        TextDlg(app, this, qgraphicsitem_cast<TextItem*>(item), pos).exec();
        break;
      }
    }
  }
}

const FunItem::DataErrorStruct& KplDoc::errorModel()
{
  m_errorModel.fitErrCol = KplSettings::fitErrorColumn();
  m_errorModel.errModPath = KplSettings::errorModelPath();
  m_errorModel.errModName = KplSettings::errorModelName();
  m_errorModel.errModArg = FunItem::DataErrorStruct::
    ErrorModelArgument(KplSettings::errorModelArgument());
  FunItem::stringToPar(KplSettings::errorModelParam(), m_errorModel.parErrMod);
  return m_errorModel;
}

void KplDoc::fitItems(QModelIndexList indexes)
{
  QList<const ArrayItem*> arr;
  QList<FunItem*> fun;
  QList<SplineItem*> spl;
  if (defSel) {
    foreach (QGraphicsItem* item, m_scene->items(Qt::AscendingOrder))
      switch (item->type()) {
        case ArrayItem::Type:
          arr << qgraphicsitem_cast<ArrayItem*>(item);
          break;
        case FunItem::Type:
          fun << qgraphicsitem_cast<FunItem*>(item);
          break;
        case SplineItem::Type:
          spl << qgraphicsitem_cast<SplineItem*>(item);
      }
  } else
    foreach (QModelIndex index, indexes)
      if (index.isValid()) {
        QGraphicsItem* item =
          static_cast<QGraphicsItem*>(index.internalPointer());
if (!item) qDebug("KplDoc::fitItems(): item = 0");
        switch (item->type()) {
          case ArrayItem::Type:
            arr << qgraphicsitem_cast<ArrayItem*>(item);
            break;
          case FunItem::Type:
            fun << qgraphicsitem_cast<FunItem*>(item);
            break;
          case SplineItem::Type:
            spl << qgraphicsitem_cast<SplineItem*>(item);
            break;
        }
      }
  if (fun.count()) {
    emit displayMessage(i18n("Fitting function parameters..."));
    FitDlg dlg(app, this, &arr, fun, FitDlg::ShowDlg);
    dlg.exec();
  } else {
    emit displayMessage(i18n("Fitting spline..."));
    SplFitDlg dlg(app, this, arr.first(), spl, FitDlg::ShowDlg);
    dlg.exec();
  }
}

bool KplDoc::fittable(QModelIndexList indexes)
{
  int nArr = 0;
  int nFun = 0;
  int nSpl = 0;
  foreach (QGraphicsItem* item, m_scene->QGraphicsScene::items())
    switch (item->type()) {
      case ArrayItem::Type:
        nArr++;
        break;
      case FunItem::Type:
        nFun++;
        break;
      case SplineItem::Type:
        nSpl++;
    }
  defSel = ((nArr == 1) && ((nFun + nSpl) == 1));
  if (defSel)
    return true;
  nArr = nFun = nSpl = 0;
  foreach (QModelIndex index, indexes) {
    QGraphicsItem* item = static_cast<QGraphicsItem*>(index.internalPointer());
    switch (item->type()) {
      case ArrayItem::Type:
        nArr++;
        break;
      case FunItem::Type:
        nFun++;
        break;
      case SplineItem::Type:
        nSpl++;
    }
  }
  return (nArr && nFun && (!nSpl) && (nArr <= nFun)) ||
         ((nArr == 1) && (nSpl && (!nFun)));
}

Qt::ItemFlags KplDoc::flags(const QModelIndex& index) const
{
  Qt::ItemFlags fl = QAbstractItemModel::flags(index) | Qt::ItemIsDropEnabled;
  if (index.isValid()) {
    if (index.column() == 2)
      fl |= Qt::ItemIsDragEnabled;
    if (!index.column())
      fl |= Qt::ItemIsUserCheckable;
  }
  return fl;
}

bool KplDoc::getPar(QWidget* parent, double* p)
{
  bool success = false;
  ArrayItem::DecimalCharacter idec = KplUtils::decimalCharacter();
  QUrl url = FileDialog::getReadUrl(idec, KplSettings::currentDir(), parent,
                                    "*.par (*.par);;* (*)");
  if (!url.isEmpty())
    success = loadPar(url, parent, p, idec);
  return success;
}

QVariant KplDoc::headerData(int section, Qt::Orientation orientation,
                            int role) const
{
  if (orientation == Qt::Horizontal && role == Qt::DisplayRole)
    switch (section) {
      case 0:
        return i18n("Active");
      case 1:
        return i18n("Type");
      case 2:
        return i18n("Settings");
    }
  return QVariant();
}

QModelIndex KplDoc::index(int row, int column, const QModelIndex& parent) const
{
  if (hasIndex(row, column, parent)) {
    QGraphicsItem* item = 0;
    QList<QGraphicsItem*> list = m_scene->items(Qt::AscendingOrder);
    if (parent.isValid()) {
      QGraphicsItem* parentItem =
        static_cast<QGraphicsItem*>(parent.internalPointer());
if (!parentItem) qDebug("KplDoc::index(): parentItem = 0");
      if (parentItem) {
if (!KplScene::framelike(parentItem->type())) qDebug("KplDoc::index(): parent != framelike");
if (list.indexOf(parentItem) + row + 1 >= list.size()) qDebug("KplDoc::index(): list index out of range");
        item = list.at(list.indexOf(parentItem) + row + 1);
if (item && (KplScene::framelike(item->type()))) qDebug("KplDoc::index(): item framelike");
      }
    } else {
      int count = 0;
      bool frame = false;
      foreach (QGraphicsItem* item2, list) {
        bool framelike = KplScene::framelike(item2->type());
        if (count == row) {
          if ((!frame) || framelike) {
            item = item2;
            break;
          }
        } else
          if (framelike) {
            frame = true;
            count++;
          } else
            if (!frame)
              count++;
      }
    }
if (!item) qDebug() << "KplDoc::index(): item = 0";
    if (item)
      return createIndex(row, column, item);
  }
  return QModelIndex();
}

void KplDoc::insertItem(QGraphicsItem* item, int z)
{
  m_scene->insertItem(item, z);
  m_undoStack->push(new NewCommand(item, this));
}

void KplDoc::insertItems(const QUrl& url, int z)
{
  QList<QGraphicsItem*> list;
  if (readPlo(url, list)) {
    foreach (QGraphicsItem* item, list) {
      insertItem(item, z);
      if (z >= 0)
        z++;
    }
    reset();
  }
}

bool KplDoc::loadPar(const QUrl& url, QWidget* parent, double* p,
                     ArrayItem::DecimalCharacter idec)
{
  bool success = false;
  QString fn;
  if (Utils::download(url, fn, parent)) {
    QFile f(fn);
    if (f.open(QIODevice::ReadOnly)) {
      memset(p, 0, FunItem::NumParMax * sizeof(double));
      QTextStream t(&f);
      int i = 0;
      while (!t.atEnd() && (i < FunItem::NumParMax)){
        QString line = t.readLine();
        if ((!line.isEmpty()) && (line[0] != '#')) {
          ArrayItem::translate(line, idec);
          QStringList list = line.split(QRegExp(idec ? "[ \t]" : "[ ,\t]"),
                                                QString::SkipEmptyParts);
          p[i++] = list[0].toDouble();
        }
      }
      f.close();
      KplUtils::setCurrentDir(url);
      success = true;
    } else
      KMessageBox::error(parent,
                         i18n("while trying to open parameter file"));
    if (!url.isLocalFile())
      QFile::remove(fn);
  }
  return success;
}

QMimeData* KplDoc::mimeData(const QModelIndexList& indexes) const
{
  mimeDataIndexes.clear();
  QMimeData* md = 0;
  QList<QGraphicsItem*> list;
  foreach (QModelIndex index, indexes)
    if (index.isValid()) {
      list << static_cast<QGraphicsItem*>(index.internalPointer());
      mimeDataIndexes << index;
    }
  if (list.count()) {
    QUrl url = QUrl::fromLocalFile(Utils::tempFileName(".plo"));
    if (m_scene->savePlo(url, list, true, presentation(),
                         KplSettings::precision())) {
      md = new QMimeData;
      md->setUrls(QList<QUrl>() << url);
    } else
      KMessageBox::sorry(app, i18n("Writing to this file not possible."));
  }
  return md;
}

QStringList KplDoc::mimeTypes() const
{
  return QStringList() << "text/uri-list";
}

void KplDoc::newDocument()
{
  m_undoStack->clear();
  m_scene->clear();
  m_url.setUrl(i18n("Untitled"));
  reset();
  nyLeg = 0;
}

void KplDoc::newItem(int type, int z, const QPoint& pos)
{
  double fx = 1;
  double fy = 1;
  switch (type) {
    case KplItem::Array:
    case KplItem::Function:
    case KplItem::ParFunction:
    case KplItem::Spline:
      QList<QGraphicsItem*> list = m_scene->items(Qt::DescendingOrder);
      if (z && list.count()) {
        int z1 = (z < 0) ? list.first()->zValue() : (z - 1);
        foreach (QGraphicsItem* item, list)
          if (item->zValue() <= z1) {
            int ityp = item->type();
            if (KplScene::framelike(ityp))
              break;
            ScaledItem* sci = 0;
            switch (ityp) {
              case ArrayItem::Type:
                sci = qgraphicsitem_cast<ArrayItem*>(item);
                break;
              case FunItem::Type:
                sci = qgraphicsitem_cast<FunItem*>(item);
                break;
              case ParFunItem::Type:
                sci = qgraphicsitem_cast<ParFunItem*>(item);
                break;
              case SplineItem::Type:
                sci = qgraphicsitem_cast<SplineItem*>(item);
            }
            if (sci) {
              fx = sci->normX();
              fy = sci->normY();
              break;
            }
          }
      }
  }
  switch (type) {
    case KplItem::Arc: {
      ArcItem* ai = new ArcItem(KplSettings::x0() + 0.5 *
                                KplSettings::width(), KplSettings::y0() +
                                0.5 * KplSettings::height(),
                                KplSettings::colFrame().rgb());
      insertItem(ai, z);
      ArcDlg(app, this, ai, pos).exec();
      break;
    }
    case KplItem::Array: {
      ArrayItem* ai = new ArrayItem(KplSettings::colData().rgb(),
        KplSettings::iX(), KplSettings::iY(), KplSettings::iE(),
        KplSettings::errBars(), KplItem::Symbol(KplSettings::symbol()),
        KplUtils::decimalCharacter(), QUrl(), fx, fy);
      insertItem(ai, z);
      ArrayDlg(app, this, ai, pos).exec();
      break;
    }
    case KplItem::Array3D: {
      Array3DItem* ai = new Array3DItem(KplSettings::x0(), KplSettings::y0(),
        KplSettings::width(), KplSettings::height(),
        KplSettings::colFrame().rgb(), KplSettings::colGrid().rgb(),
        KplSettings::colData().rgb(), KplSettings::logX(), KplSettings::logY(),
        false, FrameItem::GridMode(KplSettings::gridMode()),
        ThreeDItem::Grid, KplSettings::width(), KplSettings::height(),
        KplSettings::width(), KplSettings::textX(), KplSettings::textY(), "",
        KplSettings::textH(), 0.01 * KplSettings::relSize(), KplSettings::iX(),
        KplSettings::iY(), 2, KplSettings::iE(), KplSettings::errBars(),
        KplUtils::decimalCharacter());
      insertItem(ai, z);
      Array3DDlg(app, this, ai, pos).exec();
      break;
    }
    case KplItem::Arrow: {
      ArrowItem* ai = new ArrowItem(KplSettings::x0() + 0.5 *
                                    KplSettings::width(), KplSettings::y0() +
                                    0.5 * KplSettings::height(),
                                    KplSettings::colFrame().rgb());
      insertItem(ai, z);
      ArrowDlg(app, this, ai, pos).exec();
      break;
    }
    case KplItem::Ellipse: {
      EllipseItem* ei = new EllipseItem(KplSettings::x0() + 0.5 *
                                        KplSettings::width(),
                                        KplSettings::y0() + 0.5 *
                                        KplSettings::height(),
                                        KplSettings::colFrame().rgb());
      insertItem(ei, z);
      EllipseDlg(app, this, ei, pos).exec();
      break;
    }
    case KplItem::Frame: {
      FrameItem* fi = new FrameItem(KplSettings::x0(), KplSettings::y0(),
                                    KplSettings::width(),
                                    KplSettings::height(),
                                    KplSettings::colFrame().rgb(),
                                    KplSettings::colGrid().rgb(),
                                    KplSettings::logX(), KplSettings::logY(),
                                    FrameItem::GridMode(KplSettings::gridMode()),
                                    KplSettings::width(),
                                    KplSettings::height(),
                                    KplSettings::textX(), KplSettings::textY(),
                                    KplSettings::textH(),
                                    0.01 * KplSettings::relSize());
      insertItem(fi, z);
      FrameDlg(app, this, fi, pos).exec();
      break;
    }
    case KplItem::Function: {
      FunItem* fi = new FunItem(QUrl(), "", KplSettings::colData().rgb(),
                                KplItem::Symbol(KplSettings::symbol()), fx,
                                fy);
      insertItem(fi, z);
      FuncDlg(app, this, fi, pos).exec();
      break;
    }
    case KplItem::Function3D: {
      Fun3DItem* fi = new Fun3DItem(KplSettings::x0(), KplSettings::y0(),
                                    KplSettings::width(),
                                    KplSettings::height(),
                                    KplSettings::colFrame().rgb(),
                                    KplSettings::colGrid().rgb(),
                                    KplSettings::colData().rgb(),
                                    KplSettings::logX(), KplSettings::logY(),
                                    false,
                                    FrameItem::GridMode(KplSettings::gridMode()),
                                    ThreeDItem::Grid, KplSettings::width(),
                                    KplSettings::height(),
                                    KplSettings::width(),
                                    KplSettings::textX(), KplSettings::textY(),
                                    "", KplSettings::textH(),
                                    0.01 * KplSettings::relSize());
      insertItem(fi, z);
      Fun3DDlg(app, this, fi, pos).exec();
      break;
    }
    case KplItem::Image: {
      ImageItem* ii = new ImageItem(KplSettings::x0() + 0.5 *
                                    KplSettings::width(), KplSettings::y0() +
                                    0.5 * KplSettings::height());
      insertItem(ii, z);
      ImageDlg(app, this, ii, pos).exec();
      break;
    }
    case KplItem::Legend: {
      double dx = KplSettings::width() / 15;
      LegendItem* li = new LegendItem(KplSettings::x0() + dx,
        KplSettings::y0() + KplSettings::height() - (++nyLeg) * dx,
        KplSettings::colData().rgb(), KplSettings::colFrame().rgb(),
        KplItem::Symbol(KplSettings::symbol()), dx, -0.18 * dx);
      insertItem(li, z);
      LegendDlg(app, this, li, pos).exec();
      break;
    }
    case KplItem::Line: {
      LineItem* li = new LineItem(KplSettings::x0() + 0.5 *
                                  KplSettings::width(), KplSettings::y0() +
                                  0.5 * KplSettings::height(),
                                  KplSettings::colFrame().rgb());
      insertItem(li, z);
      LineDlg(app, this, li, pos).exec();
      break;
    }
    case KplItem::ParFunction: {
      ParFunItem* fi = new ParFunItem(QUrl(), "", QUrl(), "",
        KplSettings::colData().rgb(),
        KplItem::Symbol(KplSettings::symbol()), fx, fy);
      insertItem(fi, z);
      ParFuncDlg(app, this, fi, pos).exec();
      break;
    }
    case KplItem::Rectangle: {
      RectItem* ri = new RectItem(KplSettings::x0() + 0.5 *
                                  KplSettings::width() - 1.5,
                                  KplSettings::y0() + 0.5 *
                                  KplSettings::height() - 1,
                                  KplSettings::colFrame().rgb());
      insertItem(ri, z);
      RectDlg(app, this, ri, pos).exec();
      break;
    }
    case KplItem::ScaleBar: {
      ScaleBarItem* si = new ScaleBarItem(KplSettings::x0() + 0.5 *
                                  KplSettings::width(), KplSettings::y0() +
                                  0.5 * KplSettings::height(),
                                  KplSettings::colFrame().rgb());
      insertItem(si, z);
      ScaleBarDlg(app, this, si, pos).exec();
      break;
    }
    case KplItem::Spline: {
      SplineItem* si = new SplineItem(3, 0, 0, KplSettings::colData().rgb(),
        KplItem::Symbol(KplSettings::symbol()), fx, fy);
      insertItem(si, z);
      SplineDlg(app, this, si, pos).exec();
      break;
    }
    case KplItem::Text: {
      TextItem* ti = new TextItem(KplSettings::x0() + 0.5 *
                                  KplSettings::width(), KplSettings::y0() +
                                  0.5 * KplSettings::height(),
                                  KplSettings::colFrame().rgb());
      insertItem(ti, z);
      TextDlg(app, this, ti, pos).exec();
      break;
    }
  }
  reset();
}

bool KplDoc::openDocument(const QUrl& url, ArrayItem::DecimalCharacter idec,
                          FileType type)
{
  m_fileType = (type == Unknown) ?
    ((QFileInfo(url.fileName()).suffix() == "plo") ? Plot : Data) : type;
  bool success;
  if (m_fileType == Data) {
    FrameItem* fri = 0;
    FrameItem* frt = 0;
    int z = -1;
    QList<QGraphicsItem*> list = m_scene->items(Qt::DescendingOrder);
    if (KplSettings::addItems() &&  !list.isEmpty()) {
      foreach (QGraphicsItem* item, list)
        if ((item->type() == FrameItem::Type) && item->isVisible()) {
          fri = qgraphicsitem_cast<FrameItem*>(item);
          break;
        }
      z = qRound(list.first()->zValue());
    } else
      newDocument();
    if (!fri) {
      frt = new FrameItem(KplSettings::x0(), KplSettings::y0(),
                          KplSettings::width(), KplSettings::height(),
                          KplSettings::colFrame().rgb(),
                          KplSettings::colGrid().rgb(), KplSettings::logX(),
                          KplSettings::logY(),
                          FrameItem::GridMode(KplSettings::gridMode()),
                          KplSettings::width(), KplSettings::height(),
                          KplSettings::textX(), KplSettings::textY(),
                          KplSettings::textH(), 0.01 * KplSettings::relSize());
      m_scene->insertItem(frt);
      z++;
    }
    ArrayItem* ai = new ArrayItem(KplSettings::colData().rgb(),
      KplSettings::iX(), KplSettings::iY(), KplSettings::iE(),
      KplSettings::errBars(), KplItem::Symbol(KplSettings::symbol()), idec,
      url, 1, 1, fri ? fri : frt, -1, 0,
      KplItem::FillStyle(KplSettings::fillStyle()));
    success = ai->numRows();
    if (success) {
      ai->setZValue(++z);
      ai->setProperties(m_scene->matrix());
      ai->setErrorBars(KplSettings::errBars() &&
                       (KplSettings::iE() < ai->numCols()));
      if (fri)
        swapParents(frt = new FrameItem(*fri), fri);
      frt->autoScale(KplSettings::normalization());
      setUrl(url);
      if (z == 1)
        if (KplSettings::autoPathHeader())
          frt->setHeader(m_url.isLocalFile() ? m_url.path() : m_url.url());
      QUndoCommand* macro = new QUndoCommand;
      if (fri) {
        swapParents(fri, frt);
        new NewCommand(ai, this, macro);
        new EditCommand(frt, fri, this, macro);
      } else {
        m_scene->removeItem(frt);
        new NewCommand(frt, this, macro);
        new NewCommand(ai, this, macro);
      }
      m_undoStack->push(macro);
      if (KplSettings::autoFit()) {
        double xmi, xma;
        ScaledItem::limits(xmi, xma, &ai->data()[ai->colX()][ai->rowStart()],
                           ai->count());
        FunItem* fi = new FunItem(KplSettings::fitPath(),
          KplSettings::fitName(), KplSettings::fitColor().rgb(),
          KplItem::Symbol(KplSettings::fitSymbol()), ai->normX(), ai->normY(),
          frt, xmi, xma, 0, true,
          KplItem::FillStyle(KplSettings::fitFillStyle()));
        memcpy(fi->parY(), m_pFit, FunItem::NumParMax * sizeof(double));
        fi->setZValue(++z);
        fi->setProperties(m_scene->matrix());
        int mode = FitDlg::AutoFit;
        if (KplSettings::fitShowDialog())
          mode |= FitDlg::ShowDlg;
        if (KplSettings::fitSaveParam())
          mode |= FitDlg::SavePar;
        if (KplSettings::fitFollow())
          mode |= FitDlg::Follow;
        QList<const ArrayItem*> arr;
        QList<FunItem*> fun;
        arr.append(ai);
        fun.append(fi);
        m_fitNonLinear = KplSettings::fitNonLinear();
        m_fitMaxIterations = KplSettings::fitMaxIterations();
        m_fitTolerance = KplSettings::fitTolerance();
        FitDlg dlg(app, this, &arr, fun, mode);
        dlg.slotFit();
        if (KplSettings::fitShowDialog())
          dlg.exec();
        new NewCommand(fi, this, macro);
      }
      reset();
    } else {
      delete ai;
      delete frt;
      reset();
      m_undoStack->undo();
    }
  } else {
    if (!KplSettings::addItems())
      newDocument();
    QList<QGraphicsItem*> list;
    success = readPlo(url, list);
    if (success) {
      foreach (QGraphicsItem* item, list)
        m_scene->insertItem(item);
      m_scene->setDashOffsets();
      setUrl(url);
      reset();
    }
  }
  if (!success)
    KMessageBox::error(app,
                       QString(i18n("while loading the file\n")) + url.path());
  return success;
}

QModelIndex KplDoc::parent(const QModelIndex& index) const
{
  if (index.isValid()) {
    QGraphicsItem* item = static_cast<QGraphicsItem*>(index.internalPointer());
if (!item) qDebug("KplDoc::parent(): item = 0");
    if (item && (!KplScene::framelike(item->type()))) {
      int row = 0;
      int row2 = 0;
      QGraphicsItem* parentItem = 0;
      foreach (QGraphicsItem* item2, m_scene->items(Qt::AscendingOrder)) {
        if (item2 == item)
          break;
        if (KplScene::framelike(item2->type())) {
          parentItem = item2;
          row = row2++;
        } else
          if (!parentItem)
            row2++;
      }
      if (parentItem)
        return createIndex(row, 0, parentItem);
    }
  }
  return QModelIndex();
}

char KplDoc::presentation()
{
  static const char pres[] = "eEfgG";
  return pres[KplSettings::presentation()];
}

void KplDoc::readFitPar(bool* fitPar)
{
  QStringList list = KplSettings::fitEnableParam().split(' ');
  int cnt = list.count();
  for (int i = 0; i < FunItem::NumParMax; ++i)
    fitPar[i] = (i < cnt) ? list[i].toInt() : false;
}

void KplDoc::removeItems(QModelIndexList indexes)
{
  QUndoCommand* macro = (indexes.count() > 1) ? new QUndoCommand : 0;
  foreach (QModelIndex index, indexes)
    if (index.isValid()) {
      QUndoCommand* cmd = new DeleteCommand(
        static_cast<QGraphicsItem*>(index.internalPointer()), this, macro);
      if (!macro)
        m_undoStack->push(cmd);
    }
  if (macro)
    m_undoStack->push(macro);
  reset();
}

void KplDoc::reset()
{
  endResetModel();
  for (int row = 0; row < rowCount(); ++row) {
    QModelIndex idx = index(row, 0);
    QGraphicsItem* item = static_cast<QGraphicsItem*>(idx.internalPointer());
    bool collapsed;
    switch (item->type()) {
      case Array3DItem::Type:
        collapsed = qgraphicsitem_cast<Array3DItem*>(item)->collapsed();
        break;
      case FrameItem::Type:
        collapsed = qgraphicsitem_cast<FrameItem*>(item)->collapsed();
        break;
      case Fun3DItem::Type:
        collapsed = qgraphicsitem_cast<Fun3DItem*>(item)->collapsed();
        break;
      default:
        collapsed = true;
    }
    if (!collapsed)
      emit expanded(idx);
  }
}

int KplDoc::rowCount(const QModelIndex& parent) const
{
  int count = 0;
  if (parent.column() <= 0) {
    QList<QGraphicsItem*> list = m_scene->items(Qt::AscendingOrder);
    if (parent.isValid()) {
      QGraphicsItem* parentItem =
        static_cast<QGraphicsItem*>(parent.internalPointer());
if (!parentItem) qDebug("KplDoc::rowCount(): parentItem = 0");
      if (parentItem && KplScene::framelike(parentItem->type())) {
        int z = qRound(parentItem->zValue());
        foreach (QGraphicsItem* item, list)
          if (qRound(item->zValue()) > z) {
            if (KplScene::framelike(item->type()))
              break;
            count++;
          }
      }
    } else {
      bool frame = false;
      foreach (QGraphicsItem* item, list)
        if (KplScene::framelike(item->type())) {
          frame = true;
          count++;
        } else
          if (!frame)
            count++;
    }
  }
  return count;
}

bool KplDoc::saveDocument(const QUrl& url, bool absolute)
{
  bool success = m_scene->savePlo(url, absolute, presentation(),
                                  KplSettings::precision());
  if (success)
    m_undoStack->setClean();
  else
    KMessageBox::sorry(app, i18n("Writing to this file not possible."));
  return success;
}

void KplDoc::saveFunPar(QWidget* parent, const double* p)
{
  QUrl url;
  if (KplUtils::getWriteUrl(parent, url, KplSettings::currentDir(),
                            "*.par (*.par);;* (*)"))
    saveParameter(parent, url, p, 0);
}

bool KplDoc::saveParameter(QWidget* parent, const QUrl& url, const double* p,
                           const double* err)
{
  QFile f(url.isLocalFile() ? url.path() : Utils::tempFileName());
  bool success = f.open(QIODevice::WriteOnly);
  if (success) {
    QTextStream t(&f);
    char format = presentation();
    int prec = KplSettings::precision();
    for (int i = 0; i < FunItem::NumParMax; i++) {
      t << QString::number(p[i], format, prec);
      if (err)
        t << separator() << QString::number(err[i], format, prec);
      t << "\n";
    }
    f.close();
    if (!url.isLocalFile())
      success = Utils::move(f.fileName(), url, parent);
    if (success)
      KplUtils::setCurrentDir(url);
  }
  if (!success)
    KMessageBox::error(parent, i18n("while trying to open parameter file"));
  return success;
}

char KplDoc::separator()
{
  static const char sep[] = " \t,";
  return sep[KplSettings::separator()];
}

bool KplDoc::setData(const QModelIndex& idx, const QVariant&, int role)
{
  if (idx.isValid() && role == Qt::CheckStateRole && !idx.column()) {
    m_undoStack->push(new ToggleActiveCommand(
      static_cast<QGraphicsItem*>(idx.internalPointer()), this));
    emit dataChanged(idx, idx);
    return true;
  }
  return false;
}

QString KplDoc::settings(QGraphicsItem* item)
{
  switch (item->type()) {
    case ArcItem::Type: {
      ArcItem* it = qgraphicsitem_cast<ArcItem*>(item);
      return QString("x = %1 cm, y = %2 cm, ").arg(it->x0()).arg(it->y0()) +
        i18n("size") +
        QString(" %1 cm x %2 cm").arg(it->width()).arg(it->height());
    }
    case ArrayItem::Type: {
      ArrayItem* it = qgraphicsitem_cast<ArrayItem*>(item);
      QString s = (it->internal() ?
                   i18n("internal data") :
                   (i18n("file") + QString(" %1").arg(it->url().fileName()))) +
        ", " + i18n("columns") +
        QString(" %1, %2").arg(it->colX()).arg(it->colY());
      if (it->errorBars())
        s += ", " + QString::number(it->colErr());
      return s;
    }
    case Array3DItem::Type: {
      Array3DItem* it = qgraphicsitem_cast<Array3DItem*>(item);
      return QString(it->internal() ?
                     i18n("internal data") :
                     (i18n("file") + QString(" %1").arg(it->url()
                      .fileName())));
    }
    case ArrowItem::Type: {
      ArrowItem* it = qgraphicsitem_cast<ArrowItem*>(item);
      return QString("x = %1 cm, y = %2 cm, ").arg(it->x0()).arg(it->y0()) +
        i18n("direction") + QString(" %1").arg(it->dir()) +
        QString::fromUtf8("\302\260");
    }
    case EllipseItem::Type: {
      EllipseItem* it = qgraphicsitem_cast<EllipseItem*>(item);
      return QString("x = %1 cm, y = %2 cm, ").arg(it->x0()).arg(it->y0()) +
        i18n("size") +
        QString(" %1 cm x %2 cm").arg(it->width()).arg(it->height());
    }
    case FunItem::Type: {
      FunItem* it = qgraphicsitem_cast<FunItem*>(item);
      QFileInfo fi(it->urlY().path());
      return QString("y = %1.%2(x)").arg(fi.baseName()).arg(it->nameY());
    }
    case Fun3DItem::Type: {
      Fun3DItem* it = qgraphicsitem_cast<Fun3DItem*>(item);
      QFileInfo fi(it->urlY().path());
      return QString("y = %1.%2(x,z)").arg(fi.baseName()).arg(it->nameY());
    }
    case FrameItem::Type: {
      FrameItem* it = qgraphicsitem_cast<FrameItem*>(item);
      return QString("x = %1 cm, y = %2 cm, ").arg(it->x0()).arg(it->y0()) +
        i18n("size") +
        QString(" %1 cm x %2 cm").arg(it->width()).arg(it->height());
    }
    case ImageItem::Type: {
      ImageItem* it = qgraphicsitem_cast<ImageItem*>(item);
      return i18n("file") + QString(" %1, x = %2 cm, y = %3 cm")
        .arg(it->url().fileName()).arg(it->x0()).arg(it->y0());
    }
    case LegendItem::Type: {
      LegendItem* it = qgraphicsitem_cast<LegendItem*>(item);
      return QString("\"%1\", x = %2 cm, y = %3 cm").arg(it->text())
        .arg(it->x0()).arg(it->y0());
    }
    case LineItem::Type: {
      LineItem* it = qgraphicsitem_cast<LineItem*>(item);
      return QString("x = %1 cm, y = %2 cm, ").arg(it->x0()).arg(it->y0()) +
        i18n("direction") + QString(" %1").arg(it->dir()) +
        QString::fromUtf8("\302\260");
    }
    case ParFunItem::Type: {
      ParFunItem* it = qgraphicsitem_cast<ParFunItem*>(item);
      QFileInfo fi1(it->urlX().path());
      QFileInfo fi2(it->urlY().path());
      return QString("x = %1.%2(t), y = %3.%4(t)")
        .arg(fi1.baseName()).arg(it->nameX()).arg(fi2.baseName())
        .arg(it->nameY());
    }
    case RectItem::Type: {
      RectItem* it = qgraphicsitem_cast<RectItem*>(item);
      return QString("x = %1 cm, y = %2 cm, ").arg(it->x0()).arg(it->y0()) +
        i18n("size") +
        QString(" %1 cm x %2 cm").arg(it->width()).arg(it->height());
    }
    case ScaleBarItem::Type: {
      ScaleBarItem* it = qgraphicsitem_cast<ScaleBarItem*>(item);
      return (it->vertical() ? i18n("Vertical") : i18n("Horizontal")) +
        QString(", x = %1 cm, y = %2 cm").arg(it->x0()).arg(it->y0());
    }
    case SplineItem::Type: {
      SplineItem* it = qgraphicsitem_cast<SplineItem*>(item);
      QString s = i18n("degree") + QString(" %1").arg(it->degree());
      if (it->derivOrder()) {
        s += ", ";
        switch (it->derivOrder()) {
          case -1:
            s += i18n("integral");
            break;
          case 1:
            s += i18n("1st derivative");
            break;
          case 2:
            s += i18n("2nd derivative");
            break;
          case 3:
            s += i18n("3rd derivative");
            break;
          case 4:
            s += i18n("4th derivative");
            break;
          case 5:
            s += i18n("5th derivative");
        }
      }
      return s;
    }
    case TextItem::Type: {
      TextItem* it = qgraphicsitem_cast<TextItem*>(item);
      return QString("\"%1\", x = %2 cm, y = %3 cm")
        .arg(it->text()).arg(it->x0()).arg(it->y0());
    }
    default:
      return QString();
  }
}

void KplDoc::setUrl(const QUrl& url)
{
  m_url = url;
  KplUtils::setCurrentDir(url);
  if (url.isLocalFile())
    m_time = QFileInfo(url.path()).lastModified();
}

Qt::DropActions KplDoc::supportedDropActions() const
{
  return Qt::CopyAction | Qt::MoveAction;
}

void KplDoc::swapParents(QGraphicsItem* n, QGraphicsItem* o) const
{
  foreach (QGraphicsItem* item, o->childItems())
    item->setParentItem(n);
  m_scene->removeItem(o);
  m_scene->addItem(n);
};

const QString& KplDoc::typeName(const QGraphicsItem* item)
{
  static const QString typeNames[] = {i18n("Frame"), i18n("Array"),
                                      i18n("Function"), i18n("Par. function"),
                                      i18n("Spline"), i18n("3D array"),
                                      i18n("3D function"), i18n("Legend"),
                                      i18n("Text"), i18n("Line"),
                                      i18n("Arrow"), i18n("Arc"),
                                      i18n("Rectangle"), i18n("Ellipse"),
                                      i18n("Image"), i18n("Scale bar")};
  return typeNames[qBound(0, item->type() - QGraphicsItem::UserType - 1, 15)];
}

void KplDoc::collapse(const QModelIndex& index)
{
  setCollapsed(index, true);
}

void KplDoc::expand(const QModelIndex& index)
{
  setCollapsed(index, false);
}

void KplDoc::zoom(const QRectF& rect)
{
  foreach (QGraphicsItem* item, m_scene->items(Qt::DescendingOrder))
    if ((item->type() == FrameItem::Type) && item->isVisible())
      if (item->boundingRect().intersects(rect)) {
        FrameItem* fr = qgraphicsitem_cast<FrameItem*>(item);
        QRectF r = fr->mapToUser(rect);
        FrameItem* frNew = new FrameItem(*fr);
        frNew->autoScale(r.left(), r.right(), 1, r.bottom(), r.top(), 1);
        QUndoCommand* macro = new QUndoCommand;
        foreach (QGraphicsItem* child, fr->childItems())
          switch (child->type()) {
            case FunItem::Type: {
              FunItem* fiNew =
                new FunItem(*qgraphicsitem_cast<FunItem*>(child));
              double fx = fiNew->normX();
              fiNew->setTMin(frNew->xMin() / fx);
              fiNew->setTMax(frNew->xMax() / fx);
              fiNew->setDeltaT(0);
              new EditCommand(fiNew, child, this, macro);
              break;
            }
            case SplineItem::Type: {
              SplineItem* siNew =
                new SplineItem(*qgraphicsitem_cast<SplineItem*>(child));
              double fx = siNew->normX();
              siNew->setXMin(frNew->xMin() / fx);
              siNew->setXMax(frNew->xMax() / fx);
              siNew->setDeltaX(0);
              new EditCommand(siNew, child, this, macro);
            }
        }
        new EditCommand(frNew, fr, this, macro);
        m_undoStack->push(macro);
        break;
      }
}

bool KplDoc::readPlo(const QUrl& url, QList<QGraphicsItem*>& list) const
{
  return KplScene::readPlo(url, list, app, KplSettings::colFrame().rgb(),
                           KplSettings::colGrid().rgb(),
                           KplSettings::colData().rgb(),
                           KplItem::Symbol(KplSettings::symbol()),
                           KplSettings::x0(), KplSettings::y0(),
                           KplSettings::width(), KplSettings::height(),
                           FrameItem::GridMode(KplSettings::gridMode()),
                           KplSettings::iX(), KplSettings::iY(),
                           KplSettings::iE(), KplSettings::errBars());
  return true;
}

void KplDoc::setCollapsed(const QModelIndex& index, bool state)
{
  if (index.isValid()) {
    QGraphicsItem* item = static_cast<QGraphicsItem*>(index.internalPointer());
    switch (item->type()) {
      case Array3DItem::Type: {
        Array3DItem* newitem = qgraphicsitem_cast<Array3DItem*>(item);
        if (newitem->collapsed() != state)
          m_undoStack->push(new ToggleCollapsedCommand(newitem, this));
        break;
      }
      case FrameItem::Type: {
        FrameItem* newitem = qgraphicsitem_cast<FrameItem*>(item);
        if (newitem->collapsed() != state)
          m_undoStack->push(new ToggleCollapsedCommand(newitem, this));
        break;
      }
      case Fun3DItem::Type: {
        Fun3DItem* newitem = qgraphicsitem_cast<Fun3DItem*>(item);
        if (newitem->collapsed() != state)
          m_undoStack->push(new ToggleCollapsedCommand(newitem, this));
        break;
      }
    }
  }
}

void KplDoc::updatePos(QGraphicsItem* item, const QPointF& p,
                       QUndoCommand* macro)
{
  switch (item->type()) {
    case ArcItem::Type:
      updatePos2(qgraphicsitem_cast<ArcItem*>(item), p, macro);
      break;
    case ArrowItem::Type:
      updatePos2(qgraphicsitem_cast<ArrowItem*>(item), p, macro);
      break;
    case EllipseItem::Type:
      updatePos2(qgraphicsitem_cast<EllipseItem*>(item), p, macro);
      break;
    case ImageItem::Type:
      updatePos2(qgraphicsitem_cast<ImageItem*>(item), p, macro);
      break;
    case LegendItem::Type:
      updatePos2(qgraphicsitem_cast<LegendItem*>(item), p, macro);
      break;
    case LineItem::Type:
      updatePos2(qgraphicsitem_cast<LineItem*>(item), p, macro);
      break;
    case RectItem::Type:
      updatePos2(qgraphicsitem_cast<RectItem*>(item), p, macro);
      break;
    case ScaleBarItem::Type:
      updatePos2(qgraphicsitem_cast<ScaleBarItem*>(item), p, macro);
      break;
    case TextItem::Type:
      updatePos2(qgraphicsitem_cast<TextItem*>(item), p, macro);
  }
}

void KplDoc::notify()
{
  if (m_scene->QGraphicsScene::items().count())
    KNotification::event("plotCompleted", "Kpl", i18n("Plot completed."));
}

void KplDoc::updatePos(QGraphicsItem* item)
{
  switch (item->type()) {
    case Array3DItem::Type:
      updatePos3(qgraphicsitem_cast<Array3DItem*>(item));
      break;
    case FrameItem::Type:
      updatePos3(qgraphicsitem_cast<FrameItem*>(item));
      break;
    case Fun3DItem::Type:
      updatePos3(qgraphicsitem_cast<Fun3DItem*>(item));
      break;
    default:
      updatePos(item, Snap * QPointF((m_scene->mapInv(item->pos()) /
                                      Snap).toPoint()));
  }
}
