/***************************************************************************
                          rootdlg.cpp  -  description
                             -------------------
    begin                : Sam Jan 25 2003
    copyright            : (C) 2017 by Werner Stille
    email                : kpl@werner-stille.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <QComboBox>
#include <QDesktopServices>
#include <QDialogButtonBox>
#include <QFile>
#include <QHeaderView>
#include <QKeyEvent>
#include <QLabel>
#include <QLayout>
#include <QPushButton>
#include <QTableWidget>
#include <QTextStream>
#include <KLocalizedString>
#include <KMessageBox>
#include "doubleedit.h"
#include "kpl.h"
#include "kpldoc.h"
#include "kplsettings.h"
#include "kplutils.h"
#include "rootdlg.h"
#include "splineitem.h"
#include "utils.h"

RootDlg::RootDlg(QWidget* parent, KplDoc* model, const SplineItem* spl0) :
 QDialog(parent), m(model), spl(spl0)
{
  setWindowTitle(i18n("Roots"));
  QVBoxLayout* vbox = new QVBoxLayout(this);
  QHBoxLayout* hbox = new QHBoxLayout;
  vbox->addLayout(hbox);
  hbox->addWidget(new QLabel(i18n("Condition")));
  hbox->addWidget(eCond = new QComboBox);
  eCond->addItems(QStringList() << "y(x) - c = 0" << "y'(x) - c = 0"
                                << "y\"(x) - c = 0" << i18n("Minimum")
                                << i18n("Maximum"));
  connect(eCond, SIGNAL(activated(int)), SLOT(findRoots()));
  hbox->addItem(new QSpacerItem(20, 10, QSizePolicy::MinimumExpanding));
  hbox->addWidget(new QLabel("c"));
  char format = KplDoc::presentation();
  int prec = KplSettings::precision();
  hbox->addWidget(eOff = new DoubleEdit(0, format, prec));
  eOff->setMinimumWidth(eOff->fontMetrics().width(QString::number(-1e-123 / 3,
                                                                  format,
                                                                  prec)) + 3);
  hbox->addWidget(update = new QPushButton(i18n("&Update")));
  connect(update, SIGNAL(clicked()), SLOT(findRoots()));
  vbox->addWidget(table = new QTableWidget(0, 4));
  table->setHorizontalHeaderLabels(QStringList() << "x" << "y(x)" << "y'(x)"
                                                 << "y\"(x)");
  table->setMinimumHeight(100);
  vbox->addWidget(buttonBox = new QDialogButtonBox(QDialogButtonBox::Help |
                                                   QDialogButtonBox::Close));
  buttonBox->addButton(bExport = new QPushButton(i18n("&Export")),
                       QDialogButtonBox::ActionRole);
  connect(bExport, SIGNAL(clicked()), SLOT(slotExport()));
  connect(buttonBox->button(QDialogButtonBox::Help), SIGNAL(clicked()),
          SLOT(slotHelp()));
  buttonBox->button(QDialogButtonBox::Close)->setFocus();
  connect(buttonBox, SIGNAL(rejected()), SLOT(reject()));
  findRoots();
}

RootDlg::~RootDlg()
{
}

void RootDlg::keyPressEvent(QKeyEvent* e)
{
  if ((e->QInputEvent::modifiers() == Qt::NoModifier) &&
      (e->key() == Qt::Key_F1)) {
    buttonBox->button(QDialogButtonBox::Help)->animateClick();
    e->accept();
  } else
    QDialog::keyPressEvent(e);
}

void RootDlg::findRoots()
{
  QVector<double> zero;
  int cond = eCond->currentIndex();
  bool extrema = (cond > 2);
  update->setDisabled(extrema);
  spl->roots(zero, extrema ? 1 : cond, extrema ? 0 : eOff->value());
  int nz = zero.count();
  int count = 0;
  if (nz) {
    table->setRowCount(nz);
    char format = KplDoc::presentation();
    int prec = KplSettings::precision();
    for (int i = 0; i < nz; ++i) {
      double y[3];
      for (int j = 0; j < 3; ++j)
        spl->calcTable(&zero[i], &y[j], 1, j);
      bool insert = true;
      switch (cond) {
        case 3:
          insert = (y[2] > 0);
          break;
        case 4:
          insert = (y[2] < 0);
      }
      if (insert) {
        QTableWidgetItem* item;
        table->setItem(count, 0,
                       item = new QTableWidgetItem(QString::number(zero[i],
                                                                   format,
                                                                   prec)));
        item->setFlags(Qt::ItemIsEnabled);
        for (int j = 0; j < 3; ++j) {
          table->setItem(count, j + 1,
                         item = new QTableWidgetItem(QString::number(y[j],
                                                                     format,
                                                                     prec)));
          item->setFlags(Qt::ItemIsEnabled);
        }
        ++count;
      }
    }
  }
  table->setRowCount(count);
  table->resizeColumnsToContents();
  int w = 0;
  for (int i = 0; i < 4; ++i)
    w = qMax(w, table->columnWidth(i));
  for (int i = 0; i < 3; ++i)
    table->setColumnWidth(i, w);
  table->horizontalHeader()->setStretchLastSection(true);
  bExport->setEnabled(count);
}

void RootDlg::slotExport()
{
  QUrl url;
  if (KplUtils::getWriteUrl(this, url, KplSettings::currentDir())) {
    QFile f(url.isLocalFile() ? url.path() : Utils::tempFileName());
    bool success = f.open(QIODevice::WriteOnly);
    if (success) {
      QTextStream ts(&f);
      for (int i = 0; i < table->rowCount(); ++i)
        for (int j = 0; j < 4; ++j)
          ts << table->item(i, j)->text()
             << ((j == 3) ? QString("\n") : QString(m->separator()));
      f.close();
      if (!url.isLocalFile())
        success = Utils::move(f.fileName(), url, this);
      if (success)
        KplUtils::setCurrentDir(url);
    }
    if (!success)
      KMessageBox::error(this, i18n("while trying to open"));
  }
}

void RootDlg::slotHelp()
{
  QDesktopServices::openUrl(QUrl(QStringLiteral("help:/kpl/roots.html")));
}
