/***************************************************************************
                          settingsdlg.h  -  description
                             -------------------
    begin                : Fr Jun 4 2004
    copyright            : (C) 2017 by Werner Stille
    email                : kpl@werner-stille.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef SETTINGSDLG_H
#define SETTINGSDLG_H

#include <KConfigDialog>
#include "funitem.h"

class QCheckBox;
class QComboBox;
class QDoubleSpinBox;
class QLineEdit;
class QSpinBox;
class QToolButton;
class KColorButton;
class KConfigSkeleton;
class KUrlRequester;
class KplDoc;
class BrushButton;
class DoubleEdit;
class SymbolButton;

/**
  * Edit fit parameter dialog class. Provides a dialog for the editing
  * the parameters of a function.
  * @author Werner Stille
  */
class EditFitParameterDlg : public QDialog
{
  Q_OBJECT
public:
  /**
   * Constructor.
   * @param parent pointer to parent widget.
   * @param p parameter array.
   * @param bFit array indicating whether a parameter is to be fitted or fixed.
   * @param file absolute path of the function library.
   * @param func function name.
   */
  EditFitParameterDlg(QWidget* parent, double* p, bool* bFit,
                      const QString& file, const QString& func);
  /** Destructor. */
  ~EditFitParameterDlg();

protected:
  bool* bFit;
  double* p;
  DoubleEdit* par[FunItem::NumParMax];
  QDialogButtonBox* buttonBox;
  QCheckBox* enFit[FunItem::NumParMax];

protected slots:
  /**
   * Handles button clicks.
   * @param button pointer to button.
   */
  virtual void slotButtonClicked(QAbstractButton* button);
};

/**
  * Kpl settings dialog class. Provides a dialog for basic, autoplot, and
  * autofit settings.
  * @author Werner Stille
  */
class SettingsDlg : public KConfigDialog
{
  Q_OBJECT
public:
  /**
   * Constructor.
   * @param parent pointer to parent widget.
   * @param name object name.
   * @param config pointer to KConfigSkeleton object.
   * @param model pointer to KplDoc model object.
   */
  SettingsDlg(QWidget* parent, const QString& name, KConfigSkeleton* config,
              KplDoc* model);
  /** Destructor. */
  ~SettingsDlg();

protected:
  /** Returns true if list of fit parameters was changed. */
  bool fitParChanged();
  /** Reimplemented from KConfigDialog. */
  virtual bool hasChanged();
  /** Reimplemented from KConfigDialog. */
  virtual bool isDefault();
  /**
   * Reads config settings.
   * @param defaults true for reading of default values.
   */
  void useDefaults(bool defaults);
  KplDoc* m;
  bool fitPar0[FunItem::NumParMax], fitPar[FunItem::NumParMax];
  FunItem::DataErrorStruct err0, err;
  double pFit0[FunItem::NumParMax], pFit[FunItem::NumParMax];
  QComboBox *kcfg_Presentation, *kcfg_Separator, *format,
            *kcfg_DecimalCharacter;
  QDoubleSpinBox *kcfg_FormatX, *kcfg_FormatY, *kcfg_X0, *kcfg_Y0, *kcfg_Width,
                 *kcfg_Height;
  QSpinBox *kcfg_Precision, *kcfg_FitSymbol, *kcfg_FitMaxIterations, *kcfg_IX,
           *kcfg_IY, *kcfg_IE, *kcfg_Symbol, *kcfg_RelSize;
  BrushButton *kcfg_FillStyle, *kcfg_FitFillStyle;
  SymbolButton  *bSymb, *bSymbFit;
  DoubleEdit *kcfg_FitTolerance;
  KUrlRequester* kcfg_FitPath;
  QLineEdit *kcfg_TextX, *kcfg_TextY, *kcfg_TextH, *kcfg_FitName;
  QCheckBox *kcfg_ShowSource, *autoPlot, *kcfg_AutoFit, *kcfg_AddItems,
            *kcfg_BoundingBox, *kcfg_SaveAbsPath, *kcfg_Warning,
            *kcfg_Multithreading, *saveAtEnd, *kcfg_AutoPathHeader, *kcfg_LogX,
            *kcfg_LogY, *kcfg_ErrBars, *kcfg_Normalization,
            *kcfg_FitErrorColumn, *kcfg_FitShowDialog, *kcfg_FitSaveParam,
            *kcfg_FitFollow, *kcfg_FitNonLinear;
  QButtonGroup* kcfg_GridMode;
  QPushButton *errMod;
  QToolButton *editH;
  KColorButton *kcfg_ColFrame, *kcfg_ColGrid, *kcfg_ColData, *kcfg_FitColor;
  int lFormat, gridMode0, gridMode;

protected slots:
  /**
   * Disables/enables error model button.
   * @param off disables button when true.
   */
  void disableErrMod(bool off);
  /**
   * Disables/enables heading widgets.
   * @param off disables widgets when true.
   */
  void disableHeading(bool off);
  /**
   * Disables/enables widgets for nonlinear fits.
   * @param on enables widgets when true.
   */
  void enableNonLin(bool on);
  /** Displays help page. **/
  virtual void showHelp();
  /** Displays dialog to edit the x axis label. */
  void slotEditLetX();
  /** Displays dialog to edit the y axis label. */
  void slotEditLetY();
  /** Displays dialog to edit the heading. */
  void slotEditLetH();
  /** Displays the dialog for parameter editing. */
  void slotEditParameter();
  /** Displays the dialog for setting the calculation of standard deviations. */
  void slotErrMod();
  /**
   * Adjusts state of widgets when fill pattern is changed.
   * @param brush new pattern number.
   */
  void slotFillChanged(int brush);
  /**
   * Adjusts state of widgets when fill pattern for autofit is changed.
   * @param brush new pattern number.
   */
  void slotFillFitChanged(int brush);
  /**
   * Sets page size.
   * @param i index of selected format box item.
   */
  void slotFormatChanged(int i);
  /**
   * Sets grid mode.
   * @param i index of selected grid mode.
   */
  void slotGridModeChanged(int i);
  /** Displays the dialog for selection of a function. */
  void slotFunc();
  /** Displays the dialog for selection of a parameter file to read. */
  void slotGetParameter();
  /**
   * Updates dialog according to logarithmic axis option.
   * @param state true for logarithmic axis.
   */
  void slotLogToggled(bool state);
  /**
   * Updates dialog according to autonormalization option.
   * @param state true autonormalization on.
   */
  void slotNormToggled(bool state);
  /** Restores window size. */
  void slotRestoreSize();
  /** Displays the dialog for selection of a parameter file to save. */
  void slotSaveParameter();
  /** Reimplemented from KConfigDialog. */
  virtual void updateSettings();
  /** Reimplemented from KConfigDialog. */
  virtual void updateWidgets();
  /** Reimplemented from KConfigDialog. */
  virtual void updateWidgetsDefault();
};

#endif
